package org.wso2.carbon.application.deployer.persistance;

import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.application.deployer.CarbonApplication;
import org.wso2.carbon.application.deployer.AppDeployerConstants;
import org.wso2.carbon.application.deployer.ApplicationManager;
import org.wso2.carbon.application.deployer.config.ApplicationConfiguration;
import org.wso2.carbon.application.deployer.config.Component;
import org.wso2.carbon.application.deployer.config.Association;
import org.wso2.carbon.application.deployer.internal.AppDeployerServiceComponent;
import org.wso2.carbon.CarbonConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;

import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.io.File;
import java.io.Reader;
import java.io.StringReader;


public class CarbonAppPersistanceManager {

    private static Registry configRegistry;
    private static Registry rootRegistry;

    private static final Log log = LogFactory.getLog(CarbonAppPersistanceManager.class);

    // Static block to initialize configRegistry
    static {
        try {
            configRegistry = AppDeployerServiceComponent
                    .getRegistryService().getConfigSystemRegistry();
            /**
             * When we deploy apps, if they contain registry resources, they provide the absolute
             * paths. And also we don't know whether those resources go into local, config or
             * governance registry. Therefore we have to user the root registry in that case.
             */
            rootRegistry = AppDeployerServiceComponent
                    .getRegistryService().getRegistry(CarbonConstants.REGISTRY_SYSTEM_USERNAME);
        } catch (Exception e) {
            log.error("Can't access RegistryService", e);
        }
    }

    /**
     * Load all existing Applications in the provided CarbonApplicationService instance
     *
     * @param appManager - ApplicationManager instance
     * @throws Exception - on registry transaction error
     */
    public static void loadApps(ApplicationManager appManager) throws Exception {
        try {
            //check whether applications are empty
            if (!configRegistry.resourceExists(AppDeployerConstants.APPLICATIONS)) {
                return;
            }

            Collection allApps = (Collection) configRegistry.get(AppDeployerConstants.APPLICATIONS);
            String[] appPaths = allApps.getChildren();

            for (String currentAppPath : appPaths) {
                CarbonApplication carbonApp = new CarbonApplication();

                //get application collection
                Collection currentAppCollection = (Collection) configRegistry.get(currentAppPath);
                carbonApp.setAppName(currentAppCollection.getProperty(AppDeployerConstants.NAME));
                carbonApp.setAppFilePath(currentAppCollection.
                        getProperty(AppDeployerConstants.APP_FILE_PATH));

                //load aar services
                fillList(carbonApp.getServices().get(CarbonApplication.AAR_SERVICES), currentAppPath
                        + AppDeployerConstants.APP_SERVICES + CarbonApplication.AAR_SERVICES);

                //load jaxws services
                fillList(carbonApp.getServices().get(CarbonApplication.JAXWS_SERVICES),
                        currentAppPath + AppDeployerConstants.APP_SERVICES +
                                CarbonApplication.JAXWS_SERVICES);

                //load sequences
                fillList(carbonApp.getSequences(),
                        currentAppPath + AppDeployerConstants.APP_SEQUENCES);

                //load proxy services
                fillList(carbonApp.getProxyServices(), currentAppPath
                        + AppDeployerConstants.APP_PROXY_SERVICES);

                //load endpoints
                fillList(carbonApp.getEndpoints(), currentAppPath
                        + AppDeployerConstants.APP_ENDPOINTS);

                //load local entries
                fillList(carbonApp.getLocalentries(), currentAppPath
                        + AppDeployerConstants.APP_LOCAL_ENTRIES);

                //load events
                fillList(carbonApp.getEvents(), currentAppPath + AppDeployerConstants.APP_EVENTS);

                //load tasks
                fillList(carbonApp.getTasks(), currentAppPath + AppDeployerConstants.APP_TASKS);

                appManager.addCarbonApp(carbonApp);
            }
        } catch (RegistryException e) {
            String msg = "Unable to load Applications. Registry transactions failed.";
            log.error(msg, e);
            throw new Exception(msg, e);
        }

    }

    /**
     * Persist the given Carbon app in the registry
     *
     * @param carbonApp - CarbonApplication instance
     * @throws Exception - on registry transaction error
     */
    public static void persistNewApp(CarbonApplication carbonApp) throws Exception {
        try {
            //if there is an already existing app in the configRegistry, delete it..
            deleteApplication(carbonApp.getAppName());

            String appResourcePath = AppDeployerConstants.APPLICATIONS + carbonApp.getAppName();

            Collection app = configRegistry.newCollection();
            app.setProperty(AppDeployerConstants.NAME, carbonApp.getAppName());
            app.setProperty(AppDeployerConstants.APP_FILE_PATH, carbonApp.getAppFilePath());

            //set last updated time as a property
            File appFile = new File(carbonApp.getAppFilePath());
            if (appFile.exists()) {
                app.setProperty(AppDeployerConstants.LAST_UPDATED_TIME,
                        String.valueOf(appFile.lastModified()));
            }

            //persist application
            configRegistry.put(appResourcePath, app);

            //persist aar services
            persistList(carbonApp.getServices().get(CarbonApplication.AAR_SERVICES),
                    appResourcePath + AppDeployerConstants.APP_SERVICES
                            + CarbonApplication.AAR_SERVICES + "/");

            //persist jaxws services
            persistList(carbonApp.getServices().get(CarbonApplication.JAXWS_SERVICES),
                    appResourcePath + AppDeployerConstants.APP_SERVICES
                            + CarbonApplication.JAXWS_SERVICES + "/");

            //persist sequences
            persistList(carbonApp.getSequences(),
                    appResourcePath + AppDeployerConstants.APP_SEQUENCES);

            //persist proxy services
            persistList(carbonApp.getProxyServices(),
                    appResourcePath + AppDeployerConstants.APP_PROXY_SERVICES);

            //persist endpoints
            persistList(carbonApp.getEndpoints(),
                    appResourcePath + AppDeployerConstants.APP_ENDPOINTS);

            //persist endpoints
            persistList(carbonApp.getLocalentries(),
                    appResourcePath + AppDeployerConstants.APP_LOCAL_ENTRIES);

            //persist endpoints
            persistList(carbonApp.getEvents(), appResourcePath + AppDeployerConstants.APP_EVENTS);

            //persist endpoints
            persistList(carbonApp.getTasks(), appResourcePath + AppDeployerConstants.APP_TASKS);

        } catch (RegistryException e) {
            String msg = "Unable to persist new Application : " + carbonApp.getAppName()
                    + " Registry transactions failed.";
            log.error(msg, e);
            throw new Exception(msg, e);
        }
    }

    public static String getLastUpdatedTime(String appName) throws Exception {
        try {
            String appResourcePath = AppDeployerConstants.APPLICATIONS + appName;
            //if the app exists in the configRegistry, read the property..
            if (configRegistry.resourceExists(appResourcePath)) {
                Resource app = configRegistry.get(appResourcePath);
                return app.getProperty(AppDeployerConstants.LAST_UPDATED_TIME);
            }
        } catch (RegistryException e) {
            String msg = "Unable to read last updated time of Application : " + appName
                    + ". Registry transactions failed.";
            log.error(msg, e);
            throw new Exception(msg, e);
        }
        return null;
    }

    /**
     * Delete the specified app from registry if already exists
     *
     * @param appName - application name
     * @throws Exception - on registry transaction error
     */
    public static void deleteApplication(String appName) throws Exception {
        try {
            String appResourcePath = AppDeployerConstants.APPLICATIONS + appName;

            //if the app exists in the configRegistry, delete it..
            if (configRegistry.resourceExists(appResourcePath)) {
                configRegistry.delete(appResourcePath);
            }
        } catch (RegistryException e) {
            String msg = "Unable to delete the Application : " + appName
                    + ". Registry transactions failed.";
            log.error(msg, e);
            throw new Exception(msg, e);
        }
    }

    /**
     * Writes all registry contents (resources, collections and associations) of all the
     * components in the given carbon application into the registry
     *
     * @param carbonApp - CarbonApplication instance
     */
    public static void writeAppResources(CarbonApplication carbonApp) {
        ApplicationConfiguration appConfig = carbonApp.getAppConfig();
        List<Component> components = appConfig.getComponents();

        HashMap<String, String> tempMap;
        for (Component comp : components) {
            // write collections
            tempMap = comp.getCollections();
            for (String key : tempMap.keySet()) {
                String dirPath = carbonApp.getExtractedPath() + comp.getId() +
                        AppDeployerConstants.RESOURCES_DIR + tempMap.get(key);

                // check whether the collection dir exists
                File file = new File(dirPath);
                if (!file.exists()) {
                    log.error("Specified collection directory not found at : " + dirPath);
                }
                writeFromDir(dirPath, key);
            }

            // write resources
            tempMap = comp.getResources();
            for (String key : tempMap.keySet()) {
                String filePath = carbonApp.getExtractedPath() + comp.getId() +
                        AppDeployerConstants.RESOURCES_DIR + tempMap.get(key);

                // check whether the file exists
                File file = new File(filePath);
                if (!file.exists()) {
                    log.error("Specified file to be written as a resource is " +
                            "not found at : " + filePath);
                }
                writeFromFile(file, key + "/" + tempMap.get(key));
            }

            // write associations
            List<Association> associations = comp.getAssociations();
            for (Association association : associations) {
                try {
                    rootRegistry.addAssociation(association.getSourcePath(),
                            association.getTargetPath(), association.getType());
                } catch (RegistryException e) {
                    log.error("Error while adding the association. Source path : " + association
                            .getSourcePath() + " Target path : " + association.getTargetPath());
                }
            }
        }
    }

    /**
     * Check in a directory in the file system as a collection in the registry. This file system
     * directory should be a registry checkout tatken earlier with the meta information.
     *
     * @param dirPath      - directory which contains registry resources
     * @param registryPath - path to check in
     */
    public static void writeFromDir(String dirPath, String registryPath) {
        // check the existance of the meta directory
        File file = new File(dirPath + "/" + CheckinUtils.META_DIRECTORY);
        if (!file.exists()) {
            log.error("Resources found at " + dirPath +
                    " are invalid. No meta information found.");
        }
        String directoryMetaFileName = dirPath + "/" + CheckinUtils.META_DIRECTORY + "/" +
                CheckinUtils.META_FILE_PREFIX + "." + CheckinUtils.META_FILE_EXTENSION;

        file = new File(directoryMetaFileName);
        if (!file.exists()) {
            log.error("Resources found at " + dirPath +
                    " are invalid. No meta information found.");
        }
        // which mean we have to upload with the whole directory..
        OMElement dumpedElement = CheckinUtils.createDirectoryMetaElement(dirPath, null);

        // do the restoring
        try {
            Reader reader = new StringReader(dumpedElement.toString());
            rootRegistry.restore(registryPath, reader);
        } catch (RegistryException e) {
            log.error("Error checking in resources to path: " + registryPath +
                    " from directory " + dirPath, e);
        }
    }

    /**
     * Write the file content as a registry resource to the given path
     *
     * @param file         - file to be written
     * @param registryPath - path to write the resource
     */
    public static void writeFromFile(File file, String registryPath) {
        // convert the file content into bytes and then encode it as a string
        byte[] content = CheckinUtils.getBytesFromFile(file);

        try {
            rootRegistry.beginTransaction();
            Resource resource = rootRegistry.newResource();
            resource.setContent(new String(content));
            rootRegistry.put(registryPath, resource);
            rootRegistry.commitTransaction();
        } catch (RegistryException e) {
            try {
                rootRegistry.rollbackTransaction();
            } catch (RegistryException e1) {
                log.error("Error while transaction rollback");
            }
            log.error("Error while checking in resource to path: " + registryPath +
                    " from file: " + file.getAbsolutePath(), e);
        }
    }

    /**
     * Persist a list of items. This can be a list of modules, sequences, proxyservices, endpoints
     * or services
     *
     * @param itemList - list of items
     * @param itemPath - path for persisting
     * @throws RegistryException - on registry errors
     */
    private static void persistList(ArrayList<String> itemList,
                                    String itemPath) throws RegistryException {
        for (String fileName : itemList) {
            Resource resource = configRegistry.newResource();
            resource.setProperty(AppDeployerConstants.FILE_NAME, fileName);
            resource.setContent(fileName);
            configRegistry.put(itemPath + fileName, resource);
        }
    }


    /**
     * Fills the item list. This can be a list of modules, sequences, proxyservices, endpoints
     * or services
     *
     * @param itemList     - item list to fill
     * @param resourcePath - registry collection path
     * @throws RegistryException - error on registry actions
     */
    private static void fillList(ArrayList<String> itemList,
                                 String resourcePath) throws RegistryException {
        if (configRegistry.resourceExists(resourcePath)) {
            Collection itemCollection = (Collection) configRegistry.get(resourcePath);
            String[] itemPaths = itemCollection.getChildren();
            for (String itemPath : itemPaths) {
                Resource moduleResource = configRegistry.get(itemPath);
                String fileName = moduleResource
                        .getProperty(AppDeployerConstants.FILE_NAME);
                itemList.add(fileName);
            }
        }
    }

}
