package org.wso2.carbon.application.deployer.persistance;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMText;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.om.util.Base64;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import java.io.*;
import java.net.URLDecoder;

public class CheckinUtils {

    private static final Log log = LogFactory.getLog(CheckinUtils.class);

    public static final String META_DIRECTORY = ".meta";
    public static final String META_FILE_PREFIX = "~";
    public static final String META_FILE_EXTENSION = "xml";
    public static final String MINE_FILE_POSTFIX = ".mine";
    public static final String SERVER_FILE_POSTFIX = ".server";
    public static final String EMBEDDED_REGISTRY_NAME = "embedded";

    /**
     * create the dumping element from the given directory. If the path is not given
     * it is retrieved from the meta file.
     *
     * @param directoryName - directory to create element
     * @param path          - path
     * @return - meta element
     */
    public static OMElement createDirectoryMetaElement(String directoryName,
                                                       String path) {
        // first get the meta file of the directory.
        String directoryMetaFileName = directoryName + "/" +
                META_DIRECTORY + "/" + META_FILE_PREFIX +
                "." + META_FILE_EXTENSION;

        // confirm the existance of the meta file.
        OMElement metaElement = getOMElmentFromMetafile(directoryMetaFileName,
                true,
                path);
        // will update the meta file in the file system asap it is retrieved.
        long now = System.currentTimeMillis();
        createMetaFile(directoryMetaFileName, metaElement, now);

        // retrieve the path element
        if (path == null) {
            path = metaElement.getAttributeValue(new QName("path"));
        }

        // now add the childs element to the meta element

        OMFactory factory = OMAbstractFactory.getOMFactory();

        OMElement childs = factory.createOMElement(new QName("children"));

        File directory = new File(directoryName);
        String[] children = directory.list();
        if (children == null) {
            // Either dir does not exist or is not a directory
        } else {
            for (int i = 0; i < children.length; i++) {
                // Get filename of file or directory
                String filename = children[i];
                String undecodedfilename = filename;
                filename = decodeFilename(filename);

                if (filename.equals(META_DIRECTORY)) {
                    continue;
                }
                if (filename.endsWith(MINE_FILE_POSTFIX) ||
                        filename.endsWith(SERVER_FILE_POSTFIX)) {
                    // there is an conflicts
                    log.error("Please resolve the conflicts before checkin. After resolving" +
                            " conflict please remove the files with " + MINE_FILE_POSTFIX +
                            " and " + SERVER_FILE_POSTFIX + " suffixes manually");
                }

                String childPath;
                if (path.equals("/")) {
                    childPath = "/" + filename;
                } else {
                    childPath = path + "/" + filename;
                }
                OMElement childEle = createMetaElementForChild(directoryName,
                        undecodedfilename, childPath);
                childs.addChild(childEle);
            }
        }
        metaElement.addChild(childs);

        return metaElement;
    }

    public static OMElement createMetaElementForChild(String parentDirectory,
                                                      String filename, String path) {
        File file = new File(parentDirectory + "/" + filename);

        OMElement metaElement;
        if (!file.isDirectory()) {

            String metaFilename = parentDirectory + "/" + META_DIRECTORY + "/" +
                    META_FILE_PREFIX + filename + "." + META_FILE_EXTENSION;

            // confirm the existance of the meta file.
            metaElement = getOMElmentFromMetafile(metaFilename, file.isDirectory(), path);

            // check whether the file is modified or not
            long realUpdateTime = file.lastModified();
            String markedUpdateTimeStr = metaElement.getAttributeValue(new QName("physicalModifiedDate"));
            if (markedUpdateTimeStr != null) {
                long markedUpdateTime = Long.parseLong(markedUpdateTimeStr);
                if (markedUpdateTime < realUpdateTime) {
                    // need to update the last update time
                    OMElement lastUpdatedElement =
                            metaElement.getFirstChildWithName(new QName("lastModified"));
                    String newUpdateTime = Long.toString(realUpdateTime);
                    lastUpdatedElement.setText(newUpdateTime);
                    metaElement.addAttribute("physicalModifiedDate", newUpdateTime, null);

                    //long now = System.currentTimeMillis();
                    createMetaFile(metaFilename, metaElement, file.lastModified());
                }
            } else {
                //long now = System.currentTimeMillis();
                createMetaFile(metaFilename, metaElement, file.lastModified());
            }

            // adding the content
            byte[] content = getBytesFromFile(file);
            String encodedContent = Base64.encode(content);

            OMFactory factory = OMAbstractFactory.getOMFactory();
            OMElement contentEle = factory.createOMElement(new QName("content"));
            OMText contentText = factory.createOMText(encodedContent);
            contentEle.addChild(contentText);
            metaElement.addChild(contentEle);
        } else {
            metaElement = createDirectoryMetaElement(parentDirectory + "/" + filename, path);
        }

        return metaElement;
    }

    public static OMElement getOMElmentFromMetafile(String metafilename,
                                                    boolean isCollection,
                                                    String path) {
        File metafile = new File(metafilename);
        OMElement element = null;
        // if the file exists, just get it.
        FileInputStream directoryMetaFileStream = null;
        if (metafile.exists()) {
            try {
                directoryMetaFileStream = new FileInputStream(metafilename);

                //create the parser
                XMLStreamReader parser = XMLInputFactory.newInstance()
                        .createXMLStreamReader(directoryMetaFileStream);
                //create the builder
                StAXOMBuilder builder = new StAXOMBuilder(parser);
                // get the element to restore
                element = builder.getDocumentElement().cloneOMElement();

                // need to set this at the first place it found this
//                String registryUrl = element.getAttributeValue(new QName("registryURL"));
//                ClientOptions.getClientOptions().setRegistryUrl(registryUrl);

//                String checkoutPath = element.getAttributeValue(new QName("checkoutPath"));
//                if (checkoutPath != null) {
//                    ClientOptions.getClientOptions().setCheckoutPath(checkoutPath);
//                }

                if (path != null && isCollection) {
                    // normally in collection the paths are not renamed.. so doing it explictely
                    element.addAttribute("path", path, null);
                }

            } catch (IOException e) {
                log.error("Error in reading the meta file: " + metafilename, e);
            } catch (XMLStreamException e) {
                log.error("Error in reading the stream of the meta file: " + metafilename, e);
            } finally {
                if (directoryMetaFileStream != null) {
                    try {
                        directoryMetaFileStream.close();
                    } catch (IOException e) {
                        log.error("Error in closing the stream for the meta file: " +
                                metafilename, e);
                    }
                }
            }
            return element;
        }

        if (path == null) {
            log.error("You can checkin only from a directory " +
                    "that is previously checkout. Please checkout a directory first");
        }

        // if not provide the default details
        OMFactory factory = OMAbstractFactory.getOMFactory();
        OMElement root = factory.createOMElement(new QName("resource"));

        // adding path as an attribute
        root.addAttribute("path", path, null);

        // adding isCollection as an attribute
        root.addAttribute("isCollection", isCollection ? "true" : "false", null);
        OMElement child;

        // guessing the media type by extension
        String mediaType = "unknown";
        int lastIndex = path.lastIndexOf('.');
        if (lastIndex > 0) {
            mediaType = path.substring(lastIndex + 1);
        }
        child = factory.createOMElement(new QName("mediaType"));
        child.setText(mediaType);
        root.addChild(child);

        // set creator
//        String creator = ClientOptions.getClientOptions().getUsername();
//        child = factory.createOMElement(new QName("creator"));
//        child.setText(creator);
//        root.addChild(child);

        // set createdTime
        long now = System.currentTimeMillis();
        child = factory.createOMElement(new QName("createdTime"));
        child.setText(Long.toString(now));
        root.addChild(child);

        // set updator
//        String updator = ClientOptions.getClientOptions().getUsername();
//        child = factory.createOMElement(new QName("lastUpdater"));
//        child.setText(updator);
//        root.addChild(child);

        // set LastModified
        child = factory.createOMElement(new QName("lastModified"));
        child.setText(Long.toString(now));
        root.addChild(child);

        // set Description
        child = factory.createOMElement(new QName("description"));
        root.addChild(child);

        return root;
    }

    /**
     * creating the file that store the meta data of the current directory or file
     *
     * @param fileName
     * @param modifiedDateAtCO
     */
    public static void createMetaFile(String fileName,
                                      OMElement metaData,
                                      long modifiedDateAtCO) {
        // keep the physical modified date of the file
        metaData.addAttribute("physicalModifiedDate", Long.toString(modifiedDateAtCO), null);

        try {
            File file = new File(fileName);

            String metaDirName = file.getParent();
            if (metaDirName != null) {
                File metaDir = new File(metaDirName);
                if (!metaDir.exists()) {
                    metaDir.mkdir();
                }
            }
            file.createNewFile();

            FileOutputStream fileOut = new FileOutputStream(file, false);
            metaData.serialize(fileOut);

            fileOut.close();

        } catch (IOException e) {
            log.error("Error in creating the meta file: " + fileName);
        } catch (XMLStreamException e) {
            log.error("Error in writing to the meta file: " + fileName);
        }

    }

    /**
     * Returns the contents of the file in a byte array.
     */
    public static byte[] getBytesFromFile(File file) {

        InputStream is = null;
        try {
            is = new FileInputStream(file);
        } catch (FileNotFoundException e) {
            log.error("Expected file: " + file.getName() + " Not found", e);
        }

        long length = file.length();

        // to ensure that file is not larger than Integer.MAX_VALUE.
        if (length > Integer.MAX_VALUE) {
            // File is too large
            log.error("File " + file.getName() + "is too large.");
        }

        // byte array to keep the data
        byte[] bytes = new byte[(int) length];

        int offset = 0;
        int numRead = 0;
        try {
            while (offset < bytes.length
                    && (numRead = is.read(bytes, offset, bytes.length - offset)) >= 0) {
                offset += numRead;
            }
        } catch (IOException e) {
            log.error("Error in reading data");
        }

        if (offset < bytes.length) {
            log.error("Could not completely read file " + file.getName());
        }
        try {
            is.close();
        } catch (IOException e) {
            log.error("Error in closing the stream");
        }
        return bytes;
    }

    public static String extensionToMediaType(String extension) {
        String mediaType;
        extension = extension.toLowerCase();
        if (extension.equals("jpg") ||
                extension.equals("gif") ||
                extension.equals("jpeg") ||
                extension.equals("png")) {
            mediaType = "img/" + extension;
        } else {
            mediaType = "extension";
        }
        return mediaType;
    }

    public static String decodeFilename(String path) {
        String decodedName = null;
        try {
            decodedName = URLDecoder.decode(path, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            log.error("Problem in decoding the resource path: " + path + ". ", e);
        }
        return decodedName;

    }
}
