/*
*  Copyright (c) 2005-2010, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
*
*  WSO2 Inc. licenses this file to you under the Apache License,
*  Version 2.0 (the "License"); you may not use this file except
*  in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing,
* software distributed under the License is distributed on an
* "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
* KIND, either express or implied.  See the License for the
* specific language governing permissions and limitations
* under the License.
*/

package org.wso2.carbon.application.deployer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMAttribute;
import org.osgi.framework.Bundle;
import org.wso2.carbon.application.deployer.config.Artifact;
import org.wso2.carbon.application.deployer.config.Association;
import org.wso2.carbon.application.deployer.config.CappFile;
import org.wso2.carbon.application.deployer.config.RegistryConfig;
import org.wso2.carbon.application.deployer.service.ApplicationManagerService;
import org.wso2.carbon.CarbonException;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.utils.CarbonUtils;
import org.wso2.carbon.utils.ServerConfiguration;
import org.wso2.carbon.utils.ServerConstants;

import javax.xml.namespace.QName;
import java.io.*;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

public class AppDeployerUtils {

    public static final String APP_UNZIP_DIR;
    private static String INTERNAL_ARTIFACTS_DIR = "internal-artifacts";

    static {
        String javaTempDir = System.getProperty("java.io.tmpdir");
        APP_UNZIP_DIR = javaTempDir.endsWith(File.separator) ?
                javaTempDir + AppDeployerConstants.CARBON_APPS :
                javaTempDir + File.separator + AppDeployerConstants.CARBON_APPS;
        createDir(APP_UNZIP_DIR);
    }

    private static final Log log = LogFactory.getLog(AppDeployerUtils.class);

    public static String getParentAppName(Bundle bundle) {
        Dictionary dictionary = bundle.getHeaders();
        if (dictionary != null) {

            // Iterate through the headers and find the ParentApplication custom header
            for (Enumeration e = dictionary.keys(); e.hasMoreElements();) {

                String headerKey = (String) e.nextElement();
                if (AppDeployerConstants.PARENT_APP_HEADER.equals(headerKey)) {
                    // retireve the header value
                    String headerValue = (String) dictionary.get(headerKey);
                    if (headerValue != null) {
                        return headerValue;
                    }
                }
            }
        }
        return null;
    }

    /**
     * Checking whether a bundle contains the WSO2-Application-Deployer header
     *
     * @param bundle - input bundle
     * @return - if found header - true, else - false
     */
    public static boolean isAppDeployer(Bundle bundle) {
        Dictionary dictionary = bundle.getHeaders();
        if (dictionary != null) {

            // Iterate through the headers and find the WSO2-Project-Artifact custom header
            for (Enumeration e = dictionary.keys(); e.hasMoreElements();) {

                String headerKey = (String) e.nextElement();
                if (AppDeployerConstants.WSO2_APP_DEPLOYER_HEADER.equals(headerKey)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Finds the carbon repository location
     *
     * @return - repo path
     */
    public static String getAxis2Repo() {
        String axis2Repo = CarbonUtils.getAxis2Repo();
        if (axis2Repo == null) {
            axis2Repo = CarbonUtils.getCarbonRepository();
        }
        return axis2Repo;
    }

    /**
     * Copy the artifact file at the fromPath to toPath
     *
     * @param fromPath - path at which the read the file to copy
     * @param toPath   - path to which the file should be copied
     */
    public static void copyFile(String fromPath, String toPath) {
        File in = new File(fromPath);
        if (!in.exists()) {
            log.error("Artifact file not found at : " + fromPath);
            return;
        }

        File out = new File(toPath);
        FileInputStream fis = null;
        FileOutputStream fos = null;
        try {
            fis = new FileInputStream(in);
            fos = new FileOutputStream(out);
            byte[] buf = new byte[10240];
            int i;
            while ((i = fis.read(buf)) != -1) {
                fos.write(buf, 0, i);
            }
        } catch (Exception e) {
            log.error("Error occured while copying artifact : " + fromPath, e);
        } finally {
            try {
                if (fis != null) fis.close();
                if (fos != null) fos.close();
            } catch (IOException e) {
                log.error("Error occured while closing the streams", e);
            }
        }
    }

    /**
     * Builds the Artifact object when an artifact element is given
     *
     * @param artifactEle - artifact OMElement
     * @return created Artifact object
     */
    public static Artifact populateArtifact(OMElement artifactEle) {
        if (artifactEle == null) {
            return null;
        }

        Artifact artifact = new Artifact();
        // read top level attributes
        artifact.setName(readAttribute(artifactEle, Artifact.NAME));
        artifact.setVersion(readAttribute(artifactEle, Artifact.VERSION));
        artifact.setType(readAttribute(artifactEle, Artifact.TYPE));
        artifact.setServerRole(readAttribute(artifactEle, Artifact.SERVER_ROLE));

        // read the dependencies
        Iterator itr = artifactEle.getChildrenWithLocalName(Artifact.DEPENDENCY);
        while (itr.hasNext()) {
            OMElement depElement = (OMElement) itr.next();
            // create an artifact for each dependency and add to the root artifact
            Artifact.Dependency dep = new Artifact.Dependency();
            dep.setServerRole(readAttribute(depElement, Artifact.SERVER_ROLE));
            dep.setName(readAttribute(depElement, Artifact.ARTIFACT));
            dep.setVersion(readAttribute(depElement, Artifact.VERSION));
            artifact.addDependency(dep);
        }

        // read the subArtifacts
        OMElement subArtifactsElement = artifactEle
                .getFirstChildWithName(new QName(Artifact.SUB_ARTIFACTS));
        if (subArtifactsElement != null) {
            Iterator subArtItr = subArtifactsElement.getChildrenWithLocalName(Artifact.ARTIFACT);
            while (subArtItr.hasNext()) {
                // as this is also an artifact, use recursion
                Artifact subArtifact = populateArtifact((OMElement) subArtItr.next());
                artifact.addSubArtifact(subArtifact);
            }
        }

        // read the files
        Iterator fileItr = artifactEle.getChildrenWithLocalName(Artifact.FILE);
        while (fileItr.hasNext()) {
            OMElement fileElement = (OMElement) fileItr.next();
            CappFile tempFile = new CappFile();
            tempFile.setName(fileElement.getText());
            tempFile.setVersion(readAttribute(fileElement, Artifact.VERSION));
            artifact.addFile(tempFile);            
        }

        return artifact;
    }

    /**
     * Builds a RegistryConfig instance using an OMElement which is built using the registry
     * config file inside and cApp Registry/Resource artifact
     *
     * @param resourcesElement - regConfig element
     * @return - RegistryConfig instance built
     */
    public static RegistryConfig populateRegistryConfig(OMElement resourcesElement) {
        if (resourcesElement == null) {
            return null;
        }

        RegistryConfig regConfig = new RegistryConfig();

        // read Item elements under Resources
        Iterator itemItr = resourcesElement.getChildrenWithLocalName(Artifact.ITEM);
        while (itemItr.hasNext()) {
            OMElement itemElement = (OMElement) itemItr.next();
            regConfig.addResource(readChildText(itemElement, Artifact.PATH),
                    readChildText(itemElement, Artifact.FILE));
        }

        // read Collection elements under Resources
        Iterator collectionItr = resourcesElement.getChildrenWithLocalName(Artifact.COLLECTION);
        while (collectionItr.hasNext()) {
            OMElement collElement = (OMElement) collectionItr.next();
            regConfig.addCollection(readChildText(collElement, Artifact.PATH),
                    readChildText(collElement, Artifact.DIRECTORY));
        }

        // read Association elements under Resources
        Iterator associationItr = resourcesElement.getChildrenWithLocalName(Artifact.ASSOCIATION);
        while (associationItr.hasNext()) {
            OMElement assoElement = (OMElement) associationItr.next();
            Association association = new Association();
            association.setSourcePath(readChildText(assoElement, Artifact.SOURCE_PATH));
            association.setTargetPath(readChildText(assoElement, Artifact.TARGET_PATH));
            association.setType(readChildText(assoElement, Artifact.TYPE));
            regConfig.addAssociation(association);
        }
        return regConfig;
    }

    /**
     * Reads an attribute in the given element and returns the value of that attribute
     *
     * @param element - Element to search
     * @param attName - attribute name
     * @return if the attribute found, return value. else null.
     */
    public static String readAttribute(OMElement element, String attName) {
        if (element == null) {
            return null;
        }
        OMAttribute temp = element.getAttribute(new QName(attName));
        if (temp != null) {
            return temp.getAttributeValue();
        }
        return null;
    }

    public static String readChildText(OMElement element, String ln) {
        return readChildText(element, ln, null);
    }

    /**
     * Reads a text node which is in a child element of the given element and returns the text
     * value.
     *
     * @param element - Element to search
     * @param ln - Child element name
     * @param ns - Child element namespace
     * @return if the child text element found, return text value. else null.
     */
    public static String readChildText(OMElement element, String ln, String ns) {
        if (element == null) {
            return null;
        }
        OMElement temp = element.getFirstChildWithName(new QName(ns, ln));
        if (temp != null) {
            return temp.getText();
        }
        return null;
    }

    /**
     * First checks for the "serverRoles" system property. If null, reads the ServerRoles property
     * from carbon.xml.
     *
     * @return server roles array
     */
    public static String[] readServerRoles() {
        String[] serverRoles;
        // read the system property
        String temp = System.getProperty(AppDeployerConstants.SERVER_ROLES_CMD_OPTION);
        if (temp != null) {
            serverRoles = temp.split(",");
        } else {
            // now try to read from carbon.xml
            ServerConfiguration serverConfig = ServerConfiguration.getInstance();
            serverRoles = serverConfig.getProperties(AppDeployerConstants.CARBON_SERVER_ROLE);
        }
        return serverRoles;
    }

    /**
     * Computes the application artifact file path when the bundle is given
     *
     * @param b - App artifact as an OSGi bundle
     * @return - App file path
     */
    public static String getArchivePathFromBundle(Bundle b) {
        //compute bundle file path
        String bundlePath = b.getLocation();
        bundlePath = formatPath(bundlePath);
        return getComponentsRepo() + File.separator +
                bundlePath.substring(bundlePath.lastIndexOf('/') + 1);
    }

    /**
     * Finds repo/carbonapps path
     *
     * @return - path
     */
    public static String getApplicationLocation() {
        return getAxis2Repo() + File.separator + AppDeployerConstants.CARBON_APPS;
    }

    /**
     * Finds teh extension of a given file
     *
     * @param fileName - name of the file
     * @return - extension
     */
    public static String getFileExtension(String fileName) {
        int index = fileName.lastIndexOf('.');
        return fileName.substring(index + 1);
    }

    /**
     * Extract the Carbon application at the provided path to the java temp dir. Return the
     * extracted location
     *
     * @param appCarPath - Absolute path of the Carbon application .car file
     * @return - extracted location
     * @throws org.wso2.carbon.CarbonException - error on extraction
     */
    public static String extractCarbonApp(String appCarPath) throws CarbonException {
        appCarPath = formatPath(appCarPath);
        String fileName = appCarPath.substring(appCarPath.lastIndexOf('/') + 1);
        String dest = APP_UNZIP_DIR + File.separator + System.currentTimeMillis() +
                fileName + File.separator;
        createDir(dest);

        try {
            extract(appCarPath, dest);
        } catch (IOException e) {
            throw new CarbonException("Error while extracting Carbon Application : " + fileName, e);
        }
        return dest;
    }

    public static String createAppExtractionPath(String parentAppName) {
        String parentPath = APP_UNZIP_DIR + File.separator + System.currentTimeMillis() +
                parentAppName + File.separator;
        createDir(parentPath);
        return parentPath;
    }

    /**
     * Extract an individual cApp artifact at the provided path to the java temp dir. Return the
     * extracted location
     *
     * @param artifactPath - Absolute path of the cApp artifact file
     * @param parentPath - Parent's extracted path
     * @return - extracted location
     * @throws org.wso2.carbon.CarbonException - on error
     */
    public static String extractAppArtifact(String artifactPath,
                                            String parentPath) throws CarbonException {

        createDir(parentPath + INTERNAL_ARTIFACTS_DIR);

        String fileName = artifactPath.substring(artifactPath.lastIndexOf('/') + 1);
        String dest = parentPath + INTERNAL_ARTIFACTS_DIR + File.separator +
                fileName.substring(0, fileName.lastIndexOf(".")) + File.separator;
        createDir(dest);

        try {
            extract(artifactPath, dest);
        } catch (IOException e) {
            throw new CarbonException("Error while extracting cApp artifact : " + fileName, e);
        }
        return dest;
    }

    /**
     * Finds the owner application of the provided artifact file name and sets the runtime
     * object name in the corresponding artifact.
     *
     * @param appManager   - ApplicationManager instance
     * @param fileName     - file name of the artifact
     * @param artifactType - this can be a module or a service
     * @param runtimeObjectName - name of the runtime object corresponding to this file
     */
    public static void attachArtifactToOwnerApp(ApplicationManagerService appManager,
                                                String fileName,
                                                String artifactType,
                                                String runtimeObjectName) {
        if (appManager == null || fileName == null || artifactType == null) {
            return;
        }
        Artifact appArtifact;
        for (CarbonApplication carbonApp : appManager.getCarbonApps()) {
            appArtifact = carbonApp.getAppConfig().getApplicationArtifact();
            for (Artifact.Dependency dep : appArtifact.getDependencies()) {
                if (dep.getArtifact() != null) {
                    Artifact depArtifact = dep.getArtifact();
                    for (CappFile file : depArtifact.getFiles()) {
                        if (file.getName().equals(fileName) &&
                                depArtifact.getType().equals(artifactType)) {
                            depArtifact.setRuntimeObjectName(runtimeObjectName);
                        }
                    }
                }
            }
        }
    }

    /**
     * Format the string paths to match any platform.. windows, linux etc..
     *
     * @param path - input file path
     * @return formatted file path
     */
    public static String formatPath(String path) {
        // removing white spaces
        path = path.replaceAll("\\b\\s+\\b", "%20");
        // replacing all "\" with "/"
        return path.replace('\\', '/');
    }

    public static String getOSGiDropinsDir() {
        return CarbonUtils.getCarbonOSGIRepoDir() +
                File.separator + AppDeployerConstants.CARBON_DROPINS;
    }

    /**
     * Checks whether the given dependencies has library type artifacts
     * 
     * @param deps - list of dependencies
     * @return - true if found..
     */
    public static boolean hasLibs(List<Artifact.Dependency> deps) {
        for (Artifact.Dependency dep : deps) {
            Artifact artifact = dep.getArtifact();
            if (artifact != null) {
                if (artifact.getType().startsWith("lib/")) {
                    return true;
                } else {
                    if(hasLibs(artifact.getDependencies())) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    private static void extract(String sourcePath, String destPath) throws IOException {
        Enumeration entries;
        ZipFile zipFile;

        zipFile = new ZipFile(sourcePath);
        entries = zipFile.entries();

        while (entries.hasMoreElements()) {
            ZipEntry entry = (ZipEntry) entries.nextElement();
            // we don't need to copy the META-INF dir
            if (entry.getName().startsWith("META-INF/")) {
                continue;
            }
            // if the entry is a directory, create a new dir
            if (entry.isDirectory()) {
                createDir(destPath + entry.getName());
                continue;
            }
            // if the entry is a file, write the file
            copyInputStream(zipFile.getInputStream(entry),
                    new BufferedOutputStream(new FileOutputStream(destPath + entry.getName())));
        }
        zipFile.close();
    }

    public static void createDir(String path) {
        File temp = new File(path);
        if (!temp.exists() && !temp.mkdir()) {
            log.error("Error while creating directory : " + temp);
        }
    }

    private static void copyInputStream(InputStream in, OutputStream out)
            throws IOException {
        byte[] buffer = new byte[40960];
        int len;

        while ((len = in.read(buffer)) >= 0)
            out.write(buffer, 0, len);

        in.close();
        out.close();
    }

    /**
     * CarbonUtils.getComponentsRepo method returns a wrong location when the system property
     * or env variable is null. Therefore, using this method. As this is a patch, we can't
     * edit carbon.utils bundle
     *
     * @return - carbon components repo location
     */
    private static String getComponentsRepo() {
        String componentsRepo = System.getProperty(ServerConstants.COMPONENT_REP0);
        if (componentsRepo == null) {
            componentsRepo = System.getenv(CarbonConstants.COMPONENT_REP0_ENV);
            if (componentsRepo == null) {
                return CarbonUtils.getCarbonHome() + File.separator +
                        ServerConstants.FOLDER_NAME_REPOSITORY + File.separator +
                        ServerConstants.FOLDER_NAME_COMPONENTS + File.separator +
                        AppDeployerConstants.CARBON_PLUGINS;
            }
        }
        return componentsRepo;
    }
}
