/*
*  Copyright (c) 2005-2010, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
*
*  WSO2 Inc. licenses this file to you under the Apache License,
*  Version 2.0 (the "License"); you may not use this file except
*  in compliance with the License.
*  You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing,
* software distributed under the License is distributed on an
* "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
* KIND, either express or implied.  See the License for the
* specific language governing permissions and limitations
* under the License.
*/

package org.wso2.carbon.application.deployer.handler;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.application.deployer.config.Artifact;
import org.wso2.carbon.application.deployer.config.CappFile;
import org.wso2.carbon.application.deployer.CarbonApplication;
import org.wso2.carbon.application.deployer.AppDeployerUtils;
import org.wso2.carbon.application.deployer.internal.AppDeployerServiceComponent;
import org.osgi.framework.BundleException;
import org.osgi.framework.Bundle;

import java.util.List;
import java.io.File;

/**
 * This is one of the default handlers which is registered into the ApplicationManager. This
 * class deploys AAR services, JAXWS services, Data services and libs.
 */
public class DefaultAppDeployer implements AppDeploymentHandler {

    private static final Log log = LogFactory.getLog(DefaultAppDeployer.class);

    public static final String AAR_TYPE = "service/axis2";
    public static final String JAXWS_TYPE = "service/jaxws";
    public static final String DS_TYPE = "service/dataservice";

    public static final String AAR_DIR = "axis2services";
    public static final String JAXWS_DIR = "servicejars";
    public static final String DS_DIR = "dataservices";

    /**
     * Deploy the artifacts which can be deployed through this deployer (Axis2 services,
     * JAXWS services ..).
     *
     * @param carbonApp - find artifacts from this CarbonApplication instance
     */
    public void deployArtifacts(CarbonApplication carbonApp) {
        if (log.isDebugEnabled()) {
            log.debug("Default App Deployer deploying artifacts of cApp: " + carbonApp.getAppName());
        }

        List<Artifact.Dependency> dependencies = carbonApp.getAppConfig().getApplicationArtifact()
                .getDependencies();
        deployRecursively(dependencies);
    }

    /**
     * Installs an OSGi bundle into the OSGi environment through the bundle context..
     *
     * @param bundlePath - absolute path to the bundle to be installed..
     */
    private void installBundle(String bundlePath) {
        bundlePath = AppDeployerUtils.formatPath(bundlePath);

        // prepare the URL
        if (bundlePath.startsWith("/")) {
            // on linux
            bundlePath = "file://" + bundlePath;
        } else {
            // on windows
            bundlePath = "file:///" + bundlePath;
        }
        try {
            Bundle bundle = AppDeployerServiceComponent
                    .getBundleContext().installBundle(bundlePath);
            bundle.start();
        } catch (BundleException e) {
            log.error("Error while installing bundle : " + bundlePath);
        }
    }

    /**
     * Each artifact can have it's dependencies which are also artifacts. This method searches
     * the entire tree of artifacts to deploy default types..
     *
     * @param deps - list of dependencies to be searched..
     */
    private void deployRecursively(List<Artifact.Dependency> deps) {
        String artifactPath, destPath;
        for (Artifact.Dependency dependency : deps) {
            Artifact artifact = dependency.getArtifact();
            if (artifact == null) {
                continue;
            }
            if (AAR_TYPE.equals(artifact.getType())) {
                destPath = AppDeployerUtils.getAxis2Repo() + File.separator + AAR_DIR;
            } else if (JAXWS_TYPE.equals(artifact.getType())) {
                destPath = AppDeployerUtils.getAxis2Repo() + File.separator + JAXWS_DIR;
            } else if (DS_TYPE.equals(artifact.getType())) {
                destPath = AppDeployerUtils.getAxis2Repo() + File.separator + DS_DIR;
            } else if (artifact.getType().startsWith("lib/")) {
                destPath = AppDeployerUtils.getOSGiDropinsDir();
            } else {
                continue;
            }

            List<CappFile> files = artifact.getFiles();
            if (files.size() != 1) {
                log.error(artifact.getType() + " type must have a single file to " +
                        "be deployed. But " + files.size() + " files found.");
                continue;
            }
            String fileName = artifact.getFiles().get(0).getName();
            artifactPath = artifact.getExtractedPath() + File.separator + fileName;
            AppDeployerUtils.createDir(destPath);

            String destFilePath = destPath + File.separator + fileName;
            AppDeployerUtils.copyFile(artifactPath, destFilePath);

            /**
             * if the current artifact is a lib, we have to manually install it into the OSGi
             * environment for the usage of the lib before the first restart..
             */
            if (artifact.getType().startsWith("lib/")) {
                installBundle(destFilePath);
                artifact.setRuntimeObjectName(fileName);
            }
            deployRecursively(artifact.getDependencies());
        }
    }
}
