/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.services.authentication;

import org.apache.axis2.context.MessageContext;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.authenticator.proxy.AuthenticationAdminClient;
import org.wso2.carbon.core.AbstractAdmin;
import org.wso2.carbon.core.common.AuthenticationException;
import org.wso2.carbon.core.services.internal.CarbonServicesServiceComponent;
import org.wso2.carbon.core.services.util.CarbonAuthenticationUtil;
import org.wso2.carbon.core.util.AnonymousSessionUtil;
import org.wso2.carbon.registry.core.service.RegistryService;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.service.RealmService;
import org.wso2.carbon.user.core.util.UserCoreUtil;
import org.wso2.carbon.utils.ServerConfiguration;
import org.wso2.carbon.utils.ServerConstants;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

/**
 * Admin service to manage operations on the <code>AxisConfiguration</code>
 */
public class AuthenticationAdmin extends AbstractAdmin implements CarbonServerAuthenticator {

    private static final Log log = LogFactory.getLog(AuthenticationAdmin.class);
    protected static final String AUTHENTICATION_ADMIN_SERVICE = "AuthenticationAdminService";
    private static final int DEFAULT_PRIORITY_LEVEL = 5;
    private static final String AUTHENTICATOR_NAME = "Authenticators.DefaultCarbonAuthenticator";
    private static final String AUTHENTICATOR_PRIOROTY = AUTHENTICATOR_NAME + ".Priority";
    private static final String AUTHENTICATOR_DISABLED = AUTHENTICATOR_NAME + ".Disabled";

    public boolean login(String username, String password, String remoteAddress)
            throws AuthenticationException {
        HttpSession httpSession = getHttpSession();

        try {

            if ((username == null) || (password == null) || (remoteAddress == null)
                    || username.trim().equals("") || password.trim().equals("")
                    || remoteAddress.trim().equals("")) {
                CarbonAuthenticationUtil.onFailedAdminLogin(httpSession, username, -1,
                        remoteAddress, "Data");
                return false;
            }

            RegistryService registryService = CarbonServicesServiceComponent.getRegistryService();
            RealmService realmService = CarbonServicesServiceComponent.getRealmService();

            String tenantDomain = UserCoreUtil.getTenantDomain(realmService, username);
            String userNameWithDomain = username;
            username = UserCoreUtil.getTenantLessUsername(username);
            UserRealm realm = AnonymousSessionUtil.getRealmByTenantDomain(registryService,
                    realmService, tenantDomain);
            boolean isAuthenticated = false;
            ServerConfiguration serverConfig = CarbonServicesServiceComponent
                    .getServerConfiguration();
            if (serverConfig.getFirstProperty(CarbonConstants.IS_REMOTE_AUTHENTICATION) != null) {
                String serverUrl = serverConfig
                        .getFirstProperty(CarbonConstants.IS_REMOTE_AUTHENTICATION);
                AuthenticationAdminClient client = new AuthenticationAdminClient(
                        CarbonServicesServiceComponent.getConfigurationContextService()
                                .getClientConfigContext(), serverUrl, null, null, false);
                isAuthenticated = client.login(userNameWithDomain, password, remoteAddress);
            } else {
                isAuthenticated = realm.getUserStoreManager().authenticate(username, password);
            }
            boolean isAuthorized = realm.getAuthorizationManager().isUserAuthorized(username,
                    "/permission/admin/login", CarbonConstants.UI_PERMISSION_ACTION);
            int tenantId = realmService.getTenantManager().getTenantId(tenantDomain);

            if (isAuthenticated && isAuthorized) {
                CarbonAuthenticationUtil.onSuccessAdminLogin(httpSession, username,
                        tenantId, tenantDomain, remoteAddress);
                return true;
            } else {
                CarbonAuthenticationUtil.onFailedAdminLogin(httpSession, username, tenantId,
                        remoteAddress, "Invalid credential");
                return false;
            }
        } catch (Exception e) {
            String msg = "System error while Authenticating/Authorizing User : " + e.getMessage();
            log.error(msg, e);
            return false;
        }
    }

    public void logout() throws AuthenticationException {
        String loggedInUser;
        String delegatedBy;
        Date currentTime = Calendar.getInstance().getTime();
        SimpleDateFormat date = new SimpleDateFormat("'['yyyy-MM-dd HH:mm:ss,SSSS']'");
        HttpSession session = getHttpSession();

        if (session != null) {
            loggedInUser = (String) session.getAttribute(ServerConstants.ADMIN_USERNAME);
            delegatedBy = (String) session.getAttribute("DELEGATED_BY");
            if (delegatedBy == null) {
                log.info("'" + loggedInUser + "' logged out at " + date.format(currentTime));
            } else {
                log.info("'" + loggedInUser + "' logged out at " + date.format(currentTime)
                        + " delegated by " + delegatedBy);
            }
            session.invalidate();
        }
    }

    public String getAuthenticatorName() {
        return AUTHENTICATOR_NAME;
    }

    public int getPriority() {
        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        String priority = serverConfig.getFirstProperty(AUTHENTICATOR_PRIOROTY);
        if (priority != null && priority.length() > 0) {
            return Integer.parseInt(priority);
        }
        return DEFAULT_PRIORITY_LEVEL;
    }

    public boolean isAuthenticated(MessageContext messageContext) {
        HttpServletRequest request = (HttpServletRequest) messageContext
                .getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
        HttpSession httpSession = request.getSession();
        String loginStatus = (String) httpSession.getAttribute(ServerConstants.ADMIN_LOGGED_IN);

        if (loginStatus == null) {
            return false;
        }
        boolean isAuthenticated = "true".equals(loginStatus);
        return isAuthenticated;
    }

    public boolean isDisabled() {
        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        String isDisabled = serverConfig.getFirstProperty(AUTHENTICATOR_DISABLED);
        if (isDisabled != null && isDisabled.length() > 0) {
            return Boolean.getBoolean(isDisabled);
        }
        return false;
    }

    public boolean isHandle(MessageContext msgContext) {
        return true;
    }
}
