/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.dao.ResourceVersionDAO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.jdbc.utils.VersionRetriever;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of CollectionImpl, Instance of this class is returned when requested for an early
 * version of some collection.
 */
public class CollectionVersionImpl extends CollectionImpl {

    private static Log log = LogFactory.getLog(CollectionVersionImpl.class);

    private ResourceVersionDAO resourceVersionDAO;
    private VersionRetriever versionList;
    private int versionListIndex = 0;

    /**
     * Default Constructor. Creates an empty CollectionVersion instance.
     */
    public CollectionVersionImpl() {
        resourceVersionDAO = new ResourceVersionDAO();
    }

    /**
     * Constructor a CollectionVersionImpl for a provided path and a resourceDO.
     *
     * @param path       the path of the collection.
     * @param resourceDO the resourceDO instance.
     */
    public CollectionVersionImpl(String path, ResourceDO resourceDO) {
        super(path, resourceDO);
        resourceVersionDAO = new ResourceVersionDAO();
    }

    /**
     * Method to set the version retriever instance.
     *
     * @param versionList the version retriever.
     */
    public void setVersionList(VersionRetriever versionList) {
        this.versionList = versionList;
    }

    /**
     * Method to set the index of the current collection in the version list
     *
     * @param versionListIndex the index of the current collection in the list.
     */
    public void setVersionListIndex(int versionListIndex) {
        this.versionListIndex = versionListIndex;
    }

    /**
     * Method to return the absolute paths of the children of the collection. All the returning
     * paths will be in the same snapshot as the current collection.
     *
     * @return the array of absolute paths of the children
     * @throws RegistryException throws if the operation fail.
     */
    public String[] getChildren() throws RegistryException {
        if (getContent() instanceof String[]) {
            return fixPaths((String[])getContent());
        } else {
            return new String[0];
        }
    }

    /**
     * Method to return the paths of the selected range of children.
     *
     * @param start   the starting number of children.
     * @param pageLen the number of entries to retrieve.
     *
     * @return an array of paths of the selected range of children.
     * @throws RegistryException throws if the operation fail.
     */
    public String[] getChildren(int start, int pageLen) throws RegistryException {
        if (content != null && content instanceof String[]) {
            List<String> childPathList = new ArrayList<String>();
            String childNodes[] = (String[]) content;
            int limit = start + pageLen;
            if (start > childNodes.length) {
                return new String[0];
            }
            if (limit > childNodes.length) {
                limit = childNodes.length;
            }
            for (int i = start; i < limit; i++) {
                String childNode = childNodes[i];
                childPathList.add(childNode);
            }
            return fixPaths(childPathList.toArray(new String[childPathList.size()]));
        }
        setSessionInformation();
        try {
            String[] childPaths = new String[0];

            if (Transaction.isStarted()) {
                childPaths = resourceVersionDAO.getChildPaths(this.getResourceIDImpl(),
                        versionList, versionListIndex,
                        start, pageLen, snapshotID, Transaction.getConnection());

            } else {

                Connection conn = null;
                boolean transactionSucceeded = false;
                try {
                    conn = dataSource.getConnection();

                    // If a managed connection already exists, use that instead of a new connection.
                    Connection temp = Transaction.getManagedRegistryConnection(conn);
                    if (temp != null) {
                        conn.close();
                        conn = temp;
                    }
                    if (conn.getTransactionIsolation() != Connection.TRANSACTION_READ_COMMITTED) {
                        conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
                    }
                    conn.setAutoCommit(false);

                    childPaths = resourceVersionDAO.getChildPaths(this.getResourceIDImpl(),
                            versionList, versionListIndex,
                            start, pageLen, snapshotID, conn);
                    transactionSucceeded = true;
                } catch (SQLException e) {

                    String msg = "Failed to get the child paths " + pageLen + " child paths from " +
                            start + " of resource " + getPath() + ". " + e.getMessage();
                    log.error(msg, e);
                    throw new RegistryException(msg, e);

                } finally {
                    if (transactionSucceeded) {
                        try {
                            conn.commit();
                        } catch (SQLException e) {
                            log.error("Failed to commit the database connection used in " +
                                    "getting child paths of the collection " + getPath());
                        }
                    } else {
                        try {
                            conn.rollback();
                        } catch (SQLException e) {
                            log.error("Failed to rollback the database connection used in " +
                                    "getting child paths of the collection " + getPath());
                        }
                    }

                    if (conn != null) {
                        try {
                            conn.close();
                        } catch (SQLException e) {
                            log.error("Failed to close the database connection used in " +
                                    "getting child paths of collection " + getPath());
                        }
                    }
                }
            }

            return fixPaths(childPaths);
        } finally {
            clearSessionInformation();
        }
    }

    /**
     * Method to return the the number of children.
     *
     * @return the number of children.
     * @throws RegistryException throws if the operation fail.
     */
    public int getChildCount() throws RegistryException {
        if (content != null) {
            String[] childPaths = (String[]) content;
            return fixPaths(childPaths).length;
        }
        String[] childPaths = getChildren(0, -1);
        if (childPaths == null) {
            return 0;
        }
        return childPaths.length;
    }

    /**
     * Collection's content is a string array, which contains paths of its children. These paths are
     * loaded on demand to increase performance. It is recommended to use {@link #getChildren()}
     * method to get child paths of a collection, which provides pagination. Calling this method
     * will load all child paths.
     *
     * @return String array of child paths.
     * @throws RegistryException On any error.
     */
    public Object getContent() throws RegistryException {
        if (content == null) {
            String[] childPaths = getChildren(0, -1);
            setContentWithNoUpdate(childPaths);
        }
        return content;
    }    
}
