/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import net.sf.ehcache.Statistics;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.utils.DBUtils;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * Implementation of caching for resource paths. This is based on {@link Cache}.
 */
public class PathCache {

    private static PathCache pathCacheObject;
    private Cache pathCache = null;
    private static final Log log = LogFactory.getLog(ResourceDAO.class);

    private PathCache(int maxElementsInMemory, boolean eternal, long timeToLiveSeconds) {
        CacheManager cacheManager = new CacheManager();
        pathCache = new Cache(RegistryConstants.PATH_CACHE_ID,
                maxElementsInMemory, false, eternal, timeToLiveSeconds, 0);
        cacheManager.addCache(pathCache);
    }

    /**
     * Creates a object storing a set of cached paths.
     *
     * @param maxElementsInMemory maximum number of cached elements in memory.
     * @param eternal             whether the elements in the cache are eternal, i.e. never expire
     * @param timeToLiveSeconds   the default amount of time to live for an element from its
     *                            creation date
     *
     * @return the cache object created
     * @see Cache
     */
    public static synchronized PathCache createPathCache(int maxElementsInMemory,
                                                         boolean eternal,
                                                         long timeToLiveSeconds) {

        return new PathCache(maxElementsInMemory, eternal, timeToLiveSeconds);
    }

    /**
     * Method to obtain the created path cache instance.
     *
     * @return the created path cache instance
     */
    public static PathCache getPathCache() {
        if (pathCacheObject == null) {
            pathCacheObject = createPathCache(RegistryConstants.DEFAULT_PATH_CACHE_SIZE,
                    RegistryConstants.DEFAULT_PATH_CACHE_IS_ETERNAL,
                    RegistryConstants.DEFAULT_PATH_CACHE_TIME_TO_LIVE);
        }
        return pathCacheObject;
    }

    /**
     * Method to get the cache hit rate.
     *
     * @return the cache hit rate.
     */
    public double hitRate() {
        Statistics stats = pathCache.getStatistics();
        return (double) stats.getCacheHits() /
                ((double) (stats.getCacheHits() + stats.getCacheMisses()));
    }

    /**
     * Method to add resource path entry to the database.
     *
     * @param path         the path to add.
     * @param parentPathId the parent path's id.
     *
     * @return the path's id.
     * @throws SQLException if an error occurs while adding the entry.
     */
    public int addEntry(String path, int parentPathId) throws SQLException {
        ResultSet results = null;
        PreparedStatement ps = null;
        Connection conn;
        if (CurrentSession.getUserRegistry() != null &&
                CurrentSession.getUserRegistry().getRegistryContext() != null) {
            conn = CurrentSession.getUserRegistry().getRegistryContext().getDataSource()
                    .getConnection();
        } else {
            // TODO: This code block doesn't seem to get hit. Remove if unused.
            conn = RegistryContext.getBaseInstance().getDataSource().getConnection();
        }
        if (conn.getTransactionIsolation() != Connection.TRANSACTION_READ_COMMITTED) {
            conn.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
        }
        conn.setAutoCommit(false);
        boolean success = false;
        int pathId = 0;

        try {
            String sql =
                    "INSERT INTO REG_PATH(REG_PATH_VALUE, REG_PATH_PARENT_ID, REG_TENANT_ID) " +
                            "VALUES (?, ?, ?)";
            String dbProductName = conn.getMetaData().getDatabaseProductName();
            ps = conn.prepareStatement(sql, new String[]{
                    DBUtils.getConvertedAutoGeneratedColumnName(dbProductName, "REG_PATH_ID")});
            ps.setString(1, path);
            ps.setInt(2, parentPathId);
            ps.setInt(3, CurrentSession.getTenantId());
            ps.executeUpdate();

            results = ps.getGeneratedKeys();
            if (results.next()) {
                pathId = results.getInt(1);
                if (pathId > 0) {
                    success = true;
                    return pathId;
                }
            }
        } catch (SQLException e) {
            // we have to be expecting an exception with the duplicate value for the path value
            // which can be further checked from here..
            String msg = "Failed to insert resource to " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw e;
        } finally {
            if (success) {
                try {
                    conn.commit();
                    CacheEntry e = new CacheEntry(pathId);
                    CacheKey key = new CacheKey(path, CurrentSession.getTenantId());
                    Element cacheEle = new Element(key, e);
                    pathCache.put(cacheEle);

                } catch (SQLException e) {
                    String msg = "Failed to commit transaction. Inserting " + path + ". " +
                            e.getMessage();
                    log.error(msg, e);
                } finally {
                    try {
                        if (results != null) {
                            results.close();
                        }
                        if (ps != null) {
                            ps.close();
                        }
                        if (conn != null) {
                            conn.close();
                        }
                    } catch (SQLException e) {
                        String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR +
                                e.getMessage();
                        log.error(msg, e);
                    }
                }
            } else {
                try {
                    conn.rollback();

                } catch (SQLException e) {
                    String msg = "Failed to rollback transaction. Inserting " + path + ". " +
                            e.getMessage();
                    log.error(msg, e);
                } finally {
                    try {
                        if (results != null) {
                            results.close();
                        }
                        if (ps != null) {
                            ps.close();
                        }
                        if (conn != null) {
                            conn.close();
                        }
                    } catch (SQLException e) {
                        String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR +
                                e.getMessage();
                        log.error(msg, e);
                    }
                }
            }
        }
        return -1;
    }

    /**
     * Method to get the path id of a given path.
     *
     * @param conn the database connection to use.
     * @param path the path.
     *
     * @return the path id corresponding to the given path.
     * @throws SQLException if an error occurs while obtaining the path id.
     */
    public int getPathID(Connection conn, String path) throws SQLException {
        CacheKey key = new CacheKey(path, CurrentSession.getTenantId());
        Element cacheEle = pathCache.get(key);
        if (cacheEle != null) {
            CacheEntry result = (CacheEntry) cacheEle.getObjectValue();
            if (result != null) {
                return result.pathId;
            }
        } else {
            ResultSet results = null;
            PreparedStatement ps = null;
            try {
                String sql =
                        "SELECT REG_PATH_ID FROM REG_PATH WHERE REG_PATH_VALUE=? " +
                                "AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setString(1, path);
                ps.setInt(2, CurrentSession.getTenantId());
                results = ps.executeQuery();

                int pathId;
                if (results.next()) {
                    pathId = results.getInt(DatabaseConstants.PATH_ID_FIELD);

                    if (pathId > 0) {
                        CacheEntry e = new CacheEntry(pathId);
                        cacheEle = new Element(key, e);
                        pathCache.put(cacheEle);
                        return pathId;
                    }
                }

            } catch (SQLException e) {
                String msg = "Failed to retrieving resource from " + path + ". " + e.getMessage();
                log.error(msg, e);
                throw e;
            } finally {
                try {
                    if (results != null) {
                        results.close();
                    }
                    if (ps != null) {
                        ps.close();
                    }
                } catch (SQLException e) {
                    String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR +
                            e.getMessage();
                    log.error(msg, e);
                }
            }
        }
        return -1;
    }

    // the container class for the cache entry
    private static class CacheEntry {

        CacheEntry(int pathId) {
            this.pathId = pathId;
        }

        int pathId;
    }

    // the container class for the cache key
    private static class CacheKey {

        CacheKey(String path, int tenantId) {
            this.tenantId = tenantId;
            this.path = path;
        }

        int tenantId;
        String path;

        public boolean equals(Object key) {
            if (!(key instanceof CacheKey)) {
                return false;
            }
            CacheKey cacheKey = (CacheKey) key;

            return cacheKey.tenantId == tenantId &&
                    (cacheKey.path == null && path == null ||
                            cacheKey.path != null && cacheKey.path.equals(path));
        }

        public int hashCode() {
            if (path == null) {
                return new Integer(tenantId).hashCode();
            }
            return path.hashCode() + new Integer(tenantId).hashCode();
        }
    }
}