/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.CollectionImpl;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.ResourceIDImpl;
import org.wso2.carbon.registry.core.ResourceImpl;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.RatingDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.utils.DBUtils;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;

/**
 * Data Access Object for Ratings
 */
public class RatingsDAO {

    private static final Log log = LogFactory.getLog(RatingsDAO.class);
    private ResourceDAO resourceDAO = new ResourceDAO();

    /**
     * Method to persist a rating.
     *
     * @param resourceImpl the resource
     * @param userID       the id of the user who added the rating.
     * @param rating       the rating to be persisted.
     *
     * @throws RegistryException if some error occurs while adding a rating
     */
    public void addRating(ResourceImpl resourceImpl, String userID, int rating)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null, ps2 = null;
        ResultSet result = null;
        try {

            String sql =
                    "INSERT INTO REG_RATING (REG_RATING, REG_USER_ID, REG_RATED_TIME, " +
                            "REG_TENANT_ID) VALUES (?,?,?,?)";
            String dbProductName = conn.getMetaData().getDatabaseProductName();
            ps = conn.prepareStatement(sql, new String[]{DBUtils
                    .getConvertedAutoGeneratedColumnName(dbProductName,
                    "REG_ID")});
            ps.setInt(1, rating);
            ps.setString(2, userID);
            long now = System.currentTimeMillis();
            ps.setDate(3, new Date(now));
            ps.setInt(4, CurrentSession.getTenantId());

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if (result.next()) {
                int rateID = result.getInt(1);

                String sql2 = "INSERT INTO REG_RESOURCE_RATING (REG_RATING_ID, REG_PATH_ID, " +
                        "REG_RESOURCE_NAME, REG_TENANT_ID) " +
                        "VALUES(?,?,?,?)";
                ps2 = conn.prepareStatement(sql2);

                ps2.setInt(1, rateID);
                ps2.setInt(2, resourceImpl.getPathID());
                ps2.setString(3, resourceImpl.getName());
                ps2.setInt(4, CurrentSession.getTenantId());

                ps2.executeUpdate();

            }

        } catch (SQLException e) {

            String msg = "Failed to rate resource " + resourceImpl.getPath() +
                    " with rating " + rating + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to copy ratings.
     *
     * @param fromResource the source resource.
     * @param toResource   the target resource.
     *
     * @throws RegistryException if some error occurs while copying ratings
     */
    public void copyRatings(ResourceImpl fromResource, ResourceImpl toResource)
            throws RegistryException {
        if (fromResource == null || toResource == null || fromResource.getPath() == null
                || fromResource.getPath().equals(toResource.getPath())) {
            // no special copying needed
            return;
        }
        RatingDO[] ratingDOs = getResourceRatingDO(fromResource);
        addRatings(toResource, ratingDOs);
    }

    /**
     * Method to persist ratings.
     *
     * @param resource  the resource
     * @param ratingDOs the ratings to be persisted.
     *
     * @throws RegistryException if some error occurs while adding ratings
     */
    public void addRatings(ResourceImpl resource, RatingDO[] ratingDOs)
            throws RegistryException {
        for (RatingDO ratingDO : ratingDOs) {
            addRating(resource, ratingDO.getRatedUserName(), ratingDO.getRating());
        }
    }

    /**
     * Method to update a rating added to a resource.
     *
     * @param resourceImpl the resource.
     * @param rateID       the rate id.
     * @param rating       the rating.
     *
     * @throws RegistryException if some error occurs while updating the rating.
     */
    public void updateRating(ResourceImpl resourceImpl, int rateID, int rating)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        try {

            String sql =
                    "UPDATE REG_RATING SET REG_RATING=?, REG_RATED_TIME=? WHERE REG_ID=? " +
                            "AND REG_TENANT_ID=?";

            long now = System.currentTimeMillis();

            ps = conn.prepareStatement(sql);
            ps.setInt(1, rating);
            ps.setDate(2, new Date(now));
            ps.setInt(3, rateID);
            ps.setInt(4, CurrentSession.getTenantId());

            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to update the rating id " + rateID + " of resource " +
                    resourceImpl.getPath() +
                    " to value " + rating + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to get a id of a rating added to a given resource.
     *
     * @param resourceImpl the resource.
     * @param userID       the id of the user who added the rating.
     *
     * @return the rate id.
     * @throws RegistryException if an error occurs while getting the rate id.
     */
    public int getRateID(ResourceImpl resourceImpl, String userID) throws RegistryException {

        RatingDO ratingDO = getRatingDO(resourceImpl, userID);
        return ratingDO.getRatingID();
    }

    /**
     * Method to get the average rating added to a given resource.
     *
     * @param resourceImpl the resource.
     *
     * @return the average rating.
     * @throws RegistryException if an error occurs while getting the average rating.
     */
    public float getAverageRating(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        int sumRating;
        int countRating;

        PreparedStatement ps = null;
        ResultSet result1 = null;
        try {
            if (resourceImpl instanceof CollectionImpl) {
                String sql = "SELECT SUM(R.REG_RATING) FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME IS NULL AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
            } else {
                String sql = "SELECT SUM(R.REG_RATING) FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME = ? AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.setInt(4, CurrentSession.getTenantId());
            }

            result1 = ps.executeQuery();

            sumRating = 0;
            if (result1.next()) {
                sumRating = result1.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to get sum of all ratings on resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result1 != null) {
                    result1.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        ps = null;
        ResultSet result2 = null;
        try {

            if (resourceImpl instanceof CollectionImpl) {
                String sql =
                        "SELECT COUNT(R.REG_RATING) FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                                "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME IS NULL AND " +
                                "RR.REG_RATING_ID=R.REG_ID AND R.REG_TENANT_ID=? " +
                                "AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
            } else {
                String sql =
                        "SELECT COUNT(R.REG_RATING) FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                                "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME = ? AND " +
                                "RR.REG_RATING_ID=R.REG_ID AND R.REG_TENANT_ID=? " +
                                "AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.setInt(4, CurrentSession.getTenantId());
            }

            result2 = ps.executeQuery();

            countRating = 0;
            if (result2.next()) {
                countRating = result2.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to get ratings count on resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result2 != null) {
                    result2.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        float averageRating = 0;
        if (countRating > 0) {
            averageRating = (float) sumRating / countRating;
        }

        return averageRating;
    }

    /**
     * Method to get a rating added by the given user to the given resource.
     *
     * @param resourceImpl the resource.
     * @param userID       the id of the user who added the rating.
     *
     * @return the rating data object.
     * @throws RegistryException if an error occurs while getting the rating.
     */
    public RatingDO getRatingDO(ResourceImpl resourceImpl, String userID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        ResultSet result = null;
        RatingDO ratingDO = new RatingDO();
        try {
            if (resourceImpl instanceof CollectionImpl) {
                String sql = "SELECT R.REG_ID, R.REG_RATING, R.REG_RATED_TIME " +
                        "FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME IS NULL AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND R.REG_USER_ID=? AND " +
                        "R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, userID);
                ps.setInt(3, CurrentSession.getTenantId());
                ps.setInt(4, CurrentSession.getTenantId());
            } else {
                String sql = "SELECT R.REG_ID, R.REG_RATING, R.REG_RATED_TIME " +
                        "FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME = ? AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND R.REG_USER_ID=? AND " +
                        "R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
                ps.setString(3, userID);
                ps.setInt(4, CurrentSession.getTenantId());
                ps.setInt(5, CurrentSession.getTenantId());
            }

            result = ps.executeQuery();
            if (result.next()) {
                ratingDO.setRating(result.getInt(DatabaseConstants.RATING_FIELD));
                ratingDO.setRatedTime(new java.util.Date(
                        result.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime()));
                ratingDO.setRatedUserName(userID);
                ratingDO.setRatingID(result.getInt(DatabaseConstants.ID_FIELD));
            }
        } catch (SQLException e) {

            String msg = "Failed to get rating on resource " + resourceImpl.getPath() +
                    " done by user " + userID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
        return ratingDO;
    }

    /**
     * Method to get a rating added to a given resource.
     *
     * @param resourceImpl the resource.
     * @param userID       the id of the user who added the rating.
     *
     * @return the rating.
     * @throws RegistryException if an error occurs while getting the rating.
     */
    public int getRating(ResourceImpl resourceImpl, String userID) throws RegistryException {

        RatingDO ratingDO = getRatingDO(resourceImpl, userID);
        return ratingDO.getRating();
    }

    /**
     * Method to get a rating corresponding to the given id.
     *
     * @param ratingID the id of the rating.
     *
     * @return the rating data object.
     * @throws RegistryException if an error occurs while getting the resource path.
     * @throws SQLException      if an error occurs while getting the rating.
     */
    public RatingDO getRating(long ratingID) throws SQLException, RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT RR.REG_PATH_ID, RR.REG_RESOURCE_NAME, " +
                "R.REG_USER_ID, R.REG_RATING, R.REG_RATED_TIME " +
                "FROM REG_RATING R, REG_RESOURCE_RATING RR WHERE R.REG_ID =? AND " +
                "R.REG_ID = RR.REG_RATING_ID AND R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, ratingID);
        s.setInt(2, CurrentSession.getTenantId());
        s.setInt(3, CurrentSession.getTenantId());

        ResultSet results = s.executeQuery();
        if (results.next()) {

            java.util.Date ratedTime = new java.util.Date(
                    results.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime());

            RatingDO ratingDAO = new RatingDO();
            ratingDAO.setRatedUserName(results.getString(DatabaseConstants.USER_ID_FIELD));
            ratingDAO.setRatedTime(ratedTime);
            ratingDAO.setRating(results.getInt(DatabaseConstants.RATING_FIELD));
            // TODO:
            //ratingDAO.setResourceID(results.getString(DatabaseConstants.AID_FIELD));

            String resourcePath;

            int pathId = results.getInt(DatabaseConstants.PATH_ID_FIELD);
            String resourceName = results.getString(DatabaseConstants.RESOURCE_NAME_FIELD);
            resourcePath = resourceDAO.getPath(pathId, resourceName, true);
            if (resourcePath != null) {
                ratingDAO.setResourcePath(resourcePath);
            }
            return ratingDAO;
        }

        return null;
    }

    /**
     * Method to remove all ratings added to a resource.
     *
     * @param resourceImpl the resource.
     *
     * @throws RegistryException if some error occurs while removing ratings.
     */
    public void removeRatings(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps1 = null, ps2 = null;
        RatingDO[] ratingDOs = getResourceRatingDO(resourceImpl);
        try {
            String sql1 =
                    "DELETE FROM REG_RESOURCE_RATING WHERE REG_RATING_ID = ? AND REG_TENANT_ID=?";
            ps1 = conn.prepareStatement(sql1);

            String sql2 = "DELETE FROM REG_RATING WHERE REG_ID = ? AND REG_TENANT_ID=?";
            ps2 = conn.prepareStatement(sql2);

            for (RatingDO ratingDO : ratingDOs) {
                ps1.setLong(1, ratingDO.getRatingID());
                ps1.setInt(2, CurrentSession.getTenantId());

                ps2.setLong(1, ratingDO.getRatingID());
                ps2.setInt(2, CurrentSession.getTenantId());
                ps1.addBatch();
                ps2.addBatch();
            }

            if (ratingDOs.length > 0) {
                try {
                    ps1.executeBatch();
                    ps2.executeBatch();
                } catch (SQLException e) {
                    ps1.clearBatch();
                    ps2.clearBatch();
                    // the exception will be handled in the next catch block
                    throw e;
                }
            }

        } catch (SQLException e) {

            String msg = "Failed to remove ratings on resource " + resourceImpl.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps1 != null) {
                    ps1.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to get ratings added by all users to the given resource.
     *
     * @param resourceImpl the resource.
     *
     * @return array of rating data objects.
     * @throws RegistryException if an error occurs while getting the rating.
     */
    public RatingDO[] getResourceRatingDO(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        PreparedStatement ps = null;
        ResultSet result = null;
        ArrayList<RatingDO> ratingDOs = new ArrayList<RatingDO>();
        try {
            if (resourceImpl instanceof CollectionImpl) {
                String sql = "SELECT R.REG_ID, R.REG_RATING, R.REG_USER_ID, R.REG_RATED_TIME " +
                        "FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME IS NULL AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND " +
                        "R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
            } else {
                String sql = "SELECT R.REG_ID, R.REG_RATING, R.REG_USER_ID, R.REG_RATED_TIME " +
                        "FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME = ? AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.setInt(4, CurrentSession.getTenantId());
            }

            result = ps.executeQuery();
            while (result.next()) {
                RatingDO ratingDO = new RatingDO();
                ratingDO.setRating(result.getInt(DatabaseConstants.RATING_FIELD));
                ratingDO.setRatedTime(new java.util.Date(
                        result.getTimestamp(DatabaseConstants.RATED_TIME_FIELD).getTime()));
                ratingDO.setRatedUserName(result.getString(DatabaseConstants.USER_ID_FIELD));
                ratingDO.setRatingID(result.getInt(DatabaseConstants.ID_FIELD));

                ratingDOs.add(ratingDO);
            }
        } catch (SQLException e) {

            String msg = "Failed to get rating on resource " + resourceImpl.getPath() +
                    ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
        return ratingDOs.toArray(new RatingDO[ratingDOs.size()]);
    }

    /**
     * Method to get users who rated the given resource.
     *
     * @param resourceImpl the resource.
     *
     * @return array of user names.
     * @throws RegistryException if an error occurs while getting the rating.
     */
    public String[] getRatedUserNames(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List<String> userNameList = new ArrayList<String>();

        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            if (resourceImpl instanceof CollectionImpl) {
                String sql = "SELECT R.REG_USER_ID FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME IS NULL AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
            } else {
                String sql = "SELECT R.REG_USER_ID FROM REG_RATING R, REG_RESOURCE_RATING RR " +
                        "WHERE RR.REG_PATH_ID=? AND RR.REG_RESOURCE_NAME = ? AND " +
                        "RR.REG_RATING_ID=R.REG_ID AND R.REG_TENANT_ID=? AND RR.REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceImpl.getPathID());
                ps.setString(2, resourceImpl.getName());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.setInt(4, CurrentSession.getTenantId());
            }

            results = ps.executeQuery();
            while (results.next()) {
                String ratedUserName = results.getString(DatabaseConstants.USER_ID_FIELD);
                userNameList.add(ratedUserName);
            }

        } catch (SQLException e) {

            String msg = "Failed to users who have rated the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        String[] userNames = new String[userNameList.size()];
        for (int i = 0; i < userNames.length; i++) {
            userNames[i] = userNameList.get(i);
        }

        return userNames;
    }

    /**
     * Gets the resource with sufficient data to differentiate it from another resource. This would
     * populate a {@link ResourceImpl} with the <b>path</b>, <b>name</b> and <b>path identifier</b>
     * of a resource.
     *
     * @param path the path of the resource.
     *
     * @return the resource with minimum data.
     * @throws RegistryException if an error occurs while retrieving resource data.
     */
    public ResourceImpl getResourceWithMinimumData(String path) throws RegistryException {
        return RegistryUtils.getResourceWithMinimumData(path, resourceDAO, false);
    }

    /**
     * Method to move ratings.
     *
     * @param source the source resource.
     * @param target the target resource.
     *
     * @throws RegistryException if some error occurs while moving ratings
     */
    public void moveRatings(ResourceIDImpl source, ResourceIDImpl target) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (source.isCollection()) {
                String sql = "UPDATE REG_RESOURCE_RATING SET REG_PATH_ID=? WHERE " +
                        "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setInt(2, source.getPathID());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.executeUpdate();
            } else {
                String sql =
                        "UPDATE REG_RESOURCE_RATING SET REG_PATH_ID=?, REG_RESOURCE_NAME=? WHERE " +
                                "REG_PATH_ID=? AND REG_RESOURCE_NAME=? AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setString(2, target.getName());
                ps.setInt(3, source.getPathID());
                ps.setString(4, source.getName());
                ps.setInt(5, CurrentSession.getTenantId());
                ps.executeUpdate();
            }
        } catch (SQLException e) {
            String msg = "Failed to move ratings from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to move rating paths. This function will move the paths from one path id to another
     * regardless of the resource name.
     *
     * @param source the source resource.
     * @param target the target resource.
     *
     * @throws RegistryException if some error occurs while moving rating paths
     */
    public void moveRatingPaths(ResourceIDImpl source, ResourceIDImpl target)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql =
                    "UPDATE REG_RESOURCE_RATING SET REG_PATH_ID=? WHERE REG_PATH_ID=? " +
                            "AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, target.getPathID());
            ps.setInt(2, source.getPathID());
            ps.setInt(3, CurrentSession.getTenantId());
            ps.executeUpdate();
        } catch (SQLException e) {
            String msg = "Failed to move rating paths from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }
}
