/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.StaticConfiguration;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dataobjects.ResourceDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.utils.DBUtils;

import java.io.InputStream;
import java.sql.*;
import java.util.*;
import java.util.Date;

/**
 * The data access object for resources.
 */
public class ResourceDAO {

    private static final Log log = LogFactory.getLog(ResourceDAO.class);

    /**
     * Returns the resource ID (RID) of the resource referred by the given path. NOTE: This doesn't
     * guarantee that the path exist in the resource. It only guarantee if it is a collection: the
     * path exist in the path table. if it is a resource: the parent path exist in the path table.
     * In order to make sure the existence of the resource please use the resourceExists function
     * Note that if the same path is used for a collection and resource, this returns the resourceID
     * of the collection.
     *
     * @param path Pure path of the resource
     *
     * @return Resource id of resource exists in path cache. null if the resource does not exists.
     * @throws RegistryException throws the retrieval of resource id failed
     */
    public ResourceIDImpl getResourceID(String path) throws RegistryException {
        Connection conn = Transaction.getConnection();

        try {
            // step1: need to check whether it is a resource or collection while retrieving path id
            PathCache pathCache = PathCache.getPathCache();
            int pathID = pathCache.getPathID(conn, path);
            boolean isCollection = true;
            if (pathID == -1) {
                isCollection = false;
            }

            // step2: get the resource name + path id for non-collection
            // (for collection this is already done from the step1
            String resourceName = null;
            if (!isCollection) {
                // we have to re-get the path id for the parent path..
                String parentPath = RegistryUtils.getParentPath(path);
                resourceName = RegistryUtils.getResourceName(path);

                pathID = pathCache.getPathID(conn, parentPath);
            }

            if (pathID != -1) {
                ResourceIDImpl resourceID = new ResourceIDImpl();
                resourceID.setCollection(isCollection);
                resourceID.setName(resourceName);
                resourceID.setPathID(pathID);
                resourceID.setPath(path);
                return resourceID;
            }


        } catch (SQLException e) {
            String msg = "Failed to get ID of the resource at path " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return null;
    }


    /**
     * Returns the resource ID (RID) of the resource referred by the given path. We use this
     * overloaded function when we know our resource is a collection or non-collection NOTE: This
     * doesn't guarantee that the path exist in the resource. It only guarantee if it is a
     * collection: the path exist in the path table. if it is a resource: the parent path exist in
     * the path table. In order to make sure the existence of the resource please use the
     * resourceExists function
     *
     * @param path         Pure path of the resource
     * @param isCollection true if it is a collection
     *
     * @return Resource id of resource exists. null if the resource does not exists.
     * @throws RegistryException throws the retrieval of resource id failed
     */
    public ResourceIDImpl getResourceID(String path, boolean isCollection)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        try {
            // step1: need to check whether it is a resource or collection while retrieving path id
            PathCache pathCache = PathCache.getPathCache();

            int pathID;
            String resourceName = null;
            if (isCollection) {
                pathID = pathCache.getPathID(conn, path);
            } else {
                // we have to re-get the path id for the parent path..
                String parentPath;
                if (path.equals(RegistryConstants.ROOT_PATH)) {
                    parentPath = null;
                } else {
                    parentPath = RegistryUtils.getParentPath(path);
                }
                resourceName = RegistryUtils.getResourceName(path);

                pathID = pathCache.getPathID(conn, parentPath);
            }

            if (pathID != -1) {
                ResourceIDImpl resourceID = new ResourceIDImpl();
                resourceID.setCollection(isCollection);
                resourceID.setName(resourceName);
                resourceID.setPathID(pathID);
                resourceID.setPath(path);
                return resourceID;
            }


        } catch (SQLException e) {
            String msg = "Failed to get ID of the resource at path " + path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return null;
    }

    /**
     * Method to check the whether a resource in the provided resource id exist or not.
     *
     * @param resourceID the resource id which is checked for existence of a resource.
     *
     * @return true if the resource exists, false otherwise.
     * @throws RegistryException throws if checking existence failed.
     */
    public boolean resourceExists(ResourceIDImpl resourceID) throws RegistryException {
        return getVersion(resourceID) != -1;
    }

    /**
     * Method to check the resource existence for a given path.
     *
     * @param path the path to check the resource existence
     *
     * @return true, if the resource exists, false otherwise
     * @throws RegistryException throws if checking existence failed.
     */
    public boolean resourceExists(String path) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path);
        if (resourceID == null) {
            // resource definitely doesn't exists
            return false;
        }
        if (resourceExists(resourceID)) {
            // resource definitely exist
            return true;
        }
        // if resourceID is a collection we have to check resourceID with non-collection as well..
        return resourceID.isCollection() && resourceExists(path, false);
    }

    /**
     * Method to check the resource existence for a given path provided the resource is collection
     * or not.
     *
     * @param path         the path to check the resource existence.
     * @param isCollection provide whether the resource in the path is collection or not.
     *
     * @return true, if the resource exists, false otherwise.
     * @throws RegistryException throws if checking existence failed.
     */
    public boolean resourceExists(String path, boolean isCollection) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path, isCollection);
        return resourceID != null && resourceExists(resourceID);
    }

    /**
     * Method to return the version of a resource from resourceID
     *
     * @param resourceID the id of the resource to get the version of.
     *
     * @return the version of the resource for the given resource id.
     * @throws RegistryException throws if the version retrieval failed.
     */
    public long getVersion(ResourceIDImpl resourceID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        long version = -1;
        try {
            if (resourceID.isCollection()) {
                String sql = "SELECT REG_VERSION FROM REG_RESOURCE WHERE " +
                        "REG_PATH_ID=? AND REG_NAME IS NULL AND REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());

                result = ps.executeQuery();
                if (result.next()) {
                    version = result.getLong(DatabaseConstants.VERSION_FIELD);
                }
            } else {
                String sql = "SELECT REG_VERSION FROM REG_RESOURCE WHERE " +
                        "REG_PATH_ID=? AND REG_NAME=? AND REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
                ps.setInt(3, CurrentSession.getTenantId());

                result = ps.executeQuery();
                if (result.next()) {
                    version = result.getLong(DatabaseConstants.VERSION_FIELD);
                }
            }
        } catch (SQLException e) {
            String msg = "Failed to check the existence of the resource " +
                    resourceID.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        return version;
    }

    /**
     * Returns the resource in the given path filled with meta-data and access to the content. If a
     * resource does not exist in the given path, null is returned.
     *
     * @param path Path of the resource.
     *
     * @return ResourceImpl filled with resource meta-data and access to the resource content.
     * @throws RegistryException throws if the resource retrieval failed.
     */
    public ResourceImpl get(String path) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path);
        if (resourceID == null) {
            return null;
        }
        return get(resourceID);
    }

    /**
     * Returns the resource in the given path filled with meta-data and access to the content. If a
     * resource does not exist in the given resourceID, null is returned.
     *
     * @param resourceID the resource id
     *
     * @return resource object.
     * @throws RegistryException throws if resource retrieval failed.
     */
    public ResourceImpl get(ResourceIDImpl resourceID) throws RegistryException {
        ResourceImpl resourceImpl = getResource(resourceID);
        if (resourceImpl == null) {
            // it is possible the resource doesn't exist
            return null;
        }
        fillResourceProperties(resourceImpl);
        return resourceImpl;
    }

    /**
     * Return a collection with children only at the range of the intersect of the given range and
     * resource existence range provided the resource path. Use for resource browser pagination.
     *
     * @param path    path of the collection
     * @param start   start value of the range of children.
     * @param pageLen the length of the children to retrieve
     *
     * @return an instance of collection with child
     * @throws RegistryException throws if resource retrieval failed.
     */
    public CollectionImpl get(String path, int start, int pageLen) throws RegistryException {
        ResourceIDImpl resourceID = getResourceID(path);
        if (resourceID == null) {
            return null;
        }
        return get(resourceID, start, pageLen);
    }

    /**
     * Return a collection with children only at the range of the intersect of the given range and
     * resource existence range provided the resource id. Use for resource browser pagination.
     *
     * @param resourceID resourceID of the collection
     * @param start      start value of the range of children.
     * @param pageLen    the length of the children to retrieve
     *
     * @return an instance of collection with child
     * @throws RegistryException throws if resource retrieval failed.
     */
    public CollectionImpl get(ResourceIDImpl resourceID, int start, int pageLen)
            throws RegistryException {
        CollectionImpl collection = (CollectionImpl) getResource(resourceID);
        if (collection == null) {
            // it is possible the resource doesn't exist
            return null;
        }
        fillChildren(collection, start, pageLen);
        fillResourceProperties(collection);
        return collection;
    }

    /**
     * Fill the content (for non-collection) and the properties for a resource that already filled
     * meta data
     *
     * @param resourceImpl the resource instance to be filled with contents (if non-collection) and
     *                     fill properties.
     *
     * @throws RegistryException throws if resource filling failed.
     */
    public void fillResource(ResourceImpl resourceImpl) throws RegistryException {
        if (resourceImpl != null && !(resourceImpl instanceof CollectionImpl) &&
                resourceImpl.getDbBasedContentID() > 0) {
            fillResourceContentWithNoUpdate(resourceImpl);
        } else if (resourceImpl instanceof CollectionImpl &&
                CurrentSession.getTenantId() > 0) {
            Connection conn = Transaction.getConnection();
            fillChildren((CollectionImpl) resourceImpl, 0, -1, conn);
        }
        fillResourcePropertiesWithNoUpdate(resourceImpl);
    }

    /**
     * Fill the children and the properties for a resource that already filled with meta data.
     * Children are filled only at the at of the intersect of the given range and resource existence
     * range.
     *
     * @param collection collection to fill the children and properties.
     * @param start      start value of the range of children.
     * @param pageLen    the length of the children to retrieve
     *
     * @throws RegistryException throws if resource filling failed.
     */
    public void fillResource(CollectionImpl collection, int start, int pageLen)
            throws RegistryException {
        fillChildren(collection, start, pageLen);
        fillResourcePropertiesWithNoUpdate(collection);
    }

    /**
     * Fill the properties for a resource without making the properties modified flag.
     *
     * @param resourceImpl the resource object.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void fillResourcePropertiesWithNoUpdate(ResourceImpl resourceImpl)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            if (StaticConfiguration.isVersioningProperties()) {
                String propSQL =
                        "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, " +
                                "REG_RESOURCE_PROPERTY RP " +
                                "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_VERSION=? AND " +
                                "P.REG_TENANT_ID=? AND RP.REG_TENANT_ID=?";
                ps = conn.prepareStatement(propSQL);
                ps.setLong(1, resourceImpl.getVersionNumber());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
                results = ps.executeQuery();
            } else if (resourceImpl instanceof CollectionImpl) {
                String propSQL =
                        "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, " +
                                "REG_RESOURCE_PROPERTY RP WHERE P.REG_ID=RP.REG_PROPERTY_ID " +
                                "AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME IS NULL " +
                                "AND P.REG_TENANT_ID=? AND RP.REG_TENANT_ID=?";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
                results = ps.executeQuery();
            } else {
                String propSQL =
                        "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, " +
                                "REG_RESOURCE_PROPERTY RP WHERE P.REG_ID=RP.REG_PROPERTY_ID " +
                                "AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME=? " +
                                "AND P.REG_TENANT_ID=? AND RP.REG_TENANT_ID=?";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.setInt(4, CurrentSession.getTenantId());
                results = ps.executeQuery();
            }

            while (results.next()) {
                String name = results.getString(DatabaseConstants.NAME_FIELD);
                String value = results.getString(DatabaseConstants.VALUE_FIELD);
                resourceImpl.addPropertyWithNoUpdate(name, value);
            }
        } catch (SQLException e) {
            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }


    /**
     * Fill the properties for a resource, this will change the properties modified flag.
     *
     * @param resourceImpl the resource object.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void fillResourceProperties(ResourceImpl resourceImpl) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            if (StaticConfiguration.isVersioningProperties()) {
                String propSQL =
                        "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, " +
                                "REG_RESOURCE_PROPERTY RP " +
                                "WHERE P.REG_ID=RP.REG_PROPERTY_ID AND RP.REG_VERSION=? AND " +
                                "P.REG_TENANT_ID=? AND RP.REG_TENANT_ID=?";
                ps = conn.prepareStatement(propSQL);
                ps.setLong(1, resourceImpl.getVersionNumber());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
                results = ps.executeQuery();
            } else if (resourceImpl instanceof CollectionImpl) {
                String propSQL =
                        "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, " +
                                "REG_RESOURCE_PROPERTY RP WHERE P.REG_ID=RP.REG_PROPERTY_ID " +
                                "AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME IS NULL " +
                                "AND P.REG_TENANT_ID=? AND RP.REG_TENANT_ID=?";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());
                ps.setInt(3, CurrentSession.getTenantId());
                results = ps.executeQuery();
            } else {
                String propSQL =
                        "SELECT REG_NAME, REG_VALUE FROM REG_PROPERTY P, " +
                                "REG_RESOURCE_PROPERTY RP WHERE P.REG_ID=RP.REG_PROPERTY_ID " +
                                "AND RP.REG_PATH_ID=? AND RP.REG_RESOURCE_NAME=? " +
                                "AND P.REG_TENANT_ID=? AND RP.REG_TENANT_ID=?";
                ps = conn.prepareStatement(propSQL);
                ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
                ps.setLong(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.setInt(4, CurrentSession.getTenantId());
                results = ps.executeQuery();
            }

            while (results.next()) {
                String name = results.getString(DatabaseConstants.NAME_FIELD);
                String value = results.getString(DatabaseConstants.VALUE_FIELD);
                resourceImpl.addProperty(name, value);
            }
        } catch (SQLException e) {
            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
        resourceImpl.setPropertiesModified(false);
    }

    /**
     * Add the root collection. Only called at the very first time resource loading process.
     *
     * @param resourceImpl the resource instance
     *
     * @throws RegistryException throws if the operation failed
     */
    public void addRoot(ResourceImpl resourceImpl)
            throws RegistryException {
        String path = RegistryConstants.ROOT_PATH;
        createAndApplyResourceID(path, null, resourceImpl);

        addResource(resourceImpl, false);
        addProperties(resourceImpl);
    }

    /**
     * Add the resource to a pat when resource instance and the parent resource id is given.
     *
     * @param path         path of the resource
     * @param parentID     parent resourceID
     * @param resourceImpl the instance of the resource to be added.
     *
     * @throws RegistryException throws if the operation failed
     */
    public void add(String path, ResourceIDImpl parentID, ResourceImpl resourceImpl)
            throws RegistryException {
        // creating the resourceID
        createAndApplyResourceID(path, parentID, resourceImpl);

        addResourceWithoutContentId(resourceImpl, false);

        if (!(resourceImpl instanceof org.wso2.carbon.registry.core.Collection)) {
            addContent(resourceImpl);
            if (resourceImpl.getDbBasedContentID() > 0) {
                updateContentId(resourceImpl);
            }
        }

        addProperties(resourceImpl);
    }

    /**
     * The method to create a resource id and assign to resource instance
     *
     * @param path         path of the resource
     * @param parentID     parent path id
     * @param resourceImpl the resource instance to be assigned the resource id.
     *
     * @throws RegistryException throws if operation failed.
     */
    public void createAndApplyResourceID(String path, ResourceIDImpl parentID,
                                         ResourceImpl resourceImpl)
            throws RegistryException {
        ResourceIDImpl resourceID =
                createResourceID(path, parentID, resourceImpl instanceof CollectionImpl);
        resourceImpl.setName(resourceID.getName());
        resourceImpl.setPath(resourceID.getPath());
        resourceImpl.setPathID(resourceID.getPathID());
    }

    /**
     * Create a resource ID for a path given the parent resource id and whether it is a collection
     * or not.
     *
     * @param path         the path of the resource
     * @param parentID     parent resource id
     * @param isCollection whether the resource is a collection or not
     *
     * @return the newly created resource id.
     * @throws RegistryException throws if operation failed.
     */
    public ResourceIDImpl createResourceID(String path, ResourceIDImpl parentID,
                                           boolean isCollection)
            throws RegistryException {
        ResourceIDImpl resourceID = new ResourceIDImpl();
        try {
            int parentPathID = -1;
            if (parentID != null) {
                parentPathID = parentID.getPathID();
            }
            if (isCollection) {
                resourceID.setName(null);
                resourceID.setPath(path);

                Connection conn = Transaction.getConnection();
                int pathID = PathCache.getPathCache().getPathID(conn, path);
                if (pathID == -1) {
                    pathID = PathCache.getPathCache().addEntry(path, parentPathID);
                }
                resourceID.setPathID(pathID);
            } else {
                String resourceName = RegistryUtils.getResourceName(path);
                resourceID.setName(resourceName);
                resourceID.setPath(path);
                resourceID.setPathID(parentPathID);
            }
        } catch (SQLException e) {
            String msg = "Failed to create the resource id for the resource " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        return resourceID;
    }

    /**
     * delete the content for a given content id.
     *
     * @param contentID content id.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void deleteContentStream(int contentID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String resourceContentSQL = "DELETE FROM REG_CONTENT WHERE " +
                    "REG_CONTENT_ID = ? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(resourceContentSQL);
            ps.setLong(1, contentID);
            ps.setInt(2, CurrentSession.getTenantId());
            ps.executeUpdate();

        }
        catch (SQLException ex) {
            String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
            log.error(msg, ex);
            throw new RegistryException(msg, ex);
        }
        finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Get the content input stream for a given content id.
     *
     * @param contentID the content id as an argument.
     *
     * @return the content input stream.
     * @throws RegistryException throws if the operation failed.
     */
    public InputStream getContentStream(int contentID) throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String resourceContentSQL = "SELECT REG_CONTENT_DATA  FROM  REG_CONTENT WHERE " +
                    "REG_CONTENT_ID = ? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(resourceContentSQL);
            ps.setLong(1, contentID);
            ps.setInt(2, CurrentSession.getTenantId());
            result = ps.executeQuery();
            if (result.next()) {
                InputStream rawInputStream =
                        result.getBinaryStream(DatabaseConstants.CONTENT_DATA_FIELD);
                if (rawInputStream != null) {
                    return RegistryUtils.getMemoryStream(rawInputStream);
                }
            }

        }
        catch (SQLException ex) {
            String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
            log.error(msg, ex);
            throw new RegistryException(msg, ex);
        }
        finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            } catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
        return null;
    }

    /**
     * Save the updates of a given resource.
     *
     * @param resourceImpl the resource to be updated.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void update(ResourceImpl resourceImpl) throws RegistryException {

        if (!(resourceImpl instanceof CollectionImpl)) {
            // adding content new
            addContent(resourceImpl);
        }
        // there is no difference of adding a resource and updating as it is always
        // a new entry to the resource table
        addResource(resourceImpl, true);

        // adding properties..
        addProperties(resourceImpl);

    }

    /**
     * Method to return a child count of a collection (database connection should also be provided)
     *
     * @param collection the collection object which the children are calculated.
     * @param conn       the database connection.
     *
     * @return the child count.
     * @throws RegistryException throws if the operation failed.
     */
    public int getChildCount(CollectionImpl collection, Connection conn)
            throws RegistryException {

        ResourceIDImpl resourceID = collection.getResourceIDImpl();
        ResultSet results1 = null;
        PreparedStatement ps1 = null;
        ResultSet results2 = null;
        PreparedStatement ps2 = null;


        int childCount = 0;
        try {
            // step1: get the child resources.
            String sql = "SELECT COUNT(R.REG_NAME) AS REG_RES_COUNT FROM REG_RESOURCE R WHERE " +
                    "R.REG_PATH_ID=? AND R.REG_TENANT_ID=?";

            ps1 = conn.prepareStatement(sql);
            ps1.setInt(1, resourceID.getPathID());
            ps1.setInt(2, CurrentSession.getTenantId());
            results1 = ps1.executeQuery();

            if (results1.next()) {
                // here -1 is to avoid the parent collection itself
                childCount += results1.getInt(DatabaseConstants.RES_COUNT_FIELD);
            }

            // step2: get the child collections
            sql = "SELECT COUNT(P.REG_PATH_ID) AS REG_RES_COUNT FROM REG_PATH P, REG_RESOURCE R " +
                    "WHERE P.REG_PATH_PARENT_ID=? AND P.REG_TENANT_ID=? AND " +
                    "R.REG_PATH_ID=P.REG_PATH_ID AND R.REG_TENANT_ID=?";
            ps2 = conn.prepareStatement(sql);
            ps2.setInt(1, resourceID.getPathID());
            ps2.setInt(2, CurrentSession.getTenantId());
            ps2.setInt(3, CurrentSession.getTenantId());

            results2 = ps2.executeQuery();

            if (results2.next()) {
                childCount += results2.getInt(DatabaseConstants.RES_COUNT_FIELD);
            }

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    collection.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results1 != null) {
                    results1.close();
                }
                if (ps1 != null) {
                    ps1.close();
                }
                if (results2 != null) {
                    results2.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
        return childCount;
    }

    /**
     * Fill the children for a resource that already filled with meta data. Children are filled only
     * at the at of the intersect of the given range and resource existence range.
     *
     * @param collection collection to fill the children and properties.
     * @param start      start value of the range of children.
     * @param pageLen    the length of the children to retrieve
     *
     * @throws RegistryException if the operation failed.
     */
    public void fillChildren(CollectionImpl collection, int start, int pageLen)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        fillChildren(collection, start, pageLen, conn);
    }

    /**
     * Fill the children for a resource that already filled with meta data. Children are filled only
     * at the at of the intersect of the given range and resource existence range.
     *
     * @param collection collection to fill the children and properties.
     * @param start      start value of the range of children.
     * @param pageLen    the length of the children to retrieve
     * @param conn       the database connection
     *
     * @throws RegistryException if the operation failed.
     */
    public void fillChildren(CollectionImpl collection, int start, int pageLen, Connection conn)
            throws RegistryException {
        String[] childPaths = getChildren(collection, start, pageLen, conn);
        collection.setContent(childPaths);
    }

    /**
     * Get the children of the collection. Children are filled only at the at of the intersect of
     * the given range and resource existence range.
     *
     * @param collection collection to fill the children and properties.
     * @param start      start value of the range of children.
     * @param pageLen    the length of the children to retrieve
     * @param conn       the database connection
     *
     * @return an array of children paths
     * @throws RegistryException throws if the operation failed.
     */
    public String[] getChildren(CollectionImpl collection, int start, int pageLen, Connection conn)
            throws RegistryException {

        ResourceIDImpl resourceID = collection.getResourceIDImpl();
        ResultSet results1 = null;
        PreparedStatement ps1 = null;
        ResultSet results2 = null;
        PreparedStatement ps2 = null;

        ResourcePath parentResourcePath = new ResourcePath(resourceID.getPath());
        String parentPath = parentResourcePath.getPath();
        try {
            // step1: get the child resources.
            String sql =
                    "SELECT R.REG_NAME FROM REG_RESOURCE R WHERE R.REG_PATH_ID=? " +
                            "AND R.REG_TENANT_ID=?";

            ps1 = conn.prepareStatement(sql);
            ps1.setInt(1, resourceID.getPathID());
            ps1.setInt(2, CurrentSession.getTenantId());
            results1 = ps1.executeQuery();

            List<String> childResourcesList = new ArrayList<String>();
            List<String> childCollectionsList = new ArrayList<String>();
            while (results1.next()) {
                String childName = results1.getString(DatabaseConstants.NAME_FIELD);
                if (childName == null) {
                    // skip the null named resource (which is the parent resource
                    continue;
                }
                String childPath = parentPath +
                        (parentPath.equals(RegistryConstants.PATH_SEPARATOR) ? "" :
                                RegistryConstants.PATH_SEPARATOR) +
                        childName;

                ResourceIDImpl childResourceID = new ResourceIDImpl();
                childResourceID.setPath(childPath);
                childResourceID.setCollection(false);
                childResourceID.setName(childName);
                childResourceID.setPathID(resourceID.getPathID());
                if (AuthorizationUtils.authorize(childPath, ActionConstants.GET)) {
                    childResourcesList.add(childPath);
                }
            }

            int end = start + pageLen;
            if (childResourcesList.size() < end || pageLen == -1) {
                // step2: get the child collections
                sql = "SELECT P.REG_PATH_ID, P.REG_PATH_VALUE FROM REG_PATH P, REG_RESOURCE R " +
                        "WHERE P.REG_PATH_PARENT_ID=? AND P.REG_TENANT_ID=? AND " +
                        "R.REG_PATH_ID=P.REG_PATH_ID AND " +
                        "R.REG_NAME IS NULL AND R.REG_TENANT_ID=?";
                ps2 = conn.prepareStatement(sql);
                ps2.setInt(1, resourceID.getPathID());
                ps2.setInt(2, CurrentSession.getTenantId());
                ps2.setInt(3, CurrentSession.getTenantId());

                results2 = ps2.executeQuery();

                while (results2.next()) {
                    String childPath = results2.getString(DatabaseConstants.PATH_VALUE_FIELD);
                    int childPathId = results2.getInt(DatabaseConstants.PATH_ID_FIELD);

                    ResourceIDImpl childResourceID = new ResourceIDImpl();
                    childResourceID.setPath(childPath);
                    childResourceID.setCollection(true);
                    childResourceID.setName(null);
                    childResourceID.setPathID(childPathId);
                    if (AuthorizationUtils.authorize(childPath, ActionConstants.GET)) {
                        childCollectionsList.add(childPath);
                    }
                }
            }

            int totalSize = childResourcesList.size() + childCollectionsList.size();
            if (totalSize < start) {
                throw new RegistryException("Didn't have enough results to start at #" + start);
            }
            Collections.sort(childResourcesList);
            Collections.sort(childCollectionsList);

            if (pageLen == -1) {
                pageLen = totalSize - start;
                end = totalSize - 1;
            }

            int resultMaxSize = pageLen < (totalSize - start) ? pageLen : (totalSize - start);
            String[] childPaths = new String[resultMaxSize];
            // copying child resources
            int current = 0;
            int i;
            for (i = 0; i < childResourcesList.size(); i++) {
                if (current >= resultMaxSize) {
                    break;
                }
                if (i >= start && i <= end) {
                    childPaths[current] = childResourcesList.get(i);
                    current++;
                }
            }
            // copying child collections
            for (int j = 0; j < childCollectionsList.size(); j++, i++) {
                if (current >= resultMaxSize) {
                    break;
                }
                if (i >= start && i <= end) {
                    childPaths[current] = childCollectionsList.get(j);
                    current++;
                }
            }
            return childPaths;

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    collection.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results1 != null) {
                    results1.close();
                }
                if (ps1 != null) {
                    ps1.close();
                }
                if (results2 != null) {
                    results2.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to return the resource meta data (excluding properties, content and children)
     *
     * @param path the of the resource
     *
     * @return resource instance with the meta data filled.
     * @throws RegistryException throws if the operation failed.
     */
    public ResourceImpl getResourceMetaData(String path) throws RegistryException {
        ResourceIDImpl resourceIDImpl = getResourceID(path);
        ResourceImpl resourceImpl = null;
        if (resourceIDImpl != null) {
            resourceImpl = getResourceMetaData(resourceIDImpl);

            if (resourceImpl == null && resourceIDImpl.isCollection()) {
                // we should check the resourceID for a non-collection too.
                resourceIDImpl = getResourceID(path, false);
                if (resourceIDImpl != null) {
                    resourceImpl = getResourceMetaData(resourceIDImpl);
                }
            }
        }
        return resourceImpl;
    }

    /**
     * Method to return the resource meta data (excluding properties, content and children)
     *
     * @param resourceID the resource id
     *
     * @return resource instance with the meta data filled.
     * @throws RegistryException throws if the operation failed.
     */
    public ResourceImpl getResourceMetaData(ResourceIDImpl resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        ResourceImpl resourceImpl = null;
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            boolean isCollection = resourceID.isCollection();
            String sql;
            int pathID = resourceID.getPathID();
            String resourceName = resourceID.getName();
            if (isCollection) {
                // collection
                sql = "SELECT REG_MEDIA_TYPE, REG_CREATOR, REG_CREATED_TIME, " +
                        "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_VERSION, REG_DESCRIPTION, " +
                        "REG_CONTENT_ID FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME " +
                        "IS NULL AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, pathID);
                ps.setInt(2, CurrentSession.getTenantId());
            } else {
                // non collection
                sql = "SELECT REG_MEDIA_TYPE, REG_CREATOR, REG_CREATED_TIME, " +
                        "REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, REG_VERSION, REG_DESCRIPTION, " +
                        "REG_CONTENT_ID FROM REG_RESOURCE WHERE REG_PATH_ID=? AND REG_NAME = ? " +
                        "AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, pathID);
                ps.setString(2, resourceName);
                ps.setInt(3, CurrentSession.getTenantId());
            }

            result = ps.executeQuery();

            if (result.next()) {

                if (isCollection) {
                    resourceImpl = new CollectionImpl();
                } else {
                    resourceImpl = new ResourceImpl();
                }

                // this is always the current version of the resource
                resourceImpl.setPathID(resourceID.getPathID());
                resourceImpl.setPath(resourceID.getPath());
                resourceImpl.setName(resourceID.getName());
                resourceImpl.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
                resourceImpl.setAuthorUserName(result.getString(DatabaseConstants.CREATOR_FIELD));
                resourceImpl
                        .setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
                resourceImpl.setLastUpdaterUserName(
                        result.getString(DatabaseConstants.LAST_UPDATER_FIELD));
                resourceImpl.setLastModified(
                        result.getTimestamp(DatabaseConstants.LAST_UPDATED_TIME_FIELD));
                resourceImpl.setVersionNumber(result.getInt(DatabaseConstants.VERSION_FIELD));
                resourceImpl.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
                resourceImpl.setDbBasedContentID(result.getInt(DatabaseConstants.CONTENT_ID_FIELD));
            }
        } catch (SQLException e) {
            String msg = "Failed to get the resource at path " + resourceID.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        return resourceImpl;
    }

    /**
     * Method to return a resource when the resourceID is provided.
     *
     * @param resourceID the resource id.
     *
     * @return the resource for the given id.
     * @throws RegistryException throws if the operation failed.
     */
    private ResourceImpl getResource(ResourceIDImpl resourceID) throws RegistryException {
        ResourceImpl resourceImpl = getResourceMetaData(resourceID);

        if (resourceImpl != null && !(resourceImpl instanceof CollectionImpl) &&
                resourceImpl.getDbBasedContentID() > 0) {
            fillResourceContent(resourceImpl);
        }
        return resourceImpl;
    }

    /**
     * Method to get resource without setting the resource modified flags on.
     *
     * @param resourceID the resource id.
     *
     * @return the resource for the given id.
     * @throws RegistryException throws if the operation failed.
     */
    public ResourceImpl getResourceWithNoUpdate(ResourceIDImpl resourceID)
            throws RegistryException {
        ResourceImpl resourceImpl = getResourceMetaData(resourceID);

        if (resourceImpl != null && !(resourceImpl instanceof CollectionImpl) &&
                resourceImpl.getDbBasedContentID() > 0) {
            fillResourceContentWithNoUpdate(resourceImpl);
        }
        return resourceImpl;
    }

    /**
     * Fill resource content for a given resource implementation.
     *
     * @param resourceImpl resource object.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void fillResourceContent(ResourceImpl resourceImpl) throws RegistryException {
        int contentId = resourceImpl.getDbBasedContentID();
        InputStream contentStream = getContentStream(contentId);
        if (contentStream == null) {
            resourceImpl.setContent(null);
        } else {
            resourceImpl.setContentStream(contentStream);
        }
    }

    /**
     * Fill resource content for a given resource implementation without setting the resource
     * modified flag on.
     *
     * @param resourceImpl resource object.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void fillResourceContentWithNoUpdate(ResourceImpl resourceImpl)
            throws RegistryException {
        int contentId = resourceImpl.getDbBasedContentID();
        InputStream contentStream = getContentStream(contentId);
        if (contentStream == null) {
            resourceImpl.setContentWithNoUpdate(null);
        } else {
            resourceImpl.setContentStreamWithNoUpdate(contentStream);
        }
    }


    /**
     * Update the last updated time of a resource, This is called to update the parent's last
     * updated time when a child is created, deleted or moved out/in, copy in.
     *
     * @param resourceID the id of the resource to get the version of.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void updateCollectionLastUpdatedTime(ResourceIDImpl resourceID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "UPDATE REG_RESOURCE SET REG_LAST_UPDATED_TIME=? WHERE " +
                        "REG_PATH_ID=? AND REG_NAME IS NULL AND REG_TENANT_ID=?";

            ps = conn.prepareStatement(sql);

            long now = System.currentTimeMillis();
            ps.setTimestamp(1, new Timestamp(now));
            ps.setLong(2, resourceID.getPathID());
            ps.setInt(3, CurrentSession.getTenantId());

            ps.executeUpdate();

        } catch (SQLException e) {
            String msg = "Failed to update the last updated time for the collection. " +
                    "path: " + resourceID.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Update the content id of a resource, Normally this should be called after calling
     * addResourceWithoutContentId is called.
     *
     * @param resourceImpl       the resource object.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void updateContentId(ResourceImpl resourceImpl)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "UPDATE REG_RESOURCE SET REG_CONTENT_ID=? WHERE REG_VERSION=? " +
                    "AND REG_TENANT_ID=?";

            ps = conn.prepareStatement(sql);

            ps.setInt(1, resourceImpl.getDbBasedContentID());
            ps.setLong(2, resourceImpl.getVersionNumber());
            ps.setInt(3, CurrentSession.getTenantId());

            ps.executeUpdate();

        } catch (SQLException e) {
            String msg = "Failed to update the content id: " + resourceImpl.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Add a resource without a content id, provided whether it is overwriting existing one or not.
     * If the resource is already existing the removal of the older resource will not be handled in
     * this function.
     *
     * @param resourceImpl       the resource object.
     * @param isUpdatingExisting whether the resource is updating or not.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void addResourceWithoutContentId(ResourceImpl resourceImpl, boolean isUpdatingExisting)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
        try {
            String sql =
                    "INSERT INTO REG_RESOURCE (REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, " +
                            "REG_CREATOR, REG_CREATED_TIME, REG_LAST_UPDATOR, " +
                            "REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                            "REG_TENANT_ID) " +
                            "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)";


            long now = System.currentTimeMillis();

            String dbProductName = conn.getMetaData().getDatabaseProductName();
            ps = conn.prepareStatement(sql, new String[]{
                    DBUtils.getConvertedAutoGeneratedColumnName(dbProductName, "REG_VERSION")});

            ps.setInt(1, resourceID.getPathID());
            ps.setString(2, resourceID.getName());
            ps.setString(3, resourceImpl.getMediaType());
            if (isUpdatingExisting) {
                String authorName = resourceImpl.getAuthorUserName();
                if (authorName == null) {
                    authorName = CurrentSession.getUser();
                    resourceImpl.setAuthorUserName(authorName);
                }
                ps.setString(4, authorName);

                Date createdTime = resourceImpl.getCreatedTime();
                Timestamp createdTimestamp;
                if (createdTime == null) {
                    createdTimestamp = new Timestamp(now);
                } else {
                    createdTimestamp = new Timestamp(createdTime.getTime());
                }
                ps.setTimestamp(5, createdTimestamp);
            } else {
                ps.setString(4, CurrentSession.getUser());
                resourceImpl.setAuthorUserName(CurrentSession.getUser());
                ps.setTimestamp(5, new Timestamp(now));
            }
            ps.setString(6, CurrentSession.getUser());
            ps.setTimestamp(7, new Timestamp(now));
            ps.setString(8, resourceImpl.getDescription());
            ps.setInt(9, CurrentSession.getTenantId());

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if (result.next()) {
                long version = result.getLong(1);
                resourceImpl.setVersionNumber(version);
            }


        } catch (SQLException e) {
            String msg = "Failed to add resource to path " + resourceImpl.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Add a resource, provided whether it is overwriting existing one or not. If the resource
     * already existing the removal of the older resource will not be handled in this function.
     *
     * @param resourceImpl       the resource object.
     * @param isUpdatingExisting whether the resource is updating or not.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void addResource(ResourceImpl resourceImpl, boolean isUpdatingExisting)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
        try {
            String sql =
                    "INSERT INTO REG_RESOURCE (REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, " +
                            "REG_CREATOR, REG_CREATED_TIME, REG_LAST_UPDATOR, " +
                            "REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                            "REG_CONTENT_ID, REG_TENANT_ID) " +
                            "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";


            long now = System.currentTimeMillis();

            String dbProductName = conn.getMetaData().getDatabaseProductName();
            ps = conn.prepareStatement(sql, new String[]{
                    DBUtils.getConvertedAutoGeneratedColumnName(dbProductName, "REG_VERSION")});

            ps.setInt(1, resourceID.getPathID());
            ps.setString(2, resourceID.getName());
            ps.setString(3, resourceImpl.getMediaType());
            if (isUpdatingExisting) {
                String authorName = resourceImpl.getAuthorUserName();
                if (authorName == null) {
                    authorName = CurrentSession.getUser();
                    resourceImpl.setAuthorUserName(authorName);
                }
                ps.setString(4, authorName);

                Date createdTime = resourceImpl.getCreatedTime();
                Timestamp createdTimestamp;
                if (createdTime == null) {
                    createdTimestamp = new Timestamp(now);
                } else {
                    createdTimestamp = new Timestamp(createdTime.getTime());
                }
                ps.setTimestamp(5, createdTimestamp);
            } else {
                ps.setString(4, CurrentSession.getUser());
                resourceImpl.setAuthorUserName(CurrentSession.getUser());
                ps.setTimestamp(5, new Timestamp(now));
            }
            ps.setString(6, CurrentSession.getUser());
            ps.setTimestamp(7, new Timestamp(now));
            ps.setString(8, resourceImpl.getDescription());
            if (resourceImpl.getDbBasedContentID() > 0) {
                ps.setInt(9, resourceImpl.getDbBasedContentID());
            } else {
                ps.setNull(9, Types.INTEGER);
            }
            ps.setInt(10, CurrentSession.getTenantId());

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if (result.next()) {
                long version = result.getLong(1);
                resourceImpl.setVersionNumber(version);
            }


        } catch (SQLException e) {
            String msg = "Failed to add resource to path " + resourceImpl.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Add a resource without setting the resource modified flag on.
     *
     * @param resourceImpl the resource to be added.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void addResourceWithNoUpdate(ResourceImpl resourceImpl)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        ResourceIDImpl resourceID = resourceImpl.getResourceIDImpl();
        try {
            String sql =
                    "INSERT INTO REG_RESOURCE (REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, " +
                            "REG_CREATOR, REG_CREATED_TIME, REG_LAST_UPDATOR, " +
                            "REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                            "REG_CONTENT_ID, REG_TENANT_ID) " +
                            "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

            String dbProductName = conn.getMetaData().getDatabaseProductName();
            ps = conn.prepareStatement(sql, new String[]{
                    DBUtils.getConvertedAutoGeneratedColumnName(dbProductName, "REG_VERSION")});

            ps.setInt(1, resourceID.getPathID());
            ps.setString(2, resourceID.getName());
            ps.setString(3, resourceImpl.getMediaType());
            ps.setString(4, resourceImpl.getAuthorUserName());

            Date createdTime = resourceImpl.getCreatedTime();
            ps.setTimestamp(5, new Timestamp(createdTime.getTime()));

            ps.setString(6, resourceImpl.getLastUpdaterUserName());

            Date updatedTime = resourceImpl.getLastModified();
            ps.setTimestamp(7, new Timestamp(updatedTime.getTime()));

            ps.setString(8, resourceImpl.getDescription());
            if (resourceImpl.getDbBasedContentID() > 0) {
                ps.setInt(9, resourceImpl.getDbBasedContentID());
            } else {
                ps.setNull(9, Types.INTEGER);
            }
            ps.setInt(10, CurrentSession.getTenantId());

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if (result.next()) {
                long version = result.getLong(1);
                resourceImpl.setVersionNumber(version);
            }

        } catch (SQLException e) {
            String msg = "Failed to add resource to path " + resourceImpl.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Add the resource data object.
     *
     * @param resourceDO the resource data object.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void addResourceDO(ResourceDO resourceDO)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;

        try {
            String sql =
                    "INSERT INTO REG_RESOURCE (REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, " +
                            "REG_CREATOR, REG_CREATED_TIME, REG_LAST_UPDATOR, " +
                            "REG_LAST_UPDATED_TIME, REG_DESCRIPTION, " +
                            "REG_CONTENT_ID, REG_TENANT_ID) " +
                            "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

            String dbProductName = conn.getMetaData().getDatabaseProductName();
            ps = conn.prepareStatement(sql, new String[]{DBUtils
                    .getConvertedAutoGeneratedColumnName(dbProductName,
                    "REG_VERSION")});

            ps.setInt(1, resourceDO.getPathID());
            ps.setString(2, resourceDO.getName());
            ps.setString(3, resourceDO.getMediaType());
            ps.setString(4, resourceDO.getAuthor());
            ps.setTimestamp(5, resourceDO.getCreatedOn());
            ps.setString(6, resourceDO.getLastUpdater());
            ps.setTimestamp(7, resourceDO.getLastUpdatedOn());
            ps.setString(8, resourceDO.getDescription());
            if (resourceDO.getContentID() > 0) {
                ps.setInt(9, resourceDO.getContentID());
            } else {
                ps.setNull(9, Types.INTEGER);
            }
            ps.setInt(10, CurrentSession.getTenantId());

            ps.executeUpdate();

            result = ps.getGeneratedKeys();
            if (result.next()) {
                long version = result.getLong(1);
                resourceDO.setVersion(version);
            }

            ps.close();

        } catch (SQLException e) {
            String msg = "Failed to add resource to version " + resourceDO.getVersion() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Delete the resource provided as a resource DO
     *
     * @param resourceDO the resource to be deleted.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void deleteResource(ResourceDO resourceDO) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (resourceDO.getName() == null) {
                String sql = "DELETE FROM REG_RESOURCE WHERE REG_PATH_ID=? AND " +
                        "REG_NAME IS NULL AND REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());

                ps.executeUpdate();
            } else {
                String sql = "DELETE FROM REG_RESOURCE WHERE REG_PATH_ID=? AND " +
                        "REG_NAME=? AND REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());
                ps.setString(2, resourceDO.getName());
                ps.setInt(3, CurrentSession.getTenantId());

                ps.executeUpdate();
            }

        } catch (SQLException e) {

            String msg = "Failed to delete the resource with id " + resourceDO.getVersion() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Add the properties to the database from  given resource
     *
     * @param resource to add properties for
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void addProperties(ResourceImpl resource) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps1 = null;
        PreparedStatement ps2 = null;
        try {

            String sql1, sql2;
            boolean isVersioningProperties = StaticConfiguration.isVersioningProperties();

            sql1 = "INSERT INTO REG_PROPERTY (REG_NAME, REG_VALUE, REG_TENANT_ID) VALUES (?, ?, ?)";
            if (isVersioningProperties) {
                sql2 = "INSERT INTO REG_RESOURCE_PROPERTY (REG_PROPERTY_ID, REG_VERSION, " +
                        "REG_TENANT_ID) VALUES (?, ?, ?)";
            } else {
                sql2 = "INSERT INTO REG_RESOURCE_PROPERTY (REG_PROPERTY_ID, REG_PATH_ID, " +
                        "REG_RESOURCE_NAME, REG_TENANT_ID) VALUES (?, ?, ?, ?)";
            }

            ResourceIDImpl resourceID = resource.getResourceIDImpl();
            Properties props = resource.getProperties();

            if (props != null) {
                for (Object nameObj : props.keySet()) {
                    String name = (String) nameObj;
                    List propValues = (List) props.get(name);
                    if (propValues != null) {
                        for (Object valueObj : propValues) {
                            String value = (String) valueObj;
                            String dbProductName = conn.getMetaData().getDatabaseProductName();
                            ps1 = conn.prepareStatement(sql1,
                                    new String[]{DBUtils
                                            .getConvertedAutoGeneratedColumnName(
                                            dbProductName,
                                            "REG_ID")});
                            ps2 = conn.prepareStatement(sql2);

                            // prepare to execute query1 for the property
                            ps1.setString(1, name);
                            ps1.setString(2, value);
                            ps1.setInt(3, CurrentSession.getTenantId());
                            ps1.executeUpdate();

                            ResultSet resultSet1 = ps1.getGeneratedKeys();
                            if (resultSet1.next()) {
                                // setting the property id
                                int propertyId = resultSet1.getInt(1);
                                ps2.setInt(1, propertyId);
                                if (isVersioningProperties) {
                                    ps2.setLong(2, resource.getVersionNumber());
                                    ps2.setInt(3, CurrentSession.getTenantId());
                                } else {
                                    ps2.setLong(2, resourceID.getPathID());
                                    if (resourceID.getName() == null) {
                                        ps2.setNull(3, Types.VARCHAR);
                                    } else {
                                        ps2.setString(3, resourceID.getName());
                                    }
                                    ps2.setInt(4, CurrentSession.getTenantId());
                                }
                                ps2.executeUpdate();
                            }
                            if (ps1 != null) {
                                ps1.close();
                                ps1 = null;
                            }
                            if (ps2 != null) {
                                ps2.close();
                                ps2 = null;
                            }
                        }
                    }
                }
            }
        } catch (SQLException e) {

            String msg = "Failed to add properties to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps1 != null) {
                    ps1.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Get id values of properties of a resource.
     *
     * @param resourceDO the resource which properties are to retrieve.
     *
     * @return the array of property ids.
     * @throws RegistryException throws if the operation failed.
     */
    private Integer[] getPropertyIds(ResourceDO resourceDO) throws RegistryException {

        Connection conn = Transaction.getConnection();

        List<Integer> propertyIds = new ArrayList<Integer>();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            String sql;
            if (StaticConfiguration.isVersioningProperties()) {
                sql = "SELECT REG_PROPERTY_ID FROM REG_RESOURCE_PROPERTY WHERE " +
                        "REG_VERSION=? AND REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setLong(1, resourceDO.getVersion());
                ps.setInt(2, CurrentSession.getTenantId());

            } else if (resourceDO.getName() == null) {
                // a collection
                sql = "SELECT REG_PROPERTY_ID FROM REG_RESOURCE_PROPERTY WHERE " +
                        "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL AND REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());
            } else {
                sql = "SELECT REG_PROPERTY_ID FROM REG_RESOURCE_PROPERTY WHERE " +
                        "REG_PATH_ID=? AND REG_RESOURCE_NAME=? AND REG_TENANT_ID=?";

                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceDO.getPathID());
                ps.setString(2, resourceDO.getName());
                ps.setInt(3, CurrentSession.getTenantId());
            }

            results = ps.executeQuery();
            while (results.next()) {
                int propertyId = results.getInt(DatabaseConstants.PROPERTY_ID_FIELD);
                propertyIds.add(propertyId);
            }

        } catch (SQLException e) {

            String msg = "Failed to get properties on resource version" +
                    resourceDO.getVersion() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }

        return propertyIds.toArray(new Integer[propertyIds.size()]);
    }

    /**
     * Remove properties of a resource.
     *
     * @param resourceDO the resource DO which the properties have to be deleted.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void removeProperties(ResourceDO resourceDO) throws RegistryException {
        Connection conn = Transaction.getConnection();

        PreparedStatement ps1 = null, ps2 = null;

        Integer[] propertyIds = getPropertyIds(resourceDO);
        if (propertyIds == null) {
            return;
        }
        try {

            String sql =
                    "DELETE FROM REG_RESOURCE_PROPERTY WHERE REG_PROPERTY_ID= ? " +
                            "AND REG_TENANT_ID=?";
            ps1 = conn.prepareStatement(sql);

            sql = "DELETE FROM REG_PROPERTY WHERE REG_ID= ? AND REG_TENANT_ID=?";
            ps2 = conn.prepareStatement(sql);

            for (Integer propertyId : propertyIds) {
                ps1.setInt(1, propertyId);
                ps1.setInt(2, CurrentSession.getTenantId());
                ps2.setInt(1, propertyId);
                ps2.setInt(2, CurrentSession.getTenantId());
                ps1.addBatch();
                ps2.addBatch();
            }

            if (propertyIds.length > 0) {
                try {
                    ps1.executeBatch();
                    ps2.executeBatch();
                } catch (SQLException e) {
                    ps1.clearBatch();
                    ps2.clearBatch();
                    // the exception will be handled in the next catch block
                    throw e;
                }
            }
        } catch (SQLException e) {

            String msg = "Failed to remove properties from resource version " +
                    resourceDO.getVersion() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps1 != null) {
                    ps1.close();
                }
                if (ps2 != null) {
                    ps2.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Add the content for a resource.
     *
     * @param resourceImpl the resource to add content.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void addContent(ResourceImpl resourceImpl) throws RegistryException {

        if (resourceImpl.getContent() == null || !(resourceImpl.getContent() instanceof byte[])) {
            if (log.isDebugEnabled()) {
                String msg = "Content of the resource " + resourceImpl.getPath() + " is null or " +
                        "not a byte array. Content will not be persisted to the database.";
                log.debug(msg);
            }
            resourceImpl.setDbBasedContentID(0);
            return;
        }

        InputStream contentStream = resourceImpl.getContentStream();
        int contentID = addContentBytes(contentStream);
        resourceImpl.setDbBasedContentID(contentID);
    }

    /**
     * Add the content to the content table and return the auto generated id of content table.
     *
     * @param contentStream the input stream.
     *
     * @return the auto generated id of content table.
     * @throws RegistryException throws if the operation failed.
     */
    public int addContentBytes(InputStream contentStream) throws RegistryException {

        Connection conn = Transaction.getConnection();
        int contentID = -1;

        try {
            String sql = "INSERT INTO REG_CONTENT (REG_CONTENT_DATA, REG_TENANT_ID) VALUES (?, ?)";

            int size = contentStream.available();
            String dbProductName = conn.getMetaData().getDatabaseProductName();
            PreparedStatement ps = conn.prepareStatement(sql,
                    new String[]{DBUtils.getConvertedAutoGeneratedColumnName(
                            dbProductName, "REG_CONTENT_ID")});
            ps.setBinaryStream(1, contentStream, size);
            ps.setInt(2, CurrentSession.getTenantId());
            ps.executeUpdate();

            ResultSet result = ps.getGeneratedKeys();
            if (result.next()) {
                contentID = result.getInt(1);
            }
            ps.close();

        } catch (Exception e) {

            String msg = "Failed to write resource content to the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        return contentID;
    }

    /**
     * Method to return resourceDO from a version number.
     *
     * @param version the version of the resource.
     *
     * @return the resourceDO for the given version.
     * @throws RegistryException throws if the operation failed.
     */
    public ResourceDO getResourceDO(long version) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {

            String sql =
                    "SELECT REG_PATH_ID, REG_NAME, REG_MEDIA_TYPE, REG_CREATOR, " +
                            "REG_CREATED_TIME, REG_LAST_UPDATOR, REG_LAST_UPDATED_TIME, " +
                            "REG_DESCRIPTION, REG_CONTENT_ID " +
                            "FROM REG_RESOURCE WHERE REG_VERSION = ? AND REG_TENANT_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setLong(1, version);
            ps.setInt(2, CurrentSession.getTenantId());

            ResourceDO resourceDO = null;
            result = ps.executeQuery();
            if (result.next()) {
                resourceDO = new ResourceDO();

                resourceDO.setPathID(result.getInt(DatabaseConstants.PATH_ID_FIELD));
                resourceDO.setName(result.getString(DatabaseConstants.NAME_FIELD));
                resourceDO.setVersion(version);
                resourceDO.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
                resourceDO.setAuthor(result.getString(DatabaseConstants.CREATOR_FIELD));
                resourceDO.setCreatedOn(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
                resourceDO.setLastUpdater(result.getString(DatabaseConstants.LAST_UPDATER_FIELD));
                resourceDO.setLastUpdatedOn(
                        result.getTimestamp(DatabaseConstants.LAST_UPDATED_TIME_FIELD));
                resourceDO.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
                resourceDO.setContentID(result.getInt(DatabaseConstants.CONTENT_ID_FIELD));
            }

            return resourceDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource version:" +
                    version + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Method to get resource from resource id.
     *
     * @param resourceID the resource id.
     *
     * @return the resource DO for the resource id.
     * @throws RegistryException throws if the operation failed.
     */
    public ResourceDO getResourceDO(ResourceIDImpl resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql;
            if (resourceID.isCollection()) {
                sql = "SELECT  R.REG_PATH_ID, R.REG_NAME, R.REG_VERSION, R.REG_MEDIA_TYPE, " +
                        "R.REG_CREATOR, R.REG_CREATED_TIME, R.REG_LAST_UPDATOR, " +
                        "R.REG_LAST_UPDATED_TIME, R.REG_DESCRIPTION, R.REG_CONTENT_ID " +
                        "FROM REG_RESOURCE R WHERE R.REG_PATH_ID=? AND R.REG_NAME IS NULL " +
                        "AND R.REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                ps.setInt(2, CurrentSession.getTenantId());

            } else {
                sql = "SELECT  R.REG_PATH_ID, R.REG_NAME, R.REG_VERSION, R.REG_MEDIA_TYPE, " +
                        "R.REG_CREATOR, R.REG_CREATED_TIME, R.REG_LAST_UPDATOR, " +
                        "R.REG_LAST_UPDATED_TIME, R.REG_DESCRIPTION, R.REG_CONTENT_ID " +
                        "FROM REG_RESOURCE R WHERE R.REG_PATH_ID=? AND R.REG_NAME=? " +
                        "AND R.REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, resourceID.getPathID());
                ps.setString(2, resourceID.getName());
                ps.setInt(3, CurrentSession.getTenantId());
            }

            ResourceDO resourceDO = null;
            result = ps.executeQuery();
            if (result.next()) {
                resourceDO = new ResourceDO();

                resourceDO.setPathID(result.getInt(DatabaseConstants.PATH_ID_FIELD));
                resourceDO.setName(result.getString(DatabaseConstants.NAME_FIELD));
                resourceDO.setVersion(result.getInt(DatabaseConstants.VERSION_FIELD));
                resourceDO.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
                resourceDO.setAuthor(result.getString(DatabaseConstants.CREATOR_FIELD));
                resourceDO.setCreatedOn(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
                resourceDO.setLastUpdater(result.getString(DatabaseConstants.LAST_UPDATER_FIELD));
                resourceDO.setLastUpdatedOn(
                        result.getTimestamp(DatabaseConstants.LAST_UPDATED_TIME_FIELD));
                resourceDO.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
                resourceDO.setContentID(result.getInt(DatabaseConstants.CONTENT_ID_FIELD));
            }

            return resourceDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            if (result != null) {
                try {
                    result.close();
                } catch (SQLException e) {
                    log.error(RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR, e);
                }
            }
            if (ps != null) {
                try {
                    ps.close();
                } catch (SQLException e) {
                    log.error(RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR, e);
                }
            }
        }
    }

    /**
     * Get the child path ids of a resource, (should be a collection)
     *
     * @param resourceID the resource id of the collection.
     *
     * @return an array of child path ids.
     * @throws RegistryException throws if the operation failed.
     */
    public ArrayList<ResourceIDImpl> getChildPathIds(ResourceIDImpl resourceID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        ArrayList<ResourceIDImpl> childPathList = new ArrayList<ResourceIDImpl>();
        PreparedStatement ps = null;
        ResultSet results = null;
        try {
            // step1: get the child resources.
            String sql =
                    "SELECT R.REG_NAME FROM REG_RESOURCE R WHERE R.REG_PATH_ID=? " +
                            "AND R.REG_TENANT_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setInt(1, resourceID.getPathID());
            ps.setInt(2, CurrentSession.getTenantId());
            results = ps.executeQuery();

            while (results.next()) {
                String childName = results.getString(DatabaseConstants.NAME_FIELD);
                if (childName == null) {
                    // skip the parent resource..
                    continue;
                }
                String parentPath = resourceID.getPath();
                String childPath = parentPath +
                        (parentPath.equals(RegistryConstants.PATH_SEPARATOR) ? "" :
                                RegistryConstants.PATH_SEPARATOR) +
                        childName;
                ResourceIDImpl childResourceID = new ResourceIDImpl();
                childResourceID.setPath(childPath);
                childResourceID.setCollection(false);
                childResourceID.setName(childName);
                childResourceID.setPathID(resourceID.getPathID());
                childPathList.add(childResourceID);
            }
            if (results != null) {
                results.close();
                results = null;
            }
            if (ps != null) {
                ps.close();
                ps = null;
            }

            sql = "SELECT P.REG_PATH_ID, P.REG_PATH_VALUE FROM REG_PATH P, REG_RESOURCE R " +
                    "WHERE P.REG_PATH_PARENT_ID=? AND P.REG_TENANT_ID=? AND " +
                    "R.REG_PATH_ID=P.REG_PATH_ID AND R.REG_NAME IS NULL AND R.REG_TENANT_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, resourceID.getPathID());
            ps.setInt(2, CurrentSession.getTenantId());
            ps.setInt(3, CurrentSession.getTenantId());

            results = ps.executeQuery();

            while (results.next()) {
                String childPath = results.getString(DatabaseConstants.PATH_VALUE_FIELD);
                int childPathId = results.getInt(DatabaseConstants.PATH_ID_FIELD);

                ResourceIDImpl childResourceID = new ResourceIDImpl();
                childResourceID.setPath(childPath);
                childResourceID.setCollection(true);
                childResourceID.setName(null);
                childResourceID.setPathID(childPathId);
                childPathList.add(childResourceID);
            }
            if (results != null) {
                results.close();
                results = null;
            }
            if (ps != null) {
                ps.close();
                ps = null;
            }

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceID.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            if (results != null) {
                try {
                    results.close();
                } catch (SQLException e) {
                    log.error(RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR, e);
                }
            }
            if (ps != null) {
                try {
                    ps.close();
                } catch (SQLException e) {
                    log.error(RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR, e);
                }
            }
        }
        return childPathList;
    }

    /**
     * Get the path from the path id.
     *
     * @param pathId the path id.
     *
     * @return the path corresponding to the path id.
     * @throws RegistryException throws if operation failed.
     */
    public String getPathFromId(int pathId) throws RegistryException {

        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        ResultSet result = null;
        try {
            String sql =
                    "SELECT REG_PATH_VALUE FROM REG_PATH WHERE REG_PATH_ID=? AND REG_TENANT_ID=?";

            ps = conn.prepareStatement(sql);
            ps.setInt(1, pathId);
            ps.setInt(2, CurrentSession.getTenantId());

            result = ps.executeQuery();
            if (result.next()) {
                return result.getString(DatabaseConstants.PATH_VALUE_FIELD);
            }

        } catch (SQLException e) {

            String msg = "Failed to get the path for the path id " +
                    pathId + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (result != null) {
                    result.close();
                }
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
        return null;
    }

    /**
     * Get the path provided the resource version.
     *
     * @param version the version of the resource.
     *
     * @return the path of the resource.
     * @throws RegistryException throws if the operation failed.
     */
    public String getPath(long version) throws RegistryException {
        ResourceDO resourceDO = getResourceDO(version);
        if (resourceDO == null) {
            return null;
        }
        return getPath(resourceDO.getPathID(), resourceDO.getName(), false);
    }

    /**
     * Get the path given the path id, resource name and provided whether the resourceExistence
     * should be checked.
     *
     * @param pathId         the path id
     * @param resourceName   the resource name
     * @param checkExistence boolean to indicate whether the resource existence should be checked.
     *
     * @return the path
     * @throws RegistryException if the operation failed.
     */
    public String getPath(int pathId, String resourceName, boolean checkExistence)
            throws RegistryException {
        String pathCollection = getPathFromId(pathId);
        if (pathCollection == null) {
            return null;
        }
        String correctPath;
        if (resourceName == null) {
            // this is the parent path..
            correctPath = pathCollection;
        } else if (pathCollection.equals(RegistryConstants.ROOT_PATH)) {
            correctPath = pathCollection + resourceName;
        } else {
            correctPath = pathCollection + RegistryConstants.PATH_SEPARATOR + resourceName;
        }

        // need to check the existence of the path, otherwise return null,
        if (checkExistence) {
            ResourceIDImpl resourceID = new ResourceIDImpl();
            resourceID.setCollection(resourceName == null);
            resourceID.setName(resourceName);
            resourceID.setPath(correctPath);
            resourceID.setPathID(pathId);
            ResourceDO resourceDO = getResourceDO(resourceID);
            if (resourceDO == null) {
                return null;
            }
        }
        return correctPath;
    }

    /**
     * Move resource provided the source and target resource ids.
     *
     * @param source the resource Id of the source.
     * @param target the resource id of the target.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void moveResources(ResourceIDImpl source, ResourceIDImpl target)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (source.isCollection()) {
                String sql = "UPDATE REG_RESOURCE SET REG_PATH_ID=? WHERE " +
                        "REG_PATH_ID=? AND REG_NAME IS NULL AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setInt(2, source.getPathID());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.executeUpdate();
            } else {
                String sql = "UPDATE REG_RESOURCE SET REG_PATH_ID=?, REG_NAME =? WHERE " +
                        "REG_PATH_ID=? AND REG_NAME=? AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setString(2, target.getName());
                ps.setInt(3, source.getPathID());
                ps.setString(4, source.getName());
                ps.setInt(5, CurrentSession.getTenantId());
                ps.executeUpdate();
            }
        } catch (SQLException e) {
            String msg = "Failed to move resource from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }


    /**
     * This method will move the paths from one path id to another regardless of the resource name
     *
     * @param source the path id of the source resource.
     * @param target the path id of the target resource.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void moveResourcePaths(ResourceIDImpl source, ResourceIDImpl target)
            throws RegistryException {
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql =
                    "UPDATE REG_RESOURCE SET REG_PATH_ID=? WHERE REG_PATH_ID=? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, target.getPathID());
            ps.setInt(2, source.getPathID());
            ps.setInt(3, CurrentSession.getTenantId());
            ps.executeUpdate();
        } catch (SQLException e) {
            String msg = "Failed to move resource paths from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * Move the properties.
     *
     * @param source the resource id of the source resource.
     * @param target the resource id of the target resource.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void moveProperties(ResourceIDImpl source, ResourceIDImpl target)
            throws RegistryException {
        if (StaticConfiguration.isVersioningProperties()) {
            return;
        }
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            if (source.isCollection()) {
                String sql = "UPDATE REG_RESOURCE_PROPERTY SET REG_PATH_ID=? WHERE " +
                        "REG_PATH_ID=? AND REG_RESOURCE_NAME IS NULL AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setInt(2, source.getPathID());
                ps.setInt(3, CurrentSession.getTenantId());
                ps.executeUpdate();
            } else {
                String sql =
                        "UPDATE REG_RESOURCE_PROPERTY SET REG_PATH_ID=?, REG_RESOURCE_NAME=? " +
                                "WHERE REG_PATH_ID=? AND REG_RESOURCE_NAME=? AND REG_TENANT_ID=?";
                ps = conn.prepareStatement(sql);
                ps.setInt(1, target.getPathID());
                ps.setString(2, target.getName());
                ps.setInt(3, source.getPathID());
                ps.setString(4, source.getName());
                ps.setInt(5, CurrentSession.getTenantId());
                ps.executeUpdate();
            }
        } catch (SQLException e) {
            String msg = "Failed to move properties from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }

    /**
     * this function will move the paths from one path id to another regardless of the resource
     * name
     *
     * @param source the resource id of the source resource.
     * @param target the resource id of the target resource.
     *
     * @throws RegistryException throws if the operation failed.
     */
    public void movePropertyPaths(ResourceIDImpl source, ResourceIDImpl target)
            throws RegistryException {
        if (StaticConfiguration.isVersioningProperties()) {
            return;
        }
        Connection conn = Transaction.getConnection();
        PreparedStatement ps = null;
        try {
            String sql = "UPDATE REG_RESOURCE_PROPERTY SET REG_PATH_ID=? WHERE " +
                    "REG_PATH_ID=? AND REG_TENANT_ID=?";
            ps = conn.prepareStatement(sql);
            ps.setInt(1, target.getPathID());
            ps.setInt(2, source.getPathID());
            ps.setInt(3, CurrentSession.getTenantId());
            ps.executeUpdate();
        } catch (SQLException e) {
            String msg = "Failed to move property paths from  " + source.getPath() +
                    " to " + target.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } finally {
            try {
                if (ps != null) {
                    ps.close();
                }
            }
            catch (SQLException ex) {
                String msg = RegistryConstants.RESULT_SET_PREPARED_STATEMENT_CLOSE_ERROR;
                log.error(msg, ex);
            }
        }
    }
}
