/*
 *  Copyright (c) 2005-2009, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */
package org.wso2.carbon.registry.core.jdbc.handlers;

import org.apache.axiom.om.OMElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.handlers.filters.Filter;

import java.util.Dictionary;
import java.util.Hashtable;

/**
 * This class is used to manage handlers belonging to a particular lifecycle phase. There are three
 * defined phases, <b>default</b>, <b>reporting</b>, and <b>user</b>.
 */
public class HandlerLifecycleManager extends HandlerManager {

    private Dictionary<String, HandlerManager> handlerManagers;

    /**
     * This phase contains the default system-level handlers.
     */
    public final static String DEFAULT_SYSTEM_HANDLER_PHASE = "default";

    /**
     * This phase contains all the reporting handlers (such as WS-Eventing & Notification
     * handlers).
     */
    public final static String DEFAULT_REPORTING_HANDLER_PHASE = "reporting";

    /**
     * This phase contains user-defined handlers.
     */
    public final static String USER_DEFINED_HANDLER_PHASE = "user";

    private static final Log log = LogFactory.getLog(HandlerLifecycleManager.class);

    /**
     * Creates a Handler Manager for a given handler lifecycle phase.
     */
    public HandlerLifecycleManager() {
        handlerManagers = new Hashtable<String, HandlerManager>();
        HandlerManager defaultHandlerManager = new HandlerManager();
        defaultHandlerManager.setEvaluateAllHandlers(false);
        handlerManagers.put(DEFAULT_SYSTEM_HANDLER_PHASE, defaultHandlerManager);
        HandlerManager userDefinedHandlerManager = new HandlerManager();
        userDefinedHandlerManager.setEvaluateAllHandlers(false);
        handlerManagers.put(USER_DEFINED_HANDLER_PHASE, userDefinedHandlerManager);
        HandlerManager reportingHandlerManager = new UserDefinedHandlerManager();
        reportingHandlerManager.setEvaluateAllHandlers(true);
        handlerManagers.put(DEFAULT_REPORTING_HANDLER_PHASE, reportingHandlerManager);
    }

    @Override
    public void addHandler(String[] methods, Filter filter, Handler handler) {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).addHandler(methods, filter, handler);
    }

    @Override
    public void addHandlerWithPriority(String[] methods, Filter filter, Handler handler) {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).addHandlerWithPriority(
                methods, filter, handler);
    }

    @Override
    public void removeHandler(Handler handler) {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).removeHandler(handler);
    }

    @Override
    public void removeHandler(String[] methods, Filter filter, Handler handler) {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).removeHandler(methods, filter, handler);
    }

    @Override
    public void addHandler(String[] methods, Filter filter, Handler handler,
                           String lifecyclePhase) {
        if (lifecyclePhase == null) {
            addHandler(methods, filter, handler);
            return;
        }
        HandlerManager hm = handlerManagers.get(lifecyclePhase);
        if (hm == null) {
            log.warn("Invalid handler lifecycle phase: " + lifecyclePhase +
                    ". Adding handler to the default phase.");
            addHandler(methods, filter, handler);
        } else {
            hm.addHandler(methods, filter, handler);
        }
    }

    @Override
    public void addHandlerWithPriority(String[] methods, Filter filter, Handler handler,
                                       String lifecyclePhase) {
        if (lifecyclePhase == null) {
            addHandlerWithPriority(methods, filter, handler);
            return;
        }
        HandlerManager hm = handlerManagers.get(lifecyclePhase);
        if (hm == null) {
            log.warn("Invalid handler lifecycle phase: " + lifecyclePhase +
                    ". Adding handler to the default phase.");
            addHandlerWithPriority(methods, filter, handler);
        } else {
            hm.addHandlerWithPriority(methods, filter, handler, lifecyclePhase);
        }
    }

    @Override
    public void removeHandler(Handler handler, String lifecyclePhase) {
        if (lifecyclePhase == null) {
            removeHandler(handler);
            return;
        }
        HandlerManager hm = handlerManagers.get(lifecyclePhase);
        if (hm == null) {
            log.warn("Invalid handler lifecycle phase: " + lifecyclePhase +
                    ". Removing handler from the default phase.");
            removeHandler(handler);
        } else {
            hm.removeHandler(handler, lifecyclePhase);
        }
    }

    @Override
    public void removeHandler(String[] methods, Filter filter, Handler handler,
                              String lifecyclePhase) {
        if (lifecyclePhase == null) {
            removeHandler(methods, filter, handler);
            return;
        }
        HandlerManager hm = handlerManagers.get(lifecyclePhase);
        if (hm == null) {
            log.warn("Invalid handler lifecycle phase: " + lifecyclePhase +
                    ". Removing handler from the default phase.");
            removeHandler(methods, filter, handler);
        } else {
            hm.removeHandler(methods, filter, handler, lifecyclePhase);
        }
    }

    @Override
    public void editComment(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).editComment(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).editComment(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).editComment(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void removeComment(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).removeComment(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).removeComment(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).removeComment(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void createVersion(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).createVersion(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).createVersion(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).createVersion(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void restoreVersion(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).restoreVersion(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).restoreVersion(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).restoreVersion(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void rateResource(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).rateResource(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).rateResource(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).rateResource(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void removeTag(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).removeTag(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).removeTag(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).removeTag(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void applyTag(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).applyTag(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).applyTag(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).applyTag(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void removeAssociation(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).removeAssociation(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).removeAssociation(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).removeAssociation(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void addAssociation(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).addAssociation(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).addAssociation(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).addAssociation(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void delete(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).delete(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).delete(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).delete(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void putChild(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).putChild(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).putChild(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).putChild(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void importChild(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).importChild(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).importChild(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).importChild(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void invokeAspect(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).invokeAspect(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).invokeAspect(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).invokeAspect(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void createLink(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).createLink(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).createLink(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).createLink(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void removeLink(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).removeLink(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).removeLink(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).removeLink(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public void restore(RequestContext requestContext) throws RegistryException {
        handlerManagers.get(DEFAULT_SYSTEM_HANDLER_PHASE).restore(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(USER_DEFINED_HANDLER_PHASE).restore(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).restore(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
    }

    @Override
    public Association[] getAllAssociations(RequestContext requestContext)
            throws RegistryException {
        Association[] defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getAllAssociations(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        Association[] userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getAllAssociations(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).getAllAssociations(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public Association[] getAssociations(RequestContext requestContext) throws RegistryException {
        Association[] defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getAssociations(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        Association[] userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getAssociations(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).getAssociations(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public TaggedResourcePath[] getResourcePathsWithTag(RequestContext requestContext)
            throws RegistryException {
        TaggedResourcePath[] defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getResourcePathsWithTag(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        TaggedResourcePath[] userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getResourcePathsWithTag(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        handlerManagers.get(
                DEFAULT_REPORTING_HANDLER_PHASE).getResourcePathsWithTag(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public Tag[] getTags(RequestContext requestContext)
            throws RegistryException {
        Tag[] defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getTags(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        Tag[] userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getTags(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).getTags(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public Comment[] getComments(RequestContext requestContext) throws RegistryException {
        Comment[] defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getComments(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        Comment[] userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getComments(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).getComments(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public String[] getVersions(RequestContext requestContext) throws RegistryException {
        String[] defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getVersions(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        String[] userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getVersions(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).getVersions(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public Collection executeQuery(RequestContext requestContext) throws RegistryException {
        Collection defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).executeQuery(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        Collection userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).executeQuery(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).executeQuery(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public Collection searchContent(RequestContext requestContext) throws RegistryException {
        Collection defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).searchContent(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        Collection userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).searchContent(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).searchContent(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public String addComment(RequestContext requestContext) throws RegistryException {
        String defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).addComment(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        String userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).addComment(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).addComment(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public Resource get(RequestContext requestContext) throws RegistryException {
        Resource defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).get(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        Resource userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).get(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).get(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public String put(RequestContext requestContext) throws RegistryException {
        String defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).put(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        String userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).put(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).put(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public String importResource(RequestContext requestContext) throws RegistryException {
        String defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).importResource(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        String userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).importResource(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).importResource(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public String copy(RequestContext requestContext) throws RegistryException {
        String defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).copy(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        String userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).copy(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).copy(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public String move(RequestContext requestContext) throws RegistryException {
        String defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).move(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        String userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).move(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).move(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public String rename(RequestContext requestContext) throws RegistryException {
        String defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).rename(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        String userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).rename(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).rename(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public OMElement dump(RequestContext requestContext) throws RegistryException {
        OMElement defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).dump(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        OMElement userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).dump(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).dump(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != null) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public float getAverageRating(RequestContext requestContext) throws RegistryException {
        float defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getAverageRating(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        float userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getAverageRating(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).getAverageRating(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != -1) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public int getRating(RequestContext requestContext) throws RegistryException {
        int defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).getRating(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        int userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).getRating(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).getRating(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        if (userDefinedValue != -1) {
            return userDefinedValue;
        }
        return defaultValue;
    }

    @Override
    public boolean resourceExists(RequestContext requestContext) throws RegistryException {
        boolean defaultValue = handlerManagers.get(
                DEFAULT_SYSTEM_HANDLER_PHASE).resourceExists(requestContext);
        boolean isProcessingComplete = requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(false);
        boolean userDefinedValue = handlerManagers.get(
                USER_DEFINED_HANDLER_PHASE).resourceExists(requestContext);
        isProcessingComplete |= requestContext.isProcessingComplete();
        // The reporting handler phase needs to know about the state of processing
        requestContext.setProcessingComplete(isProcessingComplete);
        handlerManagers.get(DEFAULT_REPORTING_HANDLER_PHASE).resourceExists(requestContext);
        // The reporting handlers may change the state of processing
        isProcessingComplete |= requestContext.isProcessingComplete();
        requestContext.setProcessingComplete(isProcessingComplete);
        return userDefinedValue || defaultValue;
    }
}