/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.queries;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.config.StaticConfiguration;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.DatabaseConstants;
import org.wso2.carbon.registry.core.jdbc.dao.CommentsDAO;
import org.wso2.carbon.registry.core.jdbc.dao.RatingsDAO;
import org.wso2.carbon.registry.core.jdbc.dao.ResourceDAO;
import org.wso2.carbon.registry.core.jdbc.dao.TagsDAO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.RatingDO;
import org.wso2.carbon.registry.core.jdbc.dataobjects.TaggingDO;
import org.wso2.carbon.registry.core.jdbc.utils.Transaction;
import org.wso2.carbon.registry.core.session.CurrentSession;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * The query processor to execute sql queries.
 */
public class SQLQueryProcessor implements QueryProcessor {

    private static final Log log = LogFactory.getLog(SQLQueryProcessor.class);
    private ResourceDAO resourceDAO;
    private CommentsDAO commentsDAO;
    private RatingsDAO ratingsDAO;
    private TagsDAO tagsDAO;

    /**
     * DataSource of the registry database. URL handlers can access this to construct resources by
     * combining various tables (e.g. comments).
     */
    protected DataSource dataSource;

    /**
     * Initialize the sql query processor
     *
     * @param dataSource the dataSource to be set.
     */
    public SQLQueryProcessor(DataSource dataSource) {
        this.dataSource = dataSource;
        resourceDAO = new ResourceDAO();
        commentsDAO = StaticConfiguration.getCommentsDAO();
        ratingsDAO = StaticConfiguration.getRatingsDAO();
        tagsDAO = StaticConfiguration.getTagsDAO();
    }

    public Collection executeQuery(Registry registry, Resource query, Map parameters)
            throws RegistryException {

        Collection resultCollection = null;

        Connection conn;
        String sqlString;
        ResultSet results = null;
        PreparedStatement s = null;

        try {

            Object obj = query.getContent();
            if (parameters != null) {
                Object querySQL = parameters.get("query");
                if (querySQL != null) {
                    parameters.remove("query");
                    obj = querySQL;
                }
                Object resultType = parameters.get(RegistryConstants.RESULT_TYPE_PROPERTY_NAME);
                if (resultType != null) {
                    parameters.remove(RegistryConstants.RESULT_TYPE_PROPERTY_NAME);
                    if (resultType instanceof String) {
                        query.setProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME,
                                (String) resultType);
                    }
                }
            }
            if (obj instanceof String) {
                sqlString = (String) obj;
            } else if (obj instanceof byte[]) {
                sqlString = new String((byte[]) obj);
            } else {
                throw new RegistryException("Unable to execute query at " + query.getPath()
                        + ".Found resource content of type " +
                        (obj == null ? "null" : obj.getClass().getName())
                        + ".Expected java.lang.String or byte[]");
            }

            conn = Transaction.getConnection();

            // adding the tenant ids for the query
            TenantAwareSQLTransformer transformer = new TenantAwareSQLTransformer(sqlString);
            String transformedQuery = transformer.getTransformedQuery();
            int transformedParameterCount = transformer.getAdditionalParameterCount();
            int trailingParameterCount = transformer.getTrailingParameterCount();

            s = conn.prepareStatement(transformedQuery, ResultSet.TYPE_FORWARD_ONLY,
                    ResultSet.CONCUR_READ_ONLY);

            int nextParameterIndex = 1;
            if (parameters != null) {
                parameters.remove("content");
                for (Object parameterNumberObject : parameters.keySet()) {
                    String parameterNumber = (String) parameterNumberObject;
                    Object parameterValue = parameters.get(parameterNumber);
                    s.setObject(Integer.parseInt(parameterNumber), parameterValue);
                    nextParameterIndex++;
                }
                if (trailingParameterCount > 0 && transformedParameterCount > 0) {
                    // Move trailing parameters to the end.
                    for (int i = 0; i < trailingParameterCount; i++) {
                        nextParameterIndex--;
                        s.setObject(nextParameterIndex + transformedParameterCount,
                                parameters.get(Integer.toString(nextParameterIndex)));
                    }
                }
            }
            // adding the additional parameters caused due to adding the tenant id,
            for (int i = 0; i < transformedParameterCount; i++) {
                s.setInt(nextParameterIndex, CurrentSession.getTenantId());
                nextParameterIndex++;
            }

            results = s.executeQuery();

            String resultType = query.getProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME);

            if (resultType == null) {
                resultType = RegistryConstants.RESOURCES_RESULT_TYPE;
                query.setProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME, resultType);
            }

            if (resultType.equals(RegistryConstants.RESOURCES_RESULT_TYPE)) {

                // Result is a normal resource, which is stored in the Resources table or a
                // collection of normal resources.

                resultCollection = fillResourcesCollection(results);

            } else if (resultType.equals(RegistryConstants.COMMENTS_RESULT_TYPE)) {

                resultCollection = fillCommentsCollection(results, conn);

            } else if (resultType.equals(RegistryConstants.RATINGS_RESULT_TYPE)) {

                resultCollection = fillRatingsCollection(results);

            } else if (resultType.equals(RegistryConstants.TAGS_RESULT_TYPE)) {

                resultCollection = fillTagsCollection(results);

            } else if (resultType.equals("TagsWithCount")) {

                // Result is in the form of (tag, number of taggings) tuples
                // Fill TagCount objects with results and add them as an array


            } else if (resultType.equals("ResourcePathsWithTagCount")) {

                // Result is in the form of (resource path, number of taggings) tuples
                // Fill ResourceTagCount objects with results and add them as an array

            }

            if (resultCollection == null) {
                String msg = "Unknown result type: " + resultType + " defined for the query: " +
                        sqlString + ((query.getPath() != null) ? " located in path: " +
                        query.getPath() : "");
                log.error(msg);
                throw new RegistryException(msg);
            }

        } catch (SQLException e) {
            throw new RegistryException(e.getMessage());

        } finally {
            if (results != null) {
                try {
                    results.close();
                } catch (SQLException e) {
                    String msg = "Failed to close the result set. " + e.getMessage();
                    log.error(msg, e);
                }
            }

            if (s != null) {
                try {
                    s.close();
                } catch (SQLException e) {
                    log.error("Failed to close the statement. " + e.getMessage());
                }
            }
        }

        if (resultCollection != null && RegistryConstants.RESOURCES_RESULT_TYPE.equals(
                query.getProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME))) {

            List<String> filteredResults = new ArrayList<String>();
            String[] resultPaths = resultCollection.getChildren();
            for (String resultPath : resultPaths) {
                if (AuthorizationUtils.authorize(resultPath, ActionConstants.GET)) {
                    filteredResults.add(resultPath);
                }
            }

            String[] filteredContent = filteredResults.toArray(new String[filteredResults.size()]);
            resultCollection.setContent(filteredContent);
        }

        return resultCollection;
    }

    /**
     * Fil the resource collection from a query result set with resources
     *
     * @param results The result set object.
     *
     * @return A collection containing results as children.
     * @throws SQLException      throws if the iterating results failed.
     * @throws RegistryException throws if constructing child pas failed.
     */
    private Collection fillResourcesCollection(ResultSet results)
            throws SQLException, RegistryException {

        // Result is a normal resource, which is stored in the Resources table or a collection
        // of normal resources.

        List<String> pathList = new ArrayList<String>();
        while (results.next()) {
            int pathId = results.getInt(DatabaseConstants.PATH_ID_FIELD);
            String resourceName = results.getString(DatabaseConstants.NAME_FIELD);
            String path = resourceDAO.getPath(pathId, resourceName, true);
            if (path != null && (!pathList.contains(path))) {
                pathList.add(path);
            }
        }

        String[] paths = pathList.toArray(new String[pathList.size()]);
        return new CollectionImpl(paths);
    }


    /**
     * Fil the resource collection from a query result set with comments
     *
     * @param results The result set object.
     * @param conn    connection object.
     *
     * @return A collection containing results as children.
     * @throws SQLException      throws if the iterating results failed.
     * @throws RegistryException throws if constructing child pas failed.
     */
    private Collection fillCommentsCollection(ResultSet results, Connection conn)
            throws SQLException, RegistryException {

        // SQL query should return a list of DatabaseConstants.COMMENT_ID_FIELDs.
        // resultArtifact contains a String[] of URLs for comments.

        // URL for a comment <resource_path>?comment<comment_id>
        // e.g. /p1/r1?comment12

        List<Long> commentIDs = new ArrayList<Long>();

        while (results.next()) {
            long commentID = results.getLong(DatabaseConstants.COMMENT_ID_FIELD);
            commentIDs.add(commentID);
        }

        String[] commentPaths = commentsDAO.getResourcePathsOfComments(
                commentIDs.toArray(new Long[commentIDs.size()]), conn);

        return new CollectionImpl(commentPaths);
    }

    /**
     * Fil the resource collection from a query result set with ratings
     *
     * @param results The result set object.
     *
     * @return A collection containing results as children.
     * @throws SQLException      throws if the iterating results failed.
     * @throws RegistryException throws if constructing child pas failed.
     */
    private Collection fillRatingsCollection(ResultSet results)
            throws SQLException, RegistryException {

        // SQL query should return a list of DatabaseConstants.RATING_ID_FIELDs
        // resultArtifact contains a String[] of URLs for ratings.

        // URL for a comment <resource_path>;ratings:<userName>
        // e.g. /p1/r1;ratings:ruwan

        List<String> ratingPathList = new ArrayList<String>();

        while (results.next()) {

            long ratingID = results.getLong(DatabaseConstants.RATING_ID_FIELD);

            RatingDO ratingDAO = ratingsDAO.getRating(ratingID);

            String ratingPath =
                    ratingDAO.getResourcePath() + RegistryConstants.URL_SEPARATOR + "ratings:" +
                            ratingDAO.getRatedUserName();
            ratingPathList.add(ratingPath);
        }

        String[] ratingPaths =
                ratingPathList.toArray(new String[ratingPathList.size()]);

        return new CollectionImpl(ratingPaths);
    }

    /**
     * Fil the resource collection from a query result set with tags
     *
     * @param results The result set object.
     *
     * @return A collection containing results as children.
     * @throws SQLException      throws if the iterating results failed.
     * @throws RegistryException throws if constructing child pas failed.
     */
    private Collection fillTagsCollection(ResultSet results)
            throws SQLException, RegistryException {

        // URL for a tag /p1/r1;tags:tagName:userName

        List<String> tagPathList = new ArrayList<String>();
        while (results.next()) {

            long taggingID = results.getLong(DatabaseConstants.TAGGING_ID_FIELD);

            TaggingDO taggingDO = tagsDAO.getTagging(taggingID);

            String tagPath = taggingDO.getResourcePath() + RegistryConstants.URL_SEPARATOR +
                    "tags:" + taggingDO.getTagName() + ":" + taggingDO.getTaggedUserName();
            tagPathList.add(tagPath);
        }

        String[] tagPaths = tagPathList.toArray(new String[tagPathList.size()]);

        return new CollectionImpl(tagPaths);
    }
}