/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.utils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.sql.*;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Stack;

/**
 * This is the base class that manages the transactional database operations of the Registry. The
 * transactions API of the Registry is capable of handling nested transactions. If a nested
 * transaction is rollbacked, the whole outer transaction would also be rollbacked, so that the
 * database will not be in an inconsistent state.
 * <p/>
 * The transaction API also supports transactions across multiple databases. This allows various
 * portions of the Registry to be mounted from various databases and work as a part of a single
 * tree.
 * <p/>
 * The transaction is session scoped, and therefore, is available only on a single running thread.
 * Multiple threads would have its own transaction and thereby make it possible to manage
 * concurrency at a much lower level.
 */
public class Transaction {

    private static ThreadLocal<TransactionEntry> tCurrent =
            new ThreadLocal<TransactionEntry>() {
                protected TransactionEntry initialValue() {
                    return new TransactionEntry();
                }
            };

    private static ThreadLocal<Stack<TransactionEntry>> tStackedTransactionalEntryStack =
            new ThreadLocal<Stack<TransactionEntry>>() {
                protected Stack<TransactionEntry> initialValue() {
                    Stack<TransactionEntry> StackedTransactionalEntryStack =
                            new Stack<TransactionEntry>();
                    StackedTransactionalEntryStack.push(null);
                    return StackedTransactionalEntryStack;
                }
            };

    private static Log log = LogFactory.getLog(Transaction.class);

    // Method to obtain stacked transactional connection.
    private static TransactionEntry getStackedTransactionalEntry() {
        Stack<TransactionEntry> transactionalEntryStack =
                tStackedTransactionalEntryStack.get();
        if (transactionalEntryStack == null) {
            tStackedTransactionalEntryStack.remove();
            transactionalEntryStack = tStackedTransactionalEntryStack.get();
        }
        return transactionalEntryStack.peek();
    }

    // Method to set stacked transactional connection.
    private static void setStackedTransactionalEntry(
            TransactionEntry StackedTransactionalEntry) {
        Stack<TransactionEntry> transactionalEntryStack =
                tStackedTransactionalEntryStack.get();
        if (transactionalEntryStack == null) {
            tStackedTransactionalEntryStack.remove();
            transactionalEntryStack = tStackedTransactionalEntryStack.get();
        }
        transactionalEntryStack.push(StackedTransactionalEntry);
    }

    // Method to remove stacked transactional connection.
    private static void removeStackedTransactionalEntry() {
        Stack<TransactionEntry> transactionalEntryStack =
                tStackedTransactionalEntryStack.get();
        if (transactionalEntryStack != null) {
            transactionalEntryStack.pop();
        }
    }

    /**
     * Push current transaction to a stack and make room for a new one.
     */
    public static void pushTransaction() {
        // This method will stack the current transaction by creating a stacked transactional
        // connection object and then begin the new transaction.
        log.trace("pushing current transaction to stack");
        setStackedTransactionalEntry(tCurrent.get());
        tCurrent.set(new TransactionEntry());
    }

    /**
     * Pop stacked transaction and replace the existing one with that.
     */
    public static void popTransaction() {
        log.trace("popping current transaction from stack");
        TransactionEntry transactionEntry =
                getStackedTransactionalEntry();
        if (transactionEntry != null) {
            removeStackedTransactionalEntry();
            tCurrent.set(transactionEntry);
        }
    }

    /**
     * Method to determine whether a transaction is started or not.
     *
     * @return whether a transaction is started or not.
     */
    public static boolean isStarted() {
        if (tCurrent.get() != null) {
            return tCurrent.get().isStarted();
        } else {
            log.error("The current transaction entry has not been created.");
            return false;
        }
    }

    /**
     * Method to set whether a transaction is started or not.
     *
     * @param started whether a transaction is started or not.
     */
    public static void setStarted(boolean started) {
        if (tCurrent.get() != null) {
            tCurrent.get().setStarted(started);
        } else {
            log.error("The current transaction entry has not been created.");
        }
    }

    /**
     * Method to get connection.
     *
     * @return the connection.
     */
    public static Connection getConnection() {
        if (tCurrent.get() != null) {
            return tCurrent.get().getConnection();
        } else {
            log.error("The current transaction entry has not been created.");
            return null;
        }
    }

    /**
     * Method to set connection.
     *
     * @param connection the connection.
     */
    public static void setConnection(Connection connection) {
        if (tCurrent.get() != null) {
            if (connection != null) {
                tCurrent.get().setStarted(true);
                tCurrent.get().setConnection(new ManagedRegistryConnection(connection));
            } else {
                tCurrent.get().setConnection(null);
            }
        } else {
            log.error("The current transaction entry has not been created.");
        }
    }

    /**
     * Method to remove the connection from the transaction.
     */
    @SuppressWarnings("unused")
    public static void removeConnection() {
        if (tCurrent.get() != null) {
            tCurrent.get().setStarted(false);
            tCurrent.get().setConnection(null);
        } else {
            log.error("The current transaction entry has not been created.");
        }
    }

    /**
     * This method will increment the nested depth of the transaction on the current session.
     */
    public static void incNestedDepth() {
        if (tCurrent.get() != null) {
            int transactionDepth = tCurrent.get().getNestedDepth();
            if (transactionDepth == 0) {
                tCurrent.get().setStarted(true);
                tCurrent.get().setRollbacked(false);
            }
            transactionDepth++;
            tCurrent.get().setNestedDepth(transactionDepth);
        } else {
            log.error("The current transaction entry has not been created.");
        }
    }

    /**
     * This method will decrement the nested depth of the transaction on the current session.
     */
    public static void decNestedDepth() {
        if (tCurrent.get() != null) {
            int transactionDepth = tCurrent.get().getNestedDepth();
            transactionDepth--;
            tCurrent.get().setNestedDepth(transactionDepth);
            if (transactionDepth == 0) {
                tCurrent.get().setStarted(false);
            }
        } else {
            log.error("The current transaction entry has not been created.");
        }
    }

    /**
     * Method to obtain the depth of nesting of this transaction at a given point in time.
     *
     * @return the nested depth.
     */
    public static int getNestedDepth() {
        if (tCurrent.get() != null) {
            return tCurrent.get().getNestedDepth();
        } else {
            log.error("The current transaction entry has not been created.");
            return 0;
        }
    }

    /**
     * Method to determine whether this transaction has been rollbacked at least once. This applies
     * to the out as well as inner transactions as a whole.
     *
     * @return whether this transaction has been rollbacked.
     */
    public static boolean isRollbacked() {
        if (tCurrent.get() != null) {
            return tCurrent.get().isRollbacked();
        } else {
            log.error("The current transaction entry has not been created.");
            return false;
        }
    }

    /**
     * Method to set whether this transaction has been rollbacked or not.
     *
     * @param rollbacked whether this transaction has been rollbacked or not.
     */
    public static void setRollbacked(boolean rollbacked) {
        if (tCurrent.get() != null) {
            tCurrent.get().setRollbacked(rollbacked);
        } else {
            log.error("The current transaction entry has not been created.");
        }
    }

    /**
     * Method to obtain a connection that is managed by the registry transactions implementation.
     * If the managed transaction is already closed, committed or rollbacked, this method will
     * reinstate it.
     *
     * @param conn un-managed connection.
     *
     * @return managed connection.
     */
    public static Connection getManagedRegistryConnection(Connection conn) {
        return Transaction.ManagedRegistryConnection.getManagedRegistryConnection(conn, true);
    }

    /**
     * A structure that can stack a transactional connection, which is handy when switching
     * connections.
     */
    private static class TransactionEntry {

        private boolean started;
        private Connection connection;
        private int nestedDepth;
        private boolean rollbacked;


        public boolean isStarted() {
            return started;
        }

        public void setStarted(boolean started) {
            this.started = started;
        }

        public Connection getConnection() {
            return connection;
        }

        public void setConnection(Connection connection) {
            this.connection = connection;
        }

        public int getNestedDepth() {
            return nestedDepth;
        }

        public void setNestedDepth(int nestedDepth) {
            this.nestedDepth = nestedDepth;
        }

        public boolean isRollbacked() {
            return rollbacked;
        }

        public void setRollbacked(boolean rollbacked) {
            this.rollbacked = rollbacked;
        }
    }

    /**
     * Wrapped connection that manages connections making it possible to have connections to
     * multiple databases.
     */
    @SuppressWarnings("unused")
    private static class ManagedRegistryConnection implements Connection {
        // This holds the un-managed connection.
        private Connection connection;

        // This contains a list of all the managed connections.
        private static ThreadLocal<Map<String, ManagedRegistryConnection>>
                tManagedConnectionMap =
                new ThreadLocal<Map<String, ManagedRegistryConnection>>() {
                    protected Map<String, ManagedRegistryConnection> initialValue() {
                        return new LinkedHashMap<String, ManagedRegistryConnection>();
                    }
                };

        // This contains a list of committed or rollbacked connections.
        private static ThreadLocal<Map<String, ManagedRegistryConnection>>
                tCommittedAndRollbackedConnectionMap =
                new ThreadLocal<Map<String, ManagedRegistryConnection>>() {
                    protected Map<String, ManagedRegistryConnection> initialValue() {
                        return new LinkedHashMap<String, ManagedRegistryConnection>();
                    }
                };

        // This contains a list of connections that have been closed.
        private static ThreadLocal<Map<String, ManagedRegistryConnection>>
                tClosedConnectionMap =
                new ThreadLocal<Map<String, ManagedRegistryConnection>>() {
                    protected Map<String, ManagedRegistryConnection> initialValue() {
                        return new LinkedHashMap<String, ManagedRegistryConnection>();
                    }
                };

        // This identifies whether at least one nested transactional connection was rollbacked.
        private static ThreadLocal<Boolean> tRollbackedConnection =
                new ThreadLocal<Boolean>() {
                    protected Boolean initialValue() {
                        return false;
                    }
                };

        public ManagedRegistryConnection(Connection connection) {
            // If the connection is already managed, obtain the original connection.
            if (connection instanceof ManagedRegistryConnection) {
                this.connection = ((ManagedRegistryConnection) connection).getConnection();
            } else {
                this.connection = connection;
            }
            log.trace("Saving managed registry connection to map.");
            tManagedConnectionMap.get().put(getConnectionId(), this);

            // We start from scratch here, so remove any closed or committed connections.
            if (tClosedConnectionMap.get().get(getConnectionId()) != null) {
                tClosedConnectionMap.get().put(getConnectionId(), null);
            }
            if (tCommittedAndRollbackedConnectionMap.get().get(getConnectionId()) != null) {
                tCommittedAndRollbackedConnectionMap.get().put(getConnectionId(), null);
            }
        }

        private String getConnectionId() {
            return getConnectionId(this.connection);
        }

        private static String getConnectionId(Connection connection) {
            try {
                // The connection URL is unique enough to be used as an identifier since one thread
                // makes one connection to the given URL according to our model.
                if (connection.getMetaData() != null) {
                    return connection.getMetaData().getURL();
                }
            } catch (SQLException ignore) {
            }
            return null;
        }

        /**
         * If a managed connection for the corresponding connection is available, get that.
         *
         * @param connection the original connection
         * @param reinstate whether to instate the connection or not.
         *
         * @return the managed connection or null
         */
        public static Connection getManagedRegistryConnection(Connection connection,
                                                              boolean reinstate) {
            if (tManagedConnectionMap != null && tManagedConnectionMap.get() != null) {
                Connection mrc = tManagedConnectionMap.get().get(getConnectionId(connection));
                if (mrc != null && reinstate) {
                    if (tClosedConnectionMap != null && tClosedConnectionMap.get() != null &&
                            tClosedConnectionMap.get().get(getConnectionId(connection)) != null) {
                        tClosedConnectionMap.get().put(getConnectionId(connection), null);
                    }
                    if (tCommittedAndRollbackedConnectionMap != null &&
                            tCommittedAndRollbackedConnectionMap.get() != null &&
                            tCommittedAndRollbackedConnectionMap.get().get(
                                    getConnectionId(connection)) != null) {
                        tCommittedAndRollbackedConnectionMap.get().put(
                                getConnectionId(connection), null);
                    }
                }
                return mrc;
            }
            return null;
        }

        // This utility method returns the number of connections stored in a given map.
        private int getConnectionCount(Map<String, ManagedRegistryConnection> connections) {
            int count = 0;
            for (String k : connections.keySet()) {
                if (connections.get(k) != null) {
                    count++;
                }
            }
            return count;
        }

        public void commit() throws SQLException {
            if (tRollbackedConnection.get()) {
                // If at least one connection was rollbacked, do the same for all.
                log.trace("Rolling back the transaction(s).");
                rollback();
            } else {
                if (tManagedConnectionMap.get().size() == 1) {
                    // If there is only one connection, simply commit that.
                    connection.commit();
                    log.trace("Committed all transactions.");
                } else if ((tCommittedAndRollbackedConnectionMap.get().get(getConnectionId())
                        == null) && (tManagedConnectionMap.get().size() ==
                        getConnectionCount(tCommittedAndRollbackedConnectionMap.get()) + 1)) {
                    // If this is the outer connection, then commit all inner connections and then
                    // the outer connection.
                    Map<String, ManagedRegistryConnection> connections =
                            tCommittedAndRollbackedConnectionMap.get();
                    for (String k : connections.keySet()) {
                        if (connections.get(k) != null) {
                            connections.get(k).getConnection().commit();
                            connections.put(k, null);
                        }
                    }
                    // Clean up list of committed and rollbacked connections.
                    tCommittedAndRollbackedConnectionMap.set(new LinkedHashMap<String,
                            ManagedRegistryConnection>());
                    connection.commit();
                    log.trace("Committed all transactions.");
                } else if (tManagedConnectionMap.get().size() <
                        getConnectionCount(tCommittedAndRollbackedConnectionMap.get())) {
                    throw new SQLException("Total number of available connections are less than " +
                            "the total number of committed connections");
                } else {
                    // If this is an inner connection, keep this for later.
                    tCommittedAndRollbackedConnectionMap.get().put(getConnectionId(), this);
                }
            }
        }

        public void rollback() throws SQLException {
            // Calling this method once will set the flag indicating at least one connection was
            // rollbacked.
            tRollbackedConnection.set(true);
            if (tManagedConnectionMap.get().size() == 1) {
                // If there is only one connection, simply rollback that.
                connection.rollback();
                // Clear the flag after rollback
                tRollbackedConnection.set(false);
                log.trace("Rolled back all transactions.");
            } else if ((tCommittedAndRollbackedConnectionMap.get().get(getConnectionId())
                        == null) && (tManagedConnectionMap.get().size() ==
                        getConnectionCount(tCommittedAndRollbackedConnectionMap.get()) + 1)) {
                // If this is the outer connection, then rollback all inner connections and then
                // the outer connection.
                Map<String, ManagedRegistryConnection> connections =
                        tCommittedAndRollbackedConnectionMap.get();
                for (String k : connections.keySet()) {
                    if (connections.get(k) != null) {
                        connections.get(k).getConnection().rollback();
                    }
                }
                // Clean up list of committed and rollbacked connections.
                tCommittedAndRollbackedConnectionMap.set(new LinkedHashMap<String,
                        ManagedRegistryConnection>());
                connection.rollback();
                // Clear the flag after rollback
                tRollbackedConnection.set(false);
                log.trace("Rolled back all transactions.");
            } else if (tManagedConnectionMap.get().size() <
                    getConnectionCount(tCommittedAndRollbackedConnectionMap.get())) {
                throw new SQLException("Total number of available connections are less than the " +
                        "total number of rollbacked or committed connections");
            } else {
                // If this is an inner connection, keep this for later.
                tCommittedAndRollbackedConnectionMap.get().put(getConnectionId(), this);
            }
        }

        public void close() throws SQLException {
            if (tManagedConnectionMap.get().size() == 1) {
                // If there is only one connection, simply close that.
                connection.close();
                // Clean up list of managed connections, since we have closed the outer connection.
                tManagedConnectionMap.set(new LinkedHashMap<String, ManagedRegistryConnection>());
                log.trace("Closed all transactions.");
            } else if ((tClosedConnectionMap.get().get(getConnectionId())
                        == null) && (tManagedConnectionMap.get().size() ==
                        getConnectionCount(tClosedConnectionMap.get()) + 1)) {
                // If this is the outer connection, then close all inner connections and then
                // the outer connection.
                Map<String, ManagedRegistryConnection> connections = tClosedConnectionMap.get();
                for (String k : connections.keySet()) {
                    if (connections.get(k) != null) {
                        connections.get(k).getConnection().close();
                    }
                }
                // Clean up list of closed connections.
                tClosedConnectionMap.set(new LinkedHashMap<String, ManagedRegistryConnection>());
                connection.close();
                // Clean up list of managed connections. Closing a connection and all the nested
                // transactional connections would result in closing all of the managed connections.
                // This would leave us with nothing to manage.
                tManagedConnectionMap.set(new LinkedHashMap<String, ManagedRegistryConnection>());
                log.trace("Closed all transactions.");
            } else if (tManagedConnectionMap.get().size() <
                    getConnectionCount(tClosedConnectionMap.get())) {
                throw new SQLException("Total number of available connections are less than the " +
                        "total number of closed connections");
            } else {
                // If this is an inner connection, keep this for later.
                tClosedConnectionMap.get().put(getConnectionId(), this);
            }
        }

        public boolean isClosed() throws SQLException {
            // If there is only one connection, we don't add it to the closed connection map.
            if (tManagedConnectionMap.get().size() == 1) {
                return connection.isClosed();
            }
            // This connection is closed, if it is in the closed connections map, or if the
            // underlying connection is closed.
            return tClosedConnectionMap.get().get(getConnectionId()) != null ||
                    connection.isClosed();
        }

        public void rollback(Savepoint savepoint) throws SQLException {
            if (tManagedConnectionMap.get().size() == 1) {
                connection.rollback(savepoint);
            } else {
                throw new UnsupportedOperationException("Rollback to Save Point is not supported " +
                        "for operations involving multiple connections.");
            }
        }

        ////////////////////////////////////////////////////////
        // JDK 6 only methods.
        ////////////////////////////////////////////////////////

        public Blob createBlob() throws SQLException {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public Clob createClob() throws SQLException {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public SQLXML createSQLXML() throws SQLException {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public NClob createNClob() throws SQLException {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public Struct createStruct(String typeName, Object[] attributes) throws SQLException {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public Array createArrayOf(String typeName, Object[] elements) throws SQLException {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public Properties getClientInfo() {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public String getClientInfo(String name) {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public boolean isValid(int timeout) {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public void setClientInfo(Properties properties) {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public void setClientInfo(String name, String value) {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public boolean isWrapperFor(Class<?> i) {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        public <T> T unwrap(Class<T> i) {
            // This is a JDK6 specific operation that we don't support for JDK5 compatibility
            throw new UnsupportedOperationException("This method is not supported");
        }

        ////////////////////////////////////////////////////////
        // Other un-managed methods
        ////////////////////////////////////////////////////////

        public Statement createStatement() throws SQLException {
            return connection.createStatement();
        }

        public PreparedStatement prepareStatement(String s) throws SQLException {
            return connection.prepareStatement(s);
        }

        public CallableStatement prepareCall(String s) throws SQLException {
            return connection.prepareCall(s);
        }

        public String nativeSQL(String s) throws SQLException {
            return connection.nativeSQL(s);
        }

        public void setAutoCommit(boolean b) throws SQLException {
            connection.setAutoCommit(b);
        }

        public boolean getAutoCommit() throws SQLException {
            return connection.getAutoCommit();
        }

        public DatabaseMetaData getMetaData() throws SQLException {
            return connection.getMetaData();
        }

        public void setReadOnly(boolean b) throws SQLException {
            connection.setReadOnly(b);
        }

        public boolean isReadOnly() throws SQLException {
            return connection.isReadOnly();
        }

        public void setCatalog(String s) throws SQLException {
            connection.setCatalog(s);
        }

        public String getCatalog() throws SQLException {
            return connection.getCatalog();
        }

        public void setTransactionIsolation(int i) throws SQLException {
            connection.setTransactionIsolation(i);
        }

        public int getTransactionIsolation() throws SQLException {
            return connection.getTransactionIsolation();
        }

        public SQLWarning getWarnings() throws SQLException {
            return connection.getWarnings();
        }

        public void clearWarnings() throws SQLException {
            connection.clearWarnings();
        }

        public Statement createStatement(int i, int i1) throws SQLException {
            return connection.createStatement();
        }

        public PreparedStatement prepareStatement(String s, int i, int i1) throws SQLException {
            return connection.prepareStatement(s);
        }

        public CallableStatement prepareCall(String s, int i, int i1) throws SQLException {
            return connection.prepareCall(s, i, i1);
        }

        public Map<String, Class<?>> getTypeMap() throws SQLException {
            return connection.getTypeMap();
        }

        public void setTypeMap(Map<String, Class<?>> stringClassMap) throws SQLException {
            connection.setTypeMap(stringClassMap);
        }

        public void setHoldability(int i) throws SQLException {
            connection.setHoldability(i);
        }

        public int getHoldability() throws SQLException {
            return connection.getHoldability();
        }

        public Savepoint setSavepoint() throws SQLException {
            return connection.setSavepoint();
        }

        public Savepoint setSavepoint(String s) throws SQLException {
            return connection.setSavepoint(s);
        }

        public void releaseSavepoint(Savepoint savepoint) throws SQLException {
            connection.releaseSavepoint(savepoint);
        }

        public Statement createStatement(int i, int i1, int i2) throws SQLException {
            return connection.createStatement(i, i1, i2);
        }

        public PreparedStatement prepareStatement(String s, int i, int i1, int i2)
                throws SQLException {
            return connection.prepareStatement(s, i, i1, i2);
        }

        public CallableStatement prepareCall(String s, int i, int i1, int i2) throws SQLException {
            return connection.prepareCall(s, i, i1, i2);
        }

        public PreparedStatement prepareStatement(String s, int i) throws SQLException {
            return connection.prepareStatement(s, i);
        }

        public PreparedStatement prepareStatement(String s, int[] integers) throws SQLException {
            return connection.prepareStatement(s, integers);
        }

        public PreparedStatement prepareStatement(String s, String[] strings) throws SQLException {
            return connection.prepareStatement(s, strings);
        }

        public Connection getConnection() {
            return connection;
        }

        public void setConnection(Connection connection) {
            this.connection = connection;
        }
    }
}
