/*
 *  Copyright (c) 2005-2009, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */
package org.wso2.carbon.registry.core.utils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.utils.CarbonUtils;

import java.io.*;
import java.util.Properties;
import java.util.Set;

/**
 * Utility class for managing and fetching media types.
 */
@SuppressWarnings("unused")
// Some utility methods are used outside the kernel, in the registry components.
public class MediaTypesUtils {

    private static final Log log = LogFactory.getLog(MediaTypesUtils.class);

    private static final String RESOURCE_MEDIA_TYPE_MAPPINGS_FILE = "mime.types";

    private static final String MIME_TYPE_COLLECTION = RegistryConstants.GOVERNANCE_COMPONENT_PATH +
            "/media-types";

    private static final String RESOURCE_MIME_TYPE_INDEX = "index";

    private static final String COLLECTION_MIME_TYPE_INDEX = "collection";

    private static final String CUSTOM_UI_MIME_TYPE_INDEX = "custom.ui";

    /**
     * Method to obtain the collection media types.
     *
     * @param registryContext the registry context.
     *
     * @return a String of collection media types, in the format name:type,name:type,...
     * @throws RegistryException if the operation failed.
     */
    public static String getCollectionMediaTypeMappings(RegistryContext registryContext)
            throws RegistryException {
        if (registryContext == null) {
            return RegistryContext.getBaseInstance().getCollectionMediaTypes();
        } else {
            return registryContext.getCollectionMediaTypes();
        }
    }

    /**
     * Method to obtain the collection media types.
     *
     * @param configSystemRegistry a configuration system registry instance.
     *
     * @return a String of collection media types, in the format name:type,name:type,...
     * @throws RegistryException if the operation failed.
     */
    public static String getCollectionMediaTypeMappings(Registry configSystemRegistry)
            throws RegistryException {

        RegistryContext registryContext = configSystemRegistry.getRegistryContext();

        if (getCollectionMediaTypeMappings(registryContext) != null) {
            return getCollectionMediaTypeMappings(registryContext);
        }

        Resource resource;
        String mediaTypeString = null;

        String resourcePath = MIME_TYPE_COLLECTION + RegistryConstants.PATH_SEPARATOR +
                RESOURCE_MIME_TYPE_INDEX;

        // TODO: Adding the media types should ideally be done by the handler associated with the
        // media type
        if (!configSystemRegistry.resourceExists(resourcePath)) {
            getResourceMediaTypeMappings(configSystemRegistry);
        }
        if (!configSystemRegistry.resourceExists(resourcePath + RegistryConstants.PATH_SEPARATOR +
                COLLECTION_MIME_TYPE_INDEX)) {
            resource = configSystemRegistry.newResource();
            resource.setProperty("synapse", "application/vnd.apache.synapse");
            resource.setProperty("esb", "application/vnd.wso2.esb");
            resource.setProperty("axis2", "application/vnd.apache.axis2");
            resource.setProperty("wsas", "application/vnd.wso2.wsas");
            resource.setDescription("This resource contains the media Types associated with " +
                    "collections on the Registry. Add, Edit or Delete properties to Manage Media " +
                    "Types.");
            configSystemRegistry.put(resourcePath + RegistryConstants.PATH_SEPARATOR +
                    COLLECTION_MIME_TYPE_INDEX, resource);
        } else {
            resource = configSystemRegistry.get(resourcePath + RegistryConstants.PATH_SEPARATOR +
                    COLLECTION_MIME_TYPE_INDEX);
        }
        Properties properties = resource.getProperties();
        if (properties.size() > 0) {
            Set<Object> keySet = properties.keySet();
            for (Object key : keySet) {
                if (key instanceof String) {
                    String ext = (String) key;
                    String value = resource.getProperty(ext);
                    String mediaTypeMapping = ext + ":" + value;
                    if (mediaTypeString == null) {
                        mediaTypeString = mediaTypeMapping;
                    } else {
                        mediaTypeString = mediaTypeString + "," + mediaTypeMapping;
                    }
                }
            }
        }
        registryContext.setCollectionMediaTypes(mediaTypeString);
        return mediaTypeString;
    }

    /**
     * Method to obtain the custom UI media types.
     *
     * @param registryContext the registry context.
     *
     * @return a String of custom UI media types, in the format name:type,name:type,...
     * @throws RegistryException if the operation failed.
     */
    public static String getCustomUIMediaTypeMappings(RegistryContext registryContext)
            throws RegistryException {
        if (registryContext == null) {
            return RegistryContext.getBaseInstance().getCustomUIMediaTypes();
        } else {
            return registryContext.getCustomUIMediaTypes();
        }
    }

    /**
     * Method to obtain the custom UI media types.
     *
     * @param configSystemRegistry a configuration system registry instance.
     *
     * @return a String of custom UI media types, in the format name:type,name:type,...
     * @throws RegistryException if the operation failed.
     */
    public static String getCustomUIMediaTypeMappings(Registry configSystemRegistry)
            throws RegistryException {

        RegistryContext registryContext = configSystemRegistry.getRegistryContext();

        if (getCustomUIMediaTypeMappings(registryContext) != null) {
            return getCustomUIMediaTypeMappings(registryContext);
        }

        Resource resource;
        String mediaTypeString = null;

        String resourcePath = MIME_TYPE_COLLECTION + RegistryConstants.PATH_SEPARATOR +
                RESOURCE_MIME_TYPE_INDEX;

        // TODO: Adding the media types should ideally be done by the handler associated with the
        // media type
        if (!configSystemRegistry.resourceExists(resourcePath)) {
            getResourceMediaTypeMappings(configSystemRegistry);
        }
        if (!configSystemRegistry.resourceExists(resourcePath + RegistryConstants.PATH_SEPARATOR +
                CUSTOM_UI_MIME_TYPE_INDEX)) {
            resource = configSystemRegistry.newResource();
            resource.setProperty("profiles", "application/vnd.wso2-profiles+xml");
            //resource.setProperty("service", "application/vnd.wso2-service+xml");
            resource.setDescription("This resource contains the media Types associated with " +
                    "custom user interfaces on the Registry. Add, Edit or Delete properties to " +
                    "Manage Media Types.");
            configSystemRegistry.put(resourcePath + RegistryConstants.PATH_SEPARATOR +
                    CUSTOM_UI_MIME_TYPE_INDEX, resource);
        } else {
            resource = configSystemRegistry.get(resourcePath + RegistryConstants.PATH_SEPARATOR +
                    CUSTOM_UI_MIME_TYPE_INDEX);
        }
        Properties properties = resource.getProperties();
        if (properties.size() > 0) {
            Set<Object> keySet = properties.keySet();
            for (Object key : keySet) {
                if (key instanceof String) {
                    String ext = (String) key;
                    String value = resource.getProperty(ext);
                    String mediaTypeMapping = ext + ":" + value;
                    if (mediaTypeString == null) {
                        mediaTypeString = mediaTypeMapping;
                    } else {
                        mediaTypeString = mediaTypeString + "," + mediaTypeMapping;
                    }
                }
            }
        }
        registryContext.setCustomUIMediaTypes(mediaTypeString);
        return mediaTypeString;
    }

    /**
     * Method to obtain the resource media types.
     *
     * @param registryContext the registry context.
     *
     * @return a String of resource media types, in the format extension:type,extension:type,...
     * @throws RegistryException if the operation failed.
     */
    public static String getResourceMediaTypeMappings(RegistryContext registryContext)
            throws RegistryException {
        if (registryContext == null) {
            return RegistryContext.getBaseInstance().getResourceMediaTypes();
        } else {
            return registryContext.getResourceMediaTypes();
        }
    }

    /**
     * Method to obtain the resource media types.
     *
     * @param configSystemRegistry a configuration system registry instance.
     *
     * @return a String of resource media types, in the format extension:type,extension:type,...
     * @throws RegistryException if the operation failed.
     */
    public static String getResourceMediaTypeMappings(Registry configSystemRegistry)
            throws RegistryException {

        RegistryContext registryContext = configSystemRegistry.getRegistryContext();

        if (getResourceMediaTypeMappings(registryContext) != null) {
            return getResourceMediaTypeMappings(registryContext);
        }

        Resource resource;
        String mediaTypeString = null;

        String resourcePath = MIME_TYPE_COLLECTION + RegistryConstants.PATH_SEPARATOR +
                RESOURCE_MIME_TYPE_INDEX;

        if (!configSystemRegistry.resourceExists(resourcePath)) {
            resource = configSystemRegistry.newCollection();
        } else {
            resource = configSystemRegistry.get(resourcePath);
            Properties properties = resource.getProperties();
            if (properties.size() > 0) {
                Set<Object> keySet = properties.keySet();
                for (Object key : keySet) {
                    if (key instanceof String) {
                        String ext = (String) key;
                        String value = resource.getProperty(ext);
                        String mediaTypeMapping = ext + ":" + value;
                        if (mediaTypeString == null) {
                            mediaTypeString = mediaTypeMapping;
                        } else {
                            mediaTypeString = mediaTypeString + "," + mediaTypeMapping;
                        }
                    }
                }
            }
            registryContext.setResourceMediaTypes(mediaTypeString);
            return mediaTypeString;
        }

        BufferedReader reader;
        try {
            File mimeFile = getMediaTypesFile();
            reader = new BufferedReader(new InputStreamReader(new FileInputStream(mimeFile)));
        } catch (Exception e) {
            String msg = "Failed to read the the media type definitions file. Only a limited " +
                    "set of media type definitions will be populated. ";
            log.error(msg, e);
            mediaTypeString = "txt:text/plain,jpg:image/jpeg,gif:image/gif";
            registryContext.setResourceMediaTypes(mediaTypeString);
            return mediaTypeString;
        }


        try {
            while (reader.ready()) {
                String mediaTypeData = reader.readLine().trim();
                if (mediaTypeData.startsWith("#")) {
                    // ignore the comments
                    continue;
                }

                if (mediaTypeData.length() == 0) {
                    // ignore the blank lines
                    continue;
                }

                // mime.type file delimits media types:extensions by tabs. if there is no
                // extension associated with a media type, there are no tabs in the line. so we
                // don't need such lines.
                if (mediaTypeData.indexOf("\t") > 0) {

                    String[] parts = mediaTypeData.split("\t+");
                    if (parts.length == 2 && parts[0].length() > 0 && parts[1].length() > 0) {

                        // there can multiple extensions associated with a single media type. in
                        // that case, extensions are delimited by a space.
                        String[] extensions = parts[1].trim().split(" ");
                        for (String extension : extensions) {
                            if (extension.length() > 0) {
                                String mediaTypeMapping = extension + ":" + parts[0];
                                resource.setProperty(extension, parts[0]);
                                if (mediaTypeString == null) {
                                    mediaTypeString = mediaTypeMapping;
                                } else {
                                    mediaTypeString = mediaTypeString + "," + mediaTypeMapping;
                                }
                            }
                        }
                    }
                }
            }
            resource.setDescription("This collection contains the media Types available for " +
                    "resources on the Registry. Add, Edit or Delete properties to Manage Media " +
                    "Types.");
            Resource collection = configSystemRegistry.newCollection();
            collection.setDescription("This collection lists the media types available on the " +
                    "Registry Server. Before changing an existing media type, please make sure " +
                    "to alter existing resources/collections and related configuration details.");
            configSystemRegistry.put(MIME_TYPE_COLLECTION, collection);
            configSystemRegistry.put(resourcePath, resource);

        } catch (IOException e) {
            String msg = "Could not read the media type mappings file from the location: ";
            throw new RegistryException(msg);

        } finally {
            try {
                reader.close();
            } catch (IOException ignore) {
            }
        }
        registryContext.setResourceMediaTypes(mediaTypeString);
        return mediaTypeString;
    }

    /**
     * Method to obtain the media type of a given resource.
     *
     * @param resourceName the name of the resource.
     *
     * @return the media type.
     * @throws RegistryException if the operation failed.
     */
    public static String getMediaType(String resourceName)
            throws RegistryException {
        if (resourceName == null || resourceName.indexOf(".") == -1 ||
                resourceName.indexOf(".") + 1 >= resourceName.length()) {
            return null;
        }
        String extension = resourceName.substring(resourceName.lastIndexOf(".") + 1);
        String mediaTypes = getResourceMediaTypeMappings(RegistryContext.getBaseInstance());
        if (mediaTypes == null) {
            // We don't treat this as an error, since some collections and resources would be
            // created even before the media types have been stored into the registry, and have
            // been initialized.
            return null;
        }
        extension += ":";
        if (mediaTypes.indexOf(extension) >= 0) {
            String temp = mediaTypes.substring(mediaTypes.indexOf(extension) + extension.length());
            if (temp.indexOf(",") == -1) {
                return temp;
            } else {
                return temp.substring(0, temp.indexOf(","));
            }
        }
        return null;
    }

    // Method to obtain the media types file.

    private static File getMediaTypesFile() throws RegistryException {

        String carbonHome = System.getProperty("carbon.home");

        if (carbonHome == null) {
            carbonHome = System.getenv("CARBON_HOME");
        }

        if (carbonHome != null) {

            File mediaTypesFile = new File(CarbonUtils.getCarbonConfigDirPath(),
                    RESOURCE_MEDIA_TYPE_MAPPINGS_FILE);
            if (!mediaTypesFile.exists()) {
                String msg = "Resource media type definitions file (mime.types) file does " +
                        "not exist in the path " + CarbonUtils.getCarbonConfigDirPath();
                log.error(msg);
                throw new RegistryException(msg);
            }

            return mediaTypesFile;

        } else {
            String msg = "carbon.home system property is not set. It is required to to derive " +
                    "the path of the media type definitions file (mime.types).";
            log.error(msg);
            throw new RegistryException(msg);
        }
    }
}
