package org.wso2.carbon.registry.dump;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMText;
import org.apache.axiom.om.util.Base64;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.RegistryConstants;

import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.namespace.QName;
import java.io.*;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

/**
 * This class performs operations to dump a registry path to file system. This includes copying
 * files, storing meta data, identifying modified files when updating etc.
 */

public class RegistryToFileSystemHandler {

    private static Log log = LogFactory.getLog(RegistryToFileSystemHandler.class);

    //registry instance
    private Registry registry;

    public RegistryToFileSystemHandler(Registry registry) {
        this.registry = registry;
    }

    /**
     * Dumps a registry path into a given directory in the file system
     * @param registryPath - registry collection path
     * @param startingDir - directory to store files
     * @throws Exception - on error
     */
    public void dumpToFilesystem(String registryPath, String startingDir) throws Exception {
        this.dumpToFilesystem(registryPath, startingDir, null);
    }

    /**
     * Dumps a registry path into a given directory in the file system
     * @param registryPath - registry collection path
     * @param startingDir - directory to store files
     * @param metaDir - directory to store meta info. if this null, startingDir is used for meta
     * @throws Exception - on error
     */
    public void dumpToFilesystem(String registryPath, String startingDir,
                                 String metaDir) throws Exception {
        if (registryPath == null || startingDir == null) {
            throw new Exception("Registry path and Directory to dump can't be null.");
        }

        //first try to get the path and confirm it is a collection
        Resource r = registry.get(registryPath);
        if (r == null) {
            throw new Exception("Error in dumping the path: " + registryPath +
                    ". There is no resource in the given path.");
        }
        if (!(r instanceof Collection)) {
            throw new Exception("Error in dumping the path: " + registryPath +
                    ". No Collection found at the given path");
        }
        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump(registryPath, writer);
        Reader input = new StringReader(writer.toString());

        //create the parser
        XMLStreamReader parser = XMLInputFactory.newInstance().createXMLStreamReader(input);
        //create the builder
        StAXOMBuilder builder = new StAXOMBuilder(parser);

        //if there's no special location to store meta info, store them parallel to files
        if (metaDir == null) {
            metaDir = startingDir;
        } else {
            createDirectory(metaDir);
        }
        checkOutRecursively(builder.getDocumentElement(), startingDir, metaDir);
    }

    /**
     * Recursively checkout from the registry until there are no children
     * @param root - OMElement containing data of the collection
     * @param parentDirectoryName - dir to dump
     * @param metaParent - dir to store meta info
     * @throws Exception - on error
     */
    private void checkOutRecursively(OMElement root, String parentDirectoryName,
                                     String metaParent) throws Exception {
        Iterator children = root.getChildren();

        String isCollectionString = root.getAttributeValue(new QName("isCollection"));
        boolean isCollection = isCollectionString.equals("true");
        String path = root.getAttributeValue(new QName("path"));
        String resourceName = DumpUtils.getResourceName(path);

        //directory path to checkout
        String dirPath = parentDirectoryName;
        String metaPath = metaParent;
        if (!resourceName.equals("/")) {
            dirPath = parentDirectoryName + "/" + resourceName;
            if (isCollection) {
                metaPath = metaParent + "/" + resourceName;
            }
        }

        byte[] contentBytes = new byte[0];
        List<OMElement> childList = new ArrayList<OMElement>();

        File file = new File(dirPath);
        if (isCollection) {
            // If the dir exists, this will override it
            file.mkdir();
        } else {
            if (!file.createNewFile()) {
                throw new Exception("The file : " + dirPath + " already exist. " +
                    "Paths to checkout should be empty.");
            }
        }

        contentBytes = readChildren(children, file, contentBytes, childList);

        if (!isCollection) {
            try {
                FileOutputStream fileStream = new FileOutputStream(file);
                fileStream.write(contentBytes);
                fileStream.flush();
                fileStream.close();
            } catch (IOException e) {
                log.error("Problem in writing content to the file: " + dirPath + ".", e);
            }
        }

        // handle meta stuff
        createDirectory(metaPath);
        String metaFileName;

        if (isCollection) {
            // creating the meta directory
            createDirectory(metaPath + "/" + DumpConstants.META_DIRECTORY);

            metaFileName = metaPath + "/" + DumpConstants.META_DIRECTORY + "/" +
                    DumpConstants.META_FILE_PREFIX + "." + DumpConstants.META_FILE_EXTENSION;
        } else {
            metaFileName = metaPath + "/" + DumpConstants.META_DIRECTORY +
                    "/" + DumpConstants.META_FILE_PREFIX + DumpUtils.getResourceName(path) +
                    "." + DumpConstants.META_FILE_EXTENSION;
        }
        DumpUtils.createMetaFile(metaFileName, root, file.lastModified());

        OMElement[] childElements = childList.toArray(new OMElement[childList.size()]);
        for (OMElement child : childElements) {
            // prepare the children absolute path
            checkOutRecursively(child, dirPath, metaPath);
        }
    }

    private byte[] readChildren(Iterator itr, File file,
                                byte[] contentBytes, List<OMElement> childList) {
        while (itr.hasNext()) {
            OMElement child = (OMElement) itr.next();
            String localName = child.getLocalName();

            // LastModified
            if (localName.equals("lastModified")) {
                OMText text = (OMText) child.getFirstOMChild();
                if (text != null) {
                    long date = Long.parseLong(text.getText());
                    file.setLastModified(date);
                }
            }
            // get content
            else if (localName.equals("content")) {
                OMText text = (OMText) child.getFirstOMChild();
                // we keep content as base64 encoded
                if (text != null) {
                    contentBytes = Base64.decode(text.getText());
                }
                child.detach();
            }
            // getting children, just storing in arraylist now, will used at the end
            else if (localName.equals("children")) {
                Iterator children = child.getChildren();

                while (children.hasNext()) {
                    Object nextChild = children.next();
                    if (!(nextChild instanceof OMElement)) {
                        continue;
                    }
                    OMElement childE = (OMElement) nextChild;
                    childList.add(childE);
                }
                child.detach();
            }
        }
        return contentBytes;
    }

    /**
     * Updates an already checked out directory if there are updated files in the registry
     * @param registryPath - registry path to update
     * @param startingDir - dir to update
     * @param metaStartingDir - dir to find meta info
     * @throws Exception - on error
     */
    public void update(String registryPath,
                       String startingDir, String metaStartingDir) throws Exception {
        if (metaStartingDir == null) {
            metaStartingDir = startingDir;
        }

        // check the existence of the meta directory
        File file = new File(metaStartingDir + "/" + DumpConstants.META_DIRECTORY);
        if (!file.exists()) {
            throw new Exception("Can't find meta directory for " + startingDir);
        }

        String directoryMetaFileName = metaStartingDir + "/" + DumpConstants.META_DIRECTORY + "/" +
                DumpConstants.META_FILE_PREFIX + "." + DumpConstants.META_FILE_EXTENSION;

        file = new File(directoryMetaFileName);
        if (!file.exists()) {
            throw new Exception("Can't find meta file for " + startingDir);
        }

        OMElement directoryMetaFile = DumpUtils.
                getOMElmentFromMetafile(directoryMetaFileName, true, null);

        String path = directoryMetaFile.getAttributeValue(new QName("path"));

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump(registryPath + path, writer);
        Reader input = new StringReader(writer.toString());

        //create the parser
        XMLStreamReader parser = XMLInputFactory.newInstance().createXMLStreamReader(input);
        //create the builder
        StAXOMBuilder builder = new StAXOMBuilder(parser);

        // get the element to restore
        OMElement dumpedElement = builder.getDocumentElement();

        String startingDirParent = startingDir;
        String metaStartingParent = metaStartingDir;
        if (!path.equals(RegistryConstants.ROOT_PATH)) {
            startingDirParent = startingDir + "/" + "..";
            metaStartingParent = metaStartingDir + "/" + "..";
        }
        updateRecursively(dumpedElement, startingDirParent, metaStartingParent);
    }

    /**
     * Recursively update the given path
     * @param root - OMElement of the root meta data
     * @param parentDirectoryName - parent
     * @param metaDir - start dir for meta
     * @throws Exception - on error
     */
    private void updateRecursively(OMElement root, String parentDirectoryName,
                                   String metaDir) throws Exception {
        Iterator children = root.getChildren();

        String isCollectionString = root.getAttributeValue(new QName("isCollection"));
        boolean isCollection = isCollectionString.equals("true");
        String path = root.getAttributeValue(new QName("path"));
        String resourceName = DumpUtils.getResourceName(path);

        String dirPath = parentDirectoryName;
        String metaPath = metaDir;
        if (!resourceName.equals("/")) {
            dirPath = parentDirectoryName + "/" + resourceName;
            if (isCollection) {
                metaPath = metaDir + "/" + resourceName;
            }
        }

        byte[] contentBytes = new byte[0];

        List<OMElement> childList = new ArrayList<OMElement>();
        File file = new File(dirPath);

        boolean isUpdating = false;
        boolean noDirectoryUpdate = false;

        if (file.exists()) {
            isUpdating = true;
            // we need to check the last updated times of the current resource
            // and the updating resource
            long currentResourceUpdatedTime = file.lastModified();
            OMElement updatingLastModifiedElement =
                    root.getFirstChildWithName(new QName("lastModified"));
            long updatingLastModified = Long.parseLong(updatingLastModifiedElement.getText());

            // access the meta info of the current
            String metaFilename;

            if (isCollection) {
                metaFilename = metaPath + "/" + DumpConstants.META_DIRECTORY + "/" +
                        DumpConstants.META_FILE_PREFIX + "." + DumpConstants.META_FILE_EXTENSION;
            } else {
                metaFilename = metaPath + "/" + DumpConstants.META_DIRECTORY + "/" +
                        DumpConstants.META_FILE_PREFIX + DumpUtils.getResourceName(path) +
                        "." + DumpConstants.META_FILE_EXTENSION;
            }

            File metaFile = new File(metaFilename);
            if (metaFile.exists()) {
                // get the meta file OMElement
                OMElement metaFileElement = DumpUtils.getOMElmentFromMetafile(metaFilename, isCollection, null);
                OMElement metaFileLastModifiedElement = metaFileElement.getFirstChildWithName(new QName("lastModified"));
                long metaFileLastModified = Long.parseLong(metaFileLastModifiedElement.getText());
                String metaFilePhysicalLastModifiedString = metaFileElement.getAttributeValue(new QName("physicalModifiedDate"));
                long metaFilePhysicalLastModified = Long.parseLong(metaFilePhysicalLastModifiedString);

                if (metaFilePhysicalLastModified == currentResourceUpdatedTime) {
                    // there is no modifications happens to the current file,
                    if (metaFileLastModified == updatingLastModified) {
                        // the file in the server is not updated, so just keep the current file.
                        if (file.isDirectory()) {
                            noDirectoryUpdate = true;
                        } else {
                            return;
                        }
                    } else {
                        // there is a server update to the file, so lets allow it to update the local one
                        // localone is not updated, so it will be overwritten by the server one
                    }
                } else if (metaFileLastModified == updatingLastModified) {
                    // there is no server side changes, but there are client side changes, just don't update it
                    if (file.isDirectory()) {
                        noDirectoryUpdate = true;
                    } else {
                        return;
                    }
                }
            }
        }
        if (!isUpdating) {
            try {
                // Create file if it does not exist
                if (isCollection && !noDirectoryUpdate) {
                    file.mkdir();
                } else {
                    file.createNewFile();
                }
            } catch (IOException e) {
                throw new Exception("Problem in creating: " + dirPath + "." , e);
            }
        }

        contentBytes = readChildren(children, file, contentBytes, childList);

        if (!isCollection) {
            try {
                FileOutputStream fileStream = new FileOutputStream(file);
                fileStream.write(contentBytes);
                fileStream.flush();
                fileStream.close();
            } catch (IOException e) {
                throw new Exception("Problem in writing content to the file: " + dirPath + ".", e);
            }
        }

        // creating the meta directory
        if (isCollection) {
            File tmp = new File(metaPath);
            if (!tmp.exists()) {
                tmp.mkdir();
            }
            String metaDirectoryName = metaPath + "/" + DumpConstants.META_DIRECTORY;
            File metaDirectory = new File(metaDirectoryName);
            metaDirectory.mkdir();
        }

        OMElement[] childElements = childList.toArray(new OMElement[childList.size()]);
        ArrayList<String> childNames = new ArrayList<String>();
        for (OMElement child : childElements) {
            // prepare the children absolute path
            updateRecursively(child, dirPath, metaPath);

            String childPath = child.getAttributeValue(new QName("path"));
            String childResourceName = DumpUtils.getResourceName(childPath);
            String fileResourceName = DumpUtils.encodeResourcename(childResourceName);
            childNames.add(fileResourceName);
        }

        // now we are checking which files have been deleted at the server end.
        String[] childFileNames = file.list();
        if (childFileNames != null) {
            for (String childFileName: childFileNames) {
                if (childNames.contains(childFileName)) {
                    // this files stays on the server as well, so nothing to worry
                    continue;
                }
                File childFile = new File(file, childFileName);
                boolean isDeleted = DumpUtils.deleteFile(childFile);
                if (!isDeleted) {
                    throw new Exception("Error while trying to delete file " + childFileName);
                }
            }
        }

        if (file.isDirectory() && noDirectoryUpdate) {
            return;
        }

        // creating the meta file
        String metaFileName;
        if (isCollection) {
            metaFileName = metaPath + "/" + DumpConstants.META_DIRECTORY + "/" +
                    DumpConstants.META_FILE_PREFIX + "." + DumpConstants.META_FILE_EXTENSION;
        } else {
            metaFileName = metaPath + "/" + DumpConstants.META_DIRECTORY +
                    "/" + DumpConstants.META_FILE_PREFIX + DumpUtils.getResourceName(path) +
                    "." + DumpConstants.META_FILE_EXTENSION;
        }
        DumpUtils.createMetaFile(metaFileName, root, file.lastModified());

    }

    /**
     * Create a directory if doesn't already exists
     * @param path - dir path
     */
    private void createDirectory(String path) {
        File repoMetaDir = new File(path);
        if (!repoMetaDir.exists()) {
            repoMetaDir.mkdir();
        }
    }

}
