/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.test.jdbc;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.wso2.carbon.registry.core.jdbc.InMemoryEmbeddedRegistryService;

import org.wso2.carbon.registry.core.*;
import org.wso2.carbon.registry.core.jdbc.utils.DumpReader;
import org.wso2.carbon.registry.core.test.utils.BaseTestCase;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.utils.RegistryUtils;
import org.wso2.carbon.user.core.config.RealmConfiguration;

import java.io.StringWriter;
import java.io.Reader;
import java.io.StringReader;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;


public class DumpTest extends BaseTestCase {

    protected static InMemoryEmbeddedRegistryService embeddedRegistryService = null;
    protected static Registry registry = null;
    

    public void setUp() {
        super.setUp();
        if (embeddedRegistryService != null) {
            return;
        }
        try {
            embeddedRegistryService = new InMemoryEmbeddedRegistryService();
            RealmUnawareRegistryCoreServiceComponent comp = new RealmUnawareRegistryCoreServiceComponent();
            comp.setRealmService(embeddedRegistryService.getRealmService());
            comp.registerBuiltInHandlers(embeddedRegistryService);
            
            // get the realm config to retrieve admin username, password
            RealmConfiguration realmConfig = embeddedRegistryService.getBootstrapRealmConfiguration();
            registry = embeddedRegistryService.getUserRegistry(
                realmConfig.getAdminUserName(), realmConfig.getAdminPassword());
        } catch (RegistryException e) {
                fail("Failed to initialize the registry. Caused by: " + e.getMessage());
        }
    }

    public void testResourceDump() throws RegistryException, XMLStreamException {
        Resource r = registry.newResource();
        r.setProperty("key1", "value1");
        r.setProperty("key2", "value2");
        r.setContent("content 1");
        registry.put("/testDump", r);

        registry.addComment("/testDump", new Comment("comment1"));
        registry.addComment("/testDump", new Comment("comment2"));

        registry.applyTag("/testDump", "tag1");
        registry.applyTag("/testDump", "tag2");

        registry.rateResource("/testDump", 3);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testDump", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/testDumpDup", input);
        r = registry.get("/testDumpDup");

        assertEquals(new String((byte[])r.getContent()), "content 1");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value1");
        assertEquals(r.getProperty("key2"), "value2");

        // getting the comments
        Comment[] comments = registry.getComments("/testDumpDup");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment1");
        assertEquals(comments[1].getText(), "comment2");

        // getting the tags
        Tag[] tags = registry.getTags("/testDumpDup");
        assertEquals(tags[0].getTagName(), "tag1");
        assertEquals(tags[1].getTagName(), "tag2");

        // getting the ratings
        int rate = registry.getRating("/testDumpDup", "admin");
        assertEquals(rate, 3);
    }


    public void testCollectionDump() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value1");
        r.setProperty("key2", "value2");
        registry.put("/testDumpC", r);

        // adding children
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testDumpC/child1C", r1);

        Resource r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/testDumpC/child2R", r2);

        registry.addComment("/testDumpC", new Comment("comment1"));
        registry.addComment("/testDumpC", new Comment("comment2"));

        registry.applyTag("/testDumpC", "tag1");
        registry.applyTag("/testDumpC", "tag2");

        registry.rateResource("/testDumpC", 3);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testDumpC", writer);
        Reader input = new StringReader(writer.toString());

        // now restoring and retriving the dumped element
        registry.restore("/testDumpDupC", input);
        r = registry.get("/testDumpDupC");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value1");
        assertEquals(r.getProperty("key2"), "value2");

        // getting the comments
        Comment[] comments = registry.getComments("/testDumpDupC");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment1");
        assertEquals(comments[1].getText(), "comment2");

        // getting the tags
        Tag[] tags = registry.getTags("/testDumpDupC");
        assertEquals(tags[0].getTagName(), "tag1");
        assertEquals(tags[1].getTagName(), "tag2");

        // getting the ratings
        int rate = registry.getRating("/testDumpDupC", "admin");
        assertEquals(rate, 3);

        // getting the chils
        r1 = registry.get("/testDumpDupC/child1C");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");

        r2 = registry.get("/testDumpDupC/child2R");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }


    public void testRootDump() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value1");
        r.setProperty("key2", "value2");
        registry.put("/", r);

        // adding children
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/child1CX", r1);

        Resource r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/child2RX", r2);

        registry.addComment("/", new Comment("comment1"));
        registry.addComment("/", new Comment("comment2"));

        registry.applyTag("/", "tag1");
        registry.applyTag("/", "tag2");

        registry.rateResource("/", 3);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/testDumpDupR", input);
        r = registry.get("/testDumpDupR");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value1");
        assertEquals(r.getProperty("key2"), "value2");

        // getting the comments
        Comment[] comments = registry.getComments("/testDumpDupR");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment1");
        assertEquals(comments[1].getText(), "comment2");

        // getting the tags
        Tag[] tags = registry.getTags("/testDumpDupR");
        assertEquals(tags[0].getTagName(), "tag1");
        assertEquals(tags[1].getTagName(), "tag2");

        // getting the ratings
        int rate = registry.getRating("/testDumpDupR", "admin");
        assertEquals(rate, 3);
        
        // getting the chils
        r1 = registry.get("/testDumpDupR/child1CX");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");

        r2 = registry.get("/testDumpDupR/child2RX");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }

    public void testRootRestore() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value3");
        r.setProperty("key2", "value4");
        registry.put("/testSomewhereElse", r);

        // adding children
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY", r1);

        Resource r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/testSomewhereElse/child2RY", r2);

        registry.addComment("/testSomewhereElse", new Comment("comment3"));
        registry.addComment("/testSomewhereElse", new Comment("comment4"));

        registry.applyTag("/testSomewhereElse", "tag3");
        registry.applyTag("/testSomewhereElse", "tag4");

        registry.rateResource("/testSomewhereElse", 2);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testSomewhereElse", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/", input);
        r = registry.get("/");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value3");
        assertEquals(r.getProperty("key2"), "value4");

        // getting the comments
        Comment[] comments = registry.getComments("/");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment3");
        assertEquals(comments[1].getText(), "comment4");

        // getting the tags
        Tag[] tags = registry.getTags("/");
        assertEquals(tags[0].getTagName(), "tag3");
        assertEquals(tags[1].getTagName(), "tag4");

        // getting the ratings
        int rate = registry.getRating("/", "admin");
        assertEquals(rate, 2);

        // getting the chils
        r1 = registry.get("/child1CY");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");

        r2 = registry.get("/child2RY");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }

    public void testSimpleNewRestore() throws RegistryException, XMLStreamException {
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY/xxxx", r1);
        
        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testSomewhereElse", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/", input);
        
        Resource r2 = registry.get("/child1CY/xxxx");
        assertTrue((r2 instanceof CollectionImpl));
    }

    public void testNewRestore() throws RegistryException, XMLStreamException {
        Resource r = registry.newCollection();
        r.setProperty("key1", "value3");
        r.setProperty("key2", "value4");
        registry.put("/testSomewhereElse", r);

        // adding children
        Resource r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY", r1);
        
        r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY/xxxx", r1);
        
        r1 = registry.newCollection();
        r1.setProperty("key1", "value1C");
        r1.setProperty("key2", "value2C");
        registry.put("/testSomewhereElse/child1CY/yyyy", r1);

        Resource r2 = registry.newCollection();
        //r2.setContent("content child2R");
        registry.put("/testSomewhereElse/newwwC1", r2);

        r2 = registry.newCollection();
        //r2.setContent("content child2R");
        registry.put("/testSomewhereElse/newwwC2", r2);

        r2 = registry.newResource();
        r2.setContent("content child2R");
        registry.put("/testSomewhereElse/child2RY", r2);

        registry.addComment("/testSomewhereElse", new Comment("comment3"));
        registry.addComment("/testSomewhereElse", new Comment("comment4"));

        registry.applyTag("/testSomewhereElse", "tag3");
        registry.applyTag("/testSomewhereElse", "tag4");

        registry.rateResource("/testSomewhereElse", 2);

        // doing the dump
        StringWriter writer = new StringWriter();
        registry.dump("/testSomewhereElse", writer);
        Reader input = new StringReader(writer.toString());
        registry.restore("/", input);
        r = registry.get("/");

        // checking the properties.
        assertEquals(r.getProperties().size(), 2);
        assertEquals(r.getProperty("key1"), "value3");
        assertEquals(r.getProperty("key2"), "value4");

        // getting the comments
        Comment[] comments = registry.getComments("/");
        assertEquals(comments.length, 2);
        assertEquals(comments[0].getText(), "comment3");
        assertEquals(comments[1].getText(), "comment4");

        // getting the tags
        Tag[] tags = registry.getTags("/");
        assertEquals(tags[0].getTagName(), "tag3");
        assertEquals(tags[1].getTagName(), "tag4");

        // getting the ratings
        int rate = registry.getRating("/", "admin");
        assertEquals(rate, 2);

        // getting the chils
        r1 = registry.get("/child1CY");
        assertEquals(r1.getProperties().size(), 2);
        assertEquals(r1.getProperty("key1"), "value1C");
        assertEquals(r1.getProperty("key2"), "value2C");

        r2 = registry.get("/child1CY/xxxx");
        assertTrue((r2 instanceof CollectionImpl));
        r2 = registry.get("/child1CY/yyyy");
        assertTrue((r2 instanceof CollectionImpl));

        r2 = registry.get("/child2RY");
        assertEquals(new String((byte[])r2.getContent()), "content child2R");
    }

    public void testAbsoluteAssociationPath() throws Exception {
        assertEquals("/abc", RegistryUtils.getAbsoluteAssociationPath("../abc", "/lm/pqr"));
        assertEquals("/abc/def", RegistryUtils.getAbsoluteAssociationPath("../../../abc/def", "/lm/pqr/b/boo"));
        assertEquals("/abc/hag/def", RegistryUtils.getAbsoluteAssociationPath("../hag/def", "/abc/boo/lm"));
        assertEquals("/abc", RegistryUtils.getAbsoluteAssociationPath("abc", "/pqr"));
        assertEquals("/boolm/squi/abc", RegistryUtils.getAbsoluteAssociationPath("squi/abc", "/boolm/squi2"));
        assertEquals("/abc", RegistryUtils.getAbsoluteAssociationPath("abc", "/abc"));

        // go beyond cases
        assertEquals("//abc", RegistryUtils.getAbsoluteAssociationPath("../../abc", "/lm/pqr"));
        assertEquals("///abc", RegistryUtils.getAbsoluteAssociationPath("../../../abc", "/lm/pqr"));
        assertEquals("////abc", RegistryUtils.getAbsoluteAssociationPath("../../../../abc", "/lm/pqr"));         
    }


    public void testRelativeAssociationPath() throws Exception {
        assertEquals("../abc", RegistryUtils.getRelativeAssociationPath("/abc", "/lm/pqr"));
        assertEquals("../../../abc/def", RegistryUtils.getRelativeAssociationPath("/abc/def", "/lm/pqr/b/boo"));
        assertEquals("../hag/def", RegistryUtils.getRelativeAssociationPath("/abc/hag/def", "/abc/boo/lm"));
        assertEquals("abc", RegistryUtils.getRelativeAssociationPath("/abc", "/pqr"));
        assertEquals("squi/abc", RegistryUtils.getRelativeAssociationPath("/boolm/squi/abc", "/boolm/squi2"));
        assertEquals("abc", RegistryUtils.getRelativeAssociationPath("/abc", "/abc"));

        assertEquals("../../abc", RegistryUtils.getRelativeAssociationPath("//abc", "/lm/pqr"));
        assertEquals("../../../abc", RegistryUtils.getRelativeAssociationPath("///abc", "/lm/pqr"));
        assertEquals("../../../../abc", RegistryUtils.getRelativeAssociationPath("////abc", "/lm/pqr"));
    }

    public void testDumpWithSymLink() throws Exception {

        Resource r = registry.newResource();
        r.setProperty("key1", "value3");
        r.setProperty("key2", "value4");
        registry.put("/my/original/link/resource", r);

        registry.createLink("/my/sym/link/resource", "/my/original/link/resource");

        // just check the sym
        Resource r2 = registry.get("/my/sym/link/resource");
        assertEquals("value3", r2.getProperty("key1"));
        assertEquals("value4", r2.getProperty("key2"));

        // now get a dump of /my
        StringWriter writer = new StringWriter();
        registry.dump("/my", writer);

        StringReader reader = new StringReader(writer.toString());
        // putting reader
        registry.restore("/restored", reader);

        Resource r3 =  registry.get("/restored/sym/link/resource");
        assertEquals("value3", r3.getProperty("key1"));
        assertEquals("value4", r3.getProperty("key2"));

        // do some changes to the original and check the sym link changing
        Resource r4 = registry.get("/restored/original/link/resource");
        r4.setProperty("key3", "value5");
        registry.put("/restored/original/link/resource", r4);


        Resource r5 =  registry.get("/restored/sym/link/resource");
        assertEquals("value5", r5.getProperty("key3"));

    }
}
