/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.test.jdbc;

import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.Registry;
import org.wso2.carbon.registry.core.Resource;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.jdbc.EmbeddedRegistryService;
import org.wso2.carbon.registry.core.jdbc.InMemoryEmbeddedRegistryService;

import org.wso2.carbon.registry.core.test.utils.BaseTestCase;

public class SymlinkTest extends BaseTestCase {

    protected static Registry registry = null;
    
    protected static EmbeddedRegistryService embeddedRegistryService = null;

    public void setUp() {
        super.setUp();
        if (embeddedRegistryService != null) {
            return;
        }
        try {
            embeddedRegistryService = new InMemoryEmbeddedRegistryService();
            registry = embeddedRegistryService.getSystemRegistry();
        } catch (RegistryException e) {
            fail("Failed to initialize the registry. Caused by: " + e.getMessage());
        }
    }



    public void testSymbolicLinks() throws RegistryException {
        Collection testCollection = registry.newCollection();
        Resource testResource = registry.newResource();
        testCollection.setProperty("name", "valueC");
        registry.put("/testCollection", testCollection);
        testResource.setProperty("name", "value1");
        registry.put("/testCollection/testResource", testResource);
        
        registry.createLink("/mountCollection", "/testCollection");
        Resource mountCollection = registry.get("/mountCollection");
        assertEquals(mountCollection.getProperty("name"), "valueC");
        String [] children = (String [])mountCollection.getContent();
        assertEquals(children[0], "/mountCollection/testResource");
        Resource mountedResource = registry.get("/mountCollection/testResource");
        assertEquals(mountedResource.getProperty("name"), "value1");

        // trying to create a symlink in an already existing location
        registry.createLink("/testCollection", "/");
        boolean exceptionOccured = false;
        try {
            registry.createLink("/testCollection", "/testCollection");
        } catch (Exception e) {
            exceptionOccured = true;
        }
        assertTrue("Symlink link to itself is not valid", exceptionOccured);

    }


    public void testSymbolicLinksRoots() throws RegistryException {
        Collection testCollection = registry.newCollection();
        Resource testResource = registry.newResource();
        testCollection.setProperty("name", "valueC");
        registry.put("/", testCollection);
        testResource.setProperty("name", "value1");
        registry.put("/testCollection2", testResource);

        registry.createLink("/mountCollection2", "/");
        Resource mountCollection = registry.get("/mountCollection2");
        assertEquals(mountCollection.getProperty("name"), "valueC");
        String [] children = (String [])mountCollection.getContent();
        assertEquals(children[0], "/mountCollection2/testCollection2");
        Resource mountedResource = registry.get("/mountCollection2/testCollection2");
        assertEquals(mountedResource.getProperty("name"), "value1");
    }    

    public void testCopySourceSymLinkRoot() throws RegistryException {
        Collection testResource = registry.newCollection();
        registry.put("/a/b/c", testResource);
        registry.createLink("/p", "/a/b"); // now p is a link to b

        registry.copy("/p", "/q");

        assertTrue("q should have resource c", registry.resourceExists("/q/c"));

    }

    public void testCopySourceFromSymlink() throws RegistryException {
        Collection testResource = registry.newCollection();
        registry.put("/a1/b1/c1/d1", testResource);
        registry.createLink("/p1", "/a1/b1"); // now p is a link to b

        registry.copy("/p1/c1", "/q1");

        assertTrue("q1 should have resource d1", registry.resourceExists("/q1/d1"));

        
    }

    public void testCopyTargetToSymLink() throws RegistryException {

        Collection testResource = registry.newCollection();
        registry.put("/a2/b2/", testResource);
        registry.put("/c2/d2/", testResource);
        registry.createLink("/p2", "/a2/b2"); // now p is a link to b

        registry.copy("/c2", "/p2/c2");

        assertTrue("q2 should have resource c2", registry.resourceExists("/p2/c2/d2"));
        assertTrue("q2 should have resource c2", registry.resourceExists("/a2/b2/c2/d2"));
    }

    public void testCopyBothSymLink() throws RegistryException {

        Collection testResource = registry.newCollection();
        registry.put("/a3/b3/", testResource);
        registry.put("/c3/d3/e3", testResource);
        registry.createLink("/p3", "/a3/b3");   
        registry.createLink("/q3", "/c3/d3");

        registry.copy("/q3/e3", "/p3/e3");

        assertTrue("p3 should have resource e3", registry.resourceExists("/p3/e3"));
        assertTrue("a3/b3/ should have resource e3", registry.resourceExists("/a3/b3/e3"));
    }


    public void testMoveSourceSymLinkRoot() throws RegistryException {
        Collection testResource = registry.newCollection();
        registry.put("/a4/b4/c4", testResource);
        registry.createLink("/p4", "/a4/b4"); // now p is a link to b

        registry.move("/p4", "/q4");

        assertTrue("q4 should have resource c4", registry.resourceExists("/q4/c4"));

    }

    public void testMoveSourceFromSymlink() throws RegistryException {
        Collection testResource = registry.newCollection();
        registry.put("/a11/b11/c11/d11", testResource);
        registry.createLink("/p11", "/a11/b11"); // now p is a link to b

        registry.move("/p11/c11", "/q11");

        assertTrue("q11 should have resource d11", registry.resourceExists("/q11/d11"));


    }

    public void testMoveTargetToSymLink() throws RegistryException {

        Collection testResource = registry.newCollection();
        registry.put("/a21/b21/", testResource);
        registry.put("/c21/d21/", testResource);
        registry.createLink("/p21", "/a21/b21"); // now p is a link to b

        registry.move("/c21", "/p21/c21");

        assertTrue("q21 should have resource c21", registry.resourceExists("/p21/c21/d21"));
        assertTrue("q21 should have resource c21", registry.resourceExists("/a21/b21/c21/d21"));
    }

    public void testMoveBothSymLink() throws RegistryException {

        Collection testResource = registry.newCollection();
        registry.put("/a31/b31/", testResource);
        registry.put("/c31/d31/e31", testResource);
        registry.createLink("/p31", "/a31/b31");
        registry.createLink("/q31", "/c31/d31");

        registry.move("/q31/e31", "/p31/e31");

        assertTrue("p31 should have resource e31", registry.resourceExists("/p31/e31"));
        assertTrue("a31/b31/ should have resource e31", registry.resourceExists("/a31/b31/e31"));
    }


    public void testMoveSourceSymLinkRootCollection() throws RegistryException {
        Resource testResource = registry.newResource();
        registry.put("/a6/b6/c6", testResource);
        registry.createLink("/p6", "/a6/b6"); // now p is a link to b
        registry.createLink("/x6", "/a6/b6/c6");

        registry.move("/p6", "/q6");
        registry.move("/x6", "/y6");

        assertTrue("q6 should be a collection", registry.get("/q6") instanceof Collection);
        assertTrue("y6 should be a non-collection", !(registry.get("/y6") instanceof Collection));

    }

}
