/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.server.admin.module.handler;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.clustering.ClusteringConstants;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.handlers.AbstractHandler;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.core.services.authentication.CarbonServerAuthenticator;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.server.admin.auth.AuthenticatorServerRegistry;
import org.wso2.carbon.server.admin.internal.ServerAdminServiceComponent;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.util.UserCoreUtil;
import org.wso2.carbon.utils.ServerConstants;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.xml.namespace.QName;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 *
 */
public class AuthenticationHandler extends AbstractHandler {
    private static final Log log = LogFactory.getLog(AuthenticationHandler.class);
    private MessageContext msgContext;

    public InvocationResponse invoke(MessageContext messageContext) throws AxisFault {
        this.msgContext = messageContext;

        // do not authenticate if invoked thru local transport.
        if (Constants.TRANSPORT_LOCAL.equals(msgContext.getIncomingTransportName())) {
            return InvocationResponse.CONTINUE;
        }
        
        authenticate((String) messageContext.getProperty(MessageContext.REMOTE_ADDR));

        // Setting the tenant domain in the MessageContext since it is transport independent
        // unlike in the case of HttpSession. Used by multitenant service hosting
        HttpServletRequest request = (HttpServletRequest) msgContext.getProperty(
                HTTPConstants.MC_HTTP_SERVLETREQUEST);
        if (request != null && request.getSession(false) != null) {
            String domain = (String) request.getSession(false).getAttribute(
                    RegistryConstants.TENANT_DOMAIN);
            msgContext.setProperty(CarbonConstants.TENANT_DOMAIN, domain);
        }

        return InvocationResponse.CONTINUE;
    }

    protected void authenticate(String remoteIP) throws AxisFault {
        if (callToGeneralService()) {
            return;
        }
        try {
            if (!isAuthenticated(remoteIP)) {
                throw new AxisFault("Access Denied. Please login first.",
                        ServerConstants.AUTHENTICATION_FAULT_CODE);
            }
        } catch (AxisFault e) {
            throw e;
        } catch (AuthenticationException e) {
            throw new AxisFault("Access Denied. Session timed out.",
                    ServerConstants.AUTHENTICATION_FAULT_CODE);
        } catch (Throwable e) {
            log.error(e.getMessage(), e);
            throw new AxisFault("Authentication failure", ServerConstants.AUTHENTICATION_FAULT_CODE);
        }
    }

    private boolean isAuthenticated(String remoteIP) throws AuthenticationException {
        AxisService service = msgContext.getAxisService();
        AxisOperation operation = msgContext.getAxisOperation();

        String opName = operation.getName().getLocalPart();
        String serviceName = service.getName();

        Parameter authenticationParameter = operation.getParameter("DoAuthentication");

        if (authenticationParameter != null && "false".equals(authenticationParameter.getValue())) {
            return true;
        }

        HttpServletRequest request = (HttpServletRequest) msgContext
                .getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);

        HttpSession httpSession = request.getSession();

        if ((serviceName.equals(ServerConstants.AUTHENTICATION_ADMIN_SERVICE) || serviceName
                .equals(ClusteringConstants.NODE_MANAGER_SERVICE))) {
            if (opName.equals("login") || opName.equals("loginWithDelegation")
                    || opName.equals("getAutheticationToken")) {
                SOAPBody soapBody = msgContext.getEnvelope().getBody();

                OMElement usernameElem = soapBody.getFirstElement().getFirstChildWithName(
                        new QName(ServerConstants.AUTHENTICATION_SERVICE_NS,
                                ServerConstants.AUTHENTICATION_SERVICE_USERNAME));
                if (usernameElem != null) {
                    String userName = usernameElem.getText();
                    String tenantAwareUsername = UserCoreUtil.getTenantLessUsername(userName);
                    String tenantDomain;
                    try {
                        tenantDomain = UserCoreUtil.getTenantDomain(ServerAdminServiceComponent.getRealmService(), userName);
                    } catch (UserStoreException e) {
                        log.error("Unable to find the tenant domain :" + e.getMessage(), e);
                        throw new AuthenticationException("Unable to find the tenant domain", e);
                    }

                    httpSession.setAttribute(ServerConstants.ADMIN_USERNAME, tenantAwareUsername);
                    httpSession.setAttribute(CarbonConstants.TENANT_DOMAIN, tenantDomain);
                    return true;
                }
            } else if (opName.equals("logout")) {
                return true;
            }
        }

        CarbonServerAuthenticator authenticator = AuthenticatorServerRegistry
                .getCarbonAuthenticator(msgContext);
        boolean isAuthenticated = authenticator.isAuthenticated(msgContext);

        if (!isAuthenticated) {
            SimpleDateFormat date = new SimpleDateFormat("'['yyyy-MM-dd HH:mm:ss,SSSS']'");
            log.warn("Illegal access attempt at " + date.format(new Date()) + " from IP address "
                    + remoteIP + " : Service is " + serviceName);
        }
        // return true;
        return isAuthenticated;
    }

    private boolean callToGeneralService() {
        boolean isGeneral = true;
        AxisService service = msgContext.getAxisService();
        Parameter param = service.getParameter("adminService");
        if (param != null && "true".equals(param.getValue())) {
            isGeneral = false;
        }
        return isGeneral;
    }

}
