/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.server.admin.service;

import java.net.SocketException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.CarbonConstants;
import org.wso2.carbon.core.AbstractAdmin;
import org.wso2.carbon.core.ServerManagement;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.server.admin.common.IServerAdmin;
import org.wso2.carbon.server.admin.common.ServerData;
import org.wso2.carbon.server.admin.internal.ServerAdminServiceComponent;
import org.wso2.carbon.utils.Controllable;
import org.wso2.carbon.utils.NetworkUtils;
import org.wso2.carbon.utils.ServerConfiguration;
import org.wso2.carbon.utils.ServerConstants;


/**
 * Admin service to manage server related operations
 */
public class ServerAdmin extends AbstractAdmin implements ServerAdminMBean, IServerAdmin {
    private static final Log log = LogFactory.getLog(ServerAdmin.class);
    private SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd' 'HH:mm:ss");
    private static final int SECONDS_PER_DAY = 3600 * 24;

    public static ClassLoader classloader;

    public static ServerConfiguration serverConfig;

    public ServerAdmin() {
    }

    public ServerData getServerData() throws Exception {
        ServerData.serverConfig = serverConfig;
        boolean isRestricted = true;
        
        MessageContext msgContext = MessageContext.getCurrentMessageContext();
        if (msgContext != null) {
            HttpServletRequest request = (HttpServletRequest) msgContext
                    .getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
            HttpSession httpSession = request.getSession(false);
            if (httpSession != null) {

                String userName = (String) httpSession.getAttribute(ServerConstants.ADMIN_USERNAME);
                isRestricted = !getUserRealm().getAuthorizationManager().isUserAuthorized(userName,
                        "/permission/protected/server-admin/homepage",
                        CarbonConstants.UI_PERMISSION_ACTION);
            }
        } else { // non SOAP call
            isRestricted = false;
        }
        
        String location = null;
        if(!isRestricted) {
            location = getAxisConfig().getRepository().toString();
        }
        ServerData data = new ServerData(ServerConstants.SERVER_NAME,
                                         location,
                                         getTenantDomain() != null, isRestricted);
        
        if (!isRestricted) {
            Parameter systemStartTime = getAxisConfig().getParameter(
                    CarbonConstants.SERVER_START_TIME);
            long startTime = 0;
            if (systemStartTime != null) {
                startTime = Long.parseLong((String)systemStartTime.getValue());
            }
            Date stTime = new Date(startTime);
            data.setServerStartTime(dateFormatter.format(stTime));
            data.setSystemUpTime(getTime((System.currentTimeMillis() - startTime) / 1000));

            ServerConfiguration serverConfig = ServerAdminServiceComponent.getServerConfig();
            String registryType = serverConfig.getFirstProperty("Registry.Type");

            if (registryType == null) {
                registryType = "embedded";
            }
            data.setRegistryType(registryType);

            // Extract DB related data from RegistryContext
            if (registryType.equals("embedded")) {
                try {
                    Connection dbConnection = RegistryContext.getBaseInstance().getDataSource()
                            .getConnection();
                    DatabaseMetaData metaData = dbConnection.getMetaData();
                    if (metaData != null) {
                        data.setDbName(metaData.getDatabaseProductName());
                        data.setDbVersion(metaData.getDatabaseProductVersion());
                        data.setDbDriverName(metaData.getDriverName());
                        data.setDbDriverVersion(metaData.getDriverVersion());
                        data.setDbURL(metaData.getURL());
                    }
                    dbConnection.close();
                } catch (SQLException e) {
                    String msg = "Cannot create DB connection";
                    log.error(msg, e);
                    throw new AxisFault(msg, e);
                }
            } else if (registryType.equals("remote")) {
                data.setRemoteRegistryChroot(serverConfig.getFirstProperty("Registry.Chroot"));
                data.setRemoteRegistryURL(serverConfig.getFirstProperty("Registry.Url"));
            }
        }
        try {
            data.setServerIp(NetworkUtils.getLocalHostname());
        } catch (SocketException e) {
            throw new AxisFault(e.getMessage(), e);
        }
        return data;
    }

    public String getServerDataAsString() throws Exception {
        try {
            return getServerData().toString();
        } catch (AxisFault e) {
            String msg = "Cannot get server data";
            log.error(msg, e);
            throw new Exception(msg);
        }
    }

    public String getServerVersion() {
        return ServerConfiguration.getInstance().getFirstProperty("Version");
    }

    public boolean restart() throws Exception {
        checkRestart();
        Thread thread = Thread.currentThread();
        ClassLoader originalClassloader = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(classloader);
            ConfigurationContext configurationContext = getConfigContext();
            final Controllable controllable =
                    (Controllable) configurationContext.
                            getProperty(ServerConstants.CARBON_INSTANCE);
            Thread th = new Thread() {
                public void run() {
                    try {
                        Thread.sleep(1000);
                        controllable.restart();
                    } catch (Exception e) {
                        String msg = "Cannot restart server";
                        log.error(msg, e);
                        throw new RuntimeException(msg, e);
                    }
                }
            };
            th.start();
            invalidateSession();
        } finally {
            thread.setContextClassLoader(originalClassloader);
        }
        return true;
    }

    public boolean restartGracefully() throws Exception {
        checkRestart();
        Thread thread = Thread.currentThread();
        ClassLoader originalClassloader = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(classloader);
            ConfigurationContext configurationContext = getConfigContext();
            final Controllable controllable =
                    (Controllable) configurationContext.
                            getProperty(ServerConstants.CARBON_INSTANCE);
            Thread th = new Thread() {
                public void run() {
                    try {
                        Thread.sleep(1000);
                        controllable.restartGracefully();
                    } catch (Exception e) {
                        String msg = "Cannot restart server";
                        log.error(msg, e);
                        throw new RuntimeException(msg, e);
                    }
                }
            };
            th.start();
            invalidateSession();
        } finally {
            thread.setContextClassLoader(originalClassloader);
        }
        return true;
    }

    private void checkRestart() throws Exception {
        if (!canRestart()) {
            String msg = "cannot.restart.derby.embedded.on.windows"; // key in FE Resource file
            throw new Exception(msg);
        }
    }

    /**
     * Check whether we can restart the server. On Windows machine with the Derby Embedded Driver,
     * restarting does not work since the file handles are not released.
     *
     * @return true, if the server can be restarted. false, otherwise.
     */
    private boolean canRestart() {
        if (System.getProperty("os.name").toLowerCase().indexOf("win") == -1) {
            return true;
        }

        String registryDBDriver = ServerAdminServiceComponent.getRegistryDatabaseDriver();
        boolean registryUsesDerbyEmbeddedDriver =
                registryDBDriver != null && registryDBDriver.indexOf("Derby Embedded") != -1;

        boolean userManagerUsesDerbyEmbeddedDriver =
                "org.apache.derby.jdbc.EmbeddedDriver".equals(ServerAdminServiceComponent.getUserManagerDriver());

        return !registryUsesDerbyEmbeddedDriver && !userManagerUsesDerbyEmbeddedDriver;
    }

    public boolean shutdown() {
        Thread thread = Thread.currentThread();
        ClassLoader originalClassloader = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(classloader);
            ConfigurationContext configurationContext = getConfigContext();
            final Controllable controllable =
                    (Controllable) configurationContext.
                            getProperty(ServerConstants.CARBON_INSTANCE);
            Thread th = new Thread() {
                public void run() {
                    try {
                        Thread.sleep(1000);
                        controllable.shutdown();
                    } catch (Exception e) {
                        String msg = "Cannot shutdown server";
                        log.error(msg, e);
                        throw new RuntimeException(msg, e);
                    }
                }
            };
            th.start();
            invalidateSession();
        } finally {
            thread.setContextClassLoader(originalClassloader);
        }
        return true;
    }

    public boolean shutdownGracefully() {
        Thread thread = Thread.currentThread();
        ClassLoader originalClassloader = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(classloader);
            ConfigurationContext configurationContext = getConfigContext();
            final Controllable controllable =
                    (Controllable) configurationContext.
                            getProperty(ServerConstants.CARBON_INSTANCE);
            Thread th = new Thread() {
                public void run() {
                    try {
                        Thread.sleep(1000);
                        controllable.shutdownGracefully();
                    } catch (Exception e) {
                        String msg = "Cannot gracefully shutdown server";
                        log.error(msg, e);
                        throw new RuntimeException(msg, e);
                    }
                }
            };
            th.start();
            invalidateSession();
        } finally {
            thread.setContextClassLoader(originalClassloader);
        }
        return true;
    }

    public void startMaintenance() throws Exception {
        Thread thread = Thread.currentThread();
        ClassLoader originalClassloader = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(classloader);
            try {
                org.wso2.carbon.core.ServerStatus.setServerInMaintenance();
            } catch (AxisFault e) {
                String msg = "Cannot set server to maintenance mode";
                log.error(msg, e);
            }
            Map inTransports = getAxisConfig().getTransportsIn();
            new ServerManagement(inTransports).startMaintenance();
        } finally {
            thread.setContextClassLoader(originalClassloader);
        }
    }

    public void endMaintenance() throws Exception {
        Thread thread = Thread.currentThread();
        ClassLoader originalClassloader = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(classloader);
            Map inTransports = getAxisConfig().getTransportsIn();
            new ServerManagement(inTransports).endMaintenance();
            try {
                org.wso2.carbon.core.ServerStatus.setServerRunning();
            } catch (AxisFault e) {
                String msg = "Cannot set server to running mode";
                log.error(msg, e);
            }
        } finally {
            thread.setContextClassLoader(originalClassloader);
        }
    }

    public boolean isAlive() {
        return true;
    }

    public String getServerStatus() throws Exception {
        try {
            return org.wso2.carbon.core.ServerStatus.getCurrentStatus();
        } catch (AxisFault e) {
            String msg = "Could not get current server status";
            log.error(msg);
            throw new Exception(msg, e);
        }
    }

    private void invalidateSession() {
        MessageContext msgCtx = MessageContext.getCurrentMessageContext();
        if (msgCtx == null) {
            return;
        }
        HttpServletRequest request =
                (HttpServletRequest) msgCtx.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
        try {
            if (request != null) {
                request.getSession().invalidate();
            }
        } catch (Exception ignored) { // Ignore invalidation of invalidated sessions
        }
    }

    private String getTime(long timeInSeconds) {
        long days;
        long hours;
        long minutes;
        long seconds;
        days = timeInSeconds / SECONDS_PER_DAY;
        timeInSeconds = timeInSeconds - (days * SECONDS_PER_DAY);
        hours = timeInSeconds / 3600;
        timeInSeconds = timeInSeconds - (hours * 3600);
        minutes = timeInSeconds / 60;
        timeInSeconds = timeInSeconds - (minutes * 60);
        seconds = timeInSeconds;

        return days + " day(s) " + hours + " hr(s) " + minutes + " min(s) " + seconds + " sec(s)";
    }
}
