package org.wso2.carbon.user.core;

public interface AuthorizationManager {

    /**
     * 
     * Checks for user authorization
     * 
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource Id String
     * @param action
     *            The action user is trying to perform
     * @return Returns true when user is authorized to perform the action on the
     *         resource and false otherwise.
     * @throws UserStoreException
     */
    public boolean isUserAuthorized(String userName, String resourceId, String action)
            throws UserStoreException;

    /**
     * Checks for role authorization.
     * 
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource Id String
     * @param action
     *            The action the role is trying to perform
     * @return Returns true when the role is authorized to perform the action on
     *         the resource and false otherwise
     * @throws UserStoreException
     */
    public boolean isRoleAuthorized(String roleName, String resourceId, String action)
            throws UserStoreException;

    /**
     * 
     * Retrieves a list of users allowed to perform the given action on the
     * resource
     * 
     * @param resourceId
     *            Resource Id String
     * @param action
     *            The action that is allowed to perform
     * @return Returns a list of users allowed to perform the given action on
     *         the resource
     * @throws UserStoreException
     * 
     * @deprecated
     */
    public String[] getExplicitlyAllowedUsersForResource(String resourceId, String action)
            throws UserStoreException;

    /**
     * 
     * Retrieves a list of roles allowed to perform the given action on the
     * resource
     * 
     * @param resourceId
     *            Resource Id String
     * @param action
     *            The action that is allowed to perform
     * @return Returns a list of roles allowed to perform the given action on
     *         the resource
     * @throws UserStoreException
     */
    public String[] getAllowedRolesForResource(String resourceId, String action)
            throws UserStoreException;

    public String[] getDeniedRolesForResource(String resourceId, String action)
            throws UserStoreException;

    /**
     * 
     * @param resourceId
     * @param action
     * @return
     * @throws UserStoreException
     * 
     * @deprecated
     */
    public String[] getExplicitlyDeniedUsersForResource(String resourceId, String action)
            throws UserStoreException;

    /**
     * Grants authorizations to a user to perform an action on a resource.
     * 
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the user
     * @throws UserStoreException
     * 
     * @deprecated
     */
    public void authorizeUser(String userName, String resourceId, String action)
            throws UserStoreException;

    /**
     * Grants authorizes to a role to perform an action on a resource.
     * 
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the role
     * @throws UserStoreException
     */
    public void authorizeRole(String roleName, String resourceId, String action)
            throws UserStoreException;

    /**
     * Deny authorizations to a user to perform an action on a resource.
     * 
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the user
     * @throws UserStoreException
     * 
     * @deprecated
     */
    public void denyUser(String userName, String resourceId, String action)
            throws UserStoreException;

    /**
     * Deny authorizations to a role to perform an action on a resource.
     * 
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted to the role
     * @throws UserStoreException
     */
    public void denyRole(String roleName, String resourceId, String action)
            throws UserStoreException;

    /**
     * Deletes an already granted authorization of a user.
     * 
     * @param userName
     *            The user name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted
     * @throws UserStoreException
     * 
     * @deprecated
     */
    public void clearUserAuthorization(String userName, String resourceId, String action)
            throws UserStoreException;

    /**
     * 
     * @param userName
     * @throws UserStoreException
     * 
     * @deprecated
     */
    public void clearUserAuthorization(String userName) throws UserStoreException;

    /**
     * Deletes an already granted authorization of a role.
     * 
     * @param roleName
     *            The role name
     * @param resourceId
     *            Resource identification string
     * @param action
     *            The action granted
     * @throws UserStoreException
     */
    public void clearRoleAuthorization(String roleName, String resourceId, String action)
            throws UserStoreException;

    /**
     * Deletes the role's right to perform the action on all resources.
     * 
     * @param roleName
     *            The role name
     * @param action
     *            The action granted
     * @throws UserStoreException
     */
    public void clearRoleActionOnAllResources(String roleName, String action)
            throws UserStoreException;

    /**
     * Used when deleting roles.
     * 
     * @param roleName
     * @throws UserStoreException
     */
    public void clearRoleAuthorization(String roleName) throws UserStoreException;

    /**
     * Deletes all granted authorization on a resource.
     * 
     * @param resourceId
     *            Resource identification string
     * @throws UserStoreException
     */
    public void clearResourceAuthorizations(String resourceId) throws UserStoreException;

    /**
     * Returns the complete set of UI resources allowed for User.
     * 
     * @param userName
     * @return
     * @throws UserStoreException
     */
    public String[] getAllowedUIResourcesForUser(String userName, String permissionRootPath)
            throws UserStoreException;


    /**
     * this will get the tenant id associated with the user authorization manager
     *
     * @return the tenant id of the authorization manager
     * @throws UserStoreException if the operation failed
     */
    public int getTenantId() throws UserStoreException;    

}
