/*
 *  Copyright (c) 2005-2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 *
 */
package org.wso2.carbon.user.core.tenant;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.apache.axiom.om.OMElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.UserCoreConstants;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.utils.DBUtils;
import org.osgi.framework.BundleContext;

public class JDBCTenantManager implements TenantManager {
    DataSource dataSource;
    private static Log log = LogFactory.getLog(TenantManager.class);

    protected String superTenantDomain = null;
    
    protected BundleContext bundleContext;
    
    @SuppressWarnings("unchecked")
    public JDBCTenantManager(OMElement omElement, Map<String, Object> properties) throws Exception {
        this.dataSource = (DataSource) properties.get(UserCoreConstants.DATA_SOURCE);
        if (dataSource == null) {
            throw new Exception("Data Source is null");
        }
        Iterator ite = omElement.getChildrenWithLocalName(UserCoreConstants.LOCAL_NAME_PROPERTY);
        if (ite != null && ite.hasNext()) {
            OMElement element = (OMElement) ite.next();
            this.superTenantDomain = element.getText();
        }
    }
    
    public JDBCTenantManager(DataSource dataSource, String superTenantDomain) {
        this.dataSource = dataSource;    
        this.superTenantDomain = superTenantDomain;
    }

    public int addTenant(Tenant tenant) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        ResultSet result = null;
        int id = 0;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.ADD_TENANT_SQL;
            
            String dbProductName = dbConnection.getMetaData().getDatabaseProductName();
            prepStmt = dbConnection.prepareStatement(sqlStmt,
					new String[] { DBUtils.getConvertedAutoGeneratedColumnName(
							dbProductName, "UM_ID") });
            prepStmt.setString(1, tenant.getDomain());
            prepStmt.setString(2, tenant.getEmail());

            prepStmt.executeUpdate();

            result = prepStmt.getGeneratedKeys();
            if (result.next()) {
                id = result.getInt(1);
            }
            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in adding tenant with " +
                    "tenant domain: " + tenant.getDomain() + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
        return id;
    }


    public void updateTenant(Tenant tenant) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.UPDATE_TENANT_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setString(1, tenant.getDomain());
            prepStmt.setString(2, tenant.getEmail());

            prepStmt.setInt(3, tenant.getId());

            prepStmt.executeUpdate();

            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in adding tenant with " +
                    "tenant domain: " + tenant.getDomain() + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
    }

    public Tenant getTenant(int tenantId) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        ResultSet result = null;
        Tenant tenant = null;
        int id;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.GET_TENANT_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setInt(1, tenantId);

            result = prepStmt.executeQuery();

            if (result.next()) {
                id = result.getInt("UM_ID");
                String domain = result.getString("UM_DOMAIN_NAME");
                String email = result.getString("UM_EMAIL");
                boolean active = result.getBoolean("UM_ACTIVE");

                tenant = new Tenant();
                tenant.setId(id);
                tenant.setDomain(domain);
                tenant.setEmail(email);
                tenant.setActive(active);
            }
            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in getting the tenant with " +
                    "tenant id: " + tenantId + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
        return tenant;
    }

    public Tenant[] getAllTenants() throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        ResultSet result = null;
        List<Tenant> tenantList = new ArrayList<Tenant>();
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.GET_ALL_TENANTS_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);

            result = prepStmt.executeQuery();

            while (result.next()) {
                int id = result.getInt("UM_ID");
                String domain = result.getString("UM_DOMAIN_NAME");
                String email = result.getString("UM_EMAIL");
                boolean active = result.getBoolean("UM_ACTIVE");

                Tenant tenant = new Tenant();
                tenant.setId(id);
                tenant.setDomain(domain);
                tenant.setEmail(email);
                tenant.setActive(active);
                tenantList.add(tenant);
            }
            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in getting the tenants.";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
        return tenantList.toArray(new Tenant[tenantList.size()]);
    }

    public String getDomain(int tenantId) throws UserStoreException {
        if (tenantId == 0) {
            return null;
        }
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        ResultSet result = null;
        String domain = null;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.GET_DOMAIN_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setInt(1, tenantId);

            result = prepStmt.executeQuery();

            if (result.next()) {
                domain = result.getString("UM_DOMAIN_NAME");
            }
            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in getting the tenant with " +
                    "tenant id: " + tenantId + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
        return domain;
    }

    public int getTenantId(String domain) throws UserStoreException {
        
        if (domain == null) {
            return 0;
        } else if (superTenantDomain != null && superTenantDomain.equals(domain)){
            return 0;
        } 
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        ResultSet result = null;
        int id = -1;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.GET_TENANT_ID_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setString(1, domain);

            result = prepStmt.executeQuery();

            if (result.next()) {
                id = result.getInt("UM_ID");
            }
            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in getting the tenant id with " +
                    "tenant domain: " + domain + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (result != null) {result.close();}
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
        return id;
    }


    public void activateTenant(int tenantId) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.ACTIVATE_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setInt(1, tenantId);

            prepStmt.executeUpdate();

            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in activating the tenant with " +
                    "tenant id: " + tenantId + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
    }

    public void deactivateTenant(int tenantId) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.DEACTIVATE_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setInt(1, tenantId);

            prepStmt.executeUpdate();

            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in activating the tenant with " +
                    "tenant id: " + tenantId + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
    }

    public boolean isTenantActive(int tenantId) throws UserStoreException {
        if (tenantId == 0) {
            // this is always active..
            return true;
        }
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.IS_TENANT_ACTIVE_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setInt(1, tenantId);

            ResultSet result = prepStmt.executeQuery();

            if (result.next()) {
                boolean isActive = result.getBoolean("UM_ACTIVE");
                return isActive;
            }

            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in activating the tenant with " +
                    "tenant id: " + tenantId + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
        return false;
    }



    public void deleteTenant(int tenantId) throws UserStoreException {
        Connection dbConnection = null;
        PreparedStatement prepStmt = null;
        try {
            dbConnection = getDBConnection();
            String sqlStmt = TenantConstants.DELETE_TENANT_SQL;
            prepStmt = dbConnection.prepareStatement(sqlStmt);
            prepStmt.setInt(1, tenantId);

            prepStmt.executeUpdate();
            dbConnection.commit();
        } catch (SQLException e) {
            String msg = "Error in activating the tenant with " +
                    "tenant id: " + tenantId + ".";
            log.error(msg);
            throw new UserStoreException(e);
        }
        finally {
            try {
                if (prepStmt != null) {prepStmt.close();}
                if (dbConnection != null) {dbConnection.close();}
            }
            catch (SQLException ex)  {
                String msg = "Error in closing the connection for result and prepared statement. " + ex.getMessage();
                log.error(msg,ex);
                throw new UserStoreException(msg,ex);
            }
        }
    }
    
    
    public void setBundleContext(BundleContext bundleContext) {
        this.bundleContext = bundleContext;
    }

    private Connection getDBConnection() throws SQLException {
        Connection dbConnection = dataSource.getConnection();
        dbConnection.setAutoCommit(false);
        return dbConnection;
    }

    public String getSuperTenantDomain() throws UserStoreException {
        return superTenantDomain;
    }
}
